#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# Copyright (c) 2019-2024, AllWorldIT
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""IDMS Linux installer package."""

# pylint: disable=too-many-lines

import argparse
import logging
import os
import pathlib
import re
import shutil
import tempfile
import time
from collections.abc import Callable, MutableMapping
from typing import Any

from natsort import natsorted
from prompt_toolkit.shortcuts import checkboxlist_dialog, input_dialog, message_dialog, radiolist_dialog

from . import util
from .exceptions import IliRuntimeError
from .filesystem import IliFilesystem, IliPath
from .version import __version__

__all__ = ["IliBase", "__version__"]


RequiredConfigItem = dict[str, Any] | list[dict[str, Any]]
RequiredConfigItems = dict[int, RequiredConfigItem]


class IliBase:  # pylint: disable=too-many-instance-attributes,too-many-public-methods
    """IDMS Linux Installer class."""

    # Commandline arguments
    _args: argparse.Namespace
    # Internal config during the configuration stage
    _config: dict[str, Any]
    # Profile configurations we've pulled from file
    _profile: MutableMapping[str, Any] | None
    # Configuration items requested by targets
    _required_config: RequiredConfigItems

    # Imaging/installation stages
    _stages: dict[int, dict[str, Any]]

    # Root filesystem
    _rootfs: IliFilesystem | None
    # Mounted filesystems
    _mounts: list[IliFilesystem]
    # Kpartx devices we've created
    _kpartx_devices: list[list[Any]]
    # Loop devices we've created
    _loop_devices: list[IliPath]
    # Encrypted devices we've created
    _luks_volumes: list[str]
    # LVM devices we've created
    _lvm_vgs: list[str]
    _lvm_lvs: list[str]
    # MD raid devices we've created
    _mdraid_devices: list[IliPath]
    # Temporary paths we need to remove on exit
    _temp_paths: list[IliPath]

    # Tuples of source:destination directories we're going to be copying to the target
    _overlay_directories: list[tuple[IliPath, IliPath]]

    # Target properties
    _target_staging: bool
    _target_pacman_conf: IliPath | None
    _target_packages: list[str]
    _target_services: list[str]
    _target_hostname: str | None
    _target_root_password: str | None
    _target_sudo_username: str | None
    _target_sudo_password: str | None
    _target_root_autologin: bool
    _target_install_microcode: bool
    _target_networking: bool

    # Network devices to configure
    _target_network_devices: list[str]
    _target_network_config: dict[str, str]

    # Globs we need to remove from the target during sanitization
    _rootfs_remove_globs: list[str]

    def __init__(self, args: argparse.Namespace, profile: MutableMapping[str, Any] | None) -> None:
        """Object initialization."""

        # Commandline arguments
        self._args = args
        # Internal config during the configuration stage
        self._config = {}
        # Profile configurations we've pulled from file
        self._profile = profile
        # Configuration items requested by targets
        self._required_config = {}
        # Imaging/installation stages
        self._stages = {}

        # Root filesystem
        self._rootfs = None
        # Mounted filesystems
        self._mounts = []
        # Kpartx devices we've created
        self._kpartx_devices = []
        # Loop devices we've created
        self._loop_devices = []
        # Encrypted devices we've created
        self._luks_volumes = []
        # LVM devices we've created
        self._lvm_vgs = []
        self._lvm_lvs = []
        # MD devices we've created
        self._mdraid_devices = []
        # Temporary paths we need to remove on exit
        self._temp_paths = []

        # Overlay directories
        self._overlay_directories = []

        # Target properties
        self._target_staging = False
        self._target_pacman_conf = None
        self._target_packages = ["base"]
        self._target_services = ["systemd-networkd", "systemd-resolved"]
        self._target_hostname = None
        self._target_root_password = None
        self._target_sudo_username = None
        self._target_sudo_password = None
        self._target_root_autologin = False

        self._target_install_microcode = False  # Don't install AMD and Intel microcode by default
        self._target_networking = True
        self._target_network_devices = []
        self._target_network_config = {}

        # List of path globs to remove
        self._rootfs_remove_globs = [
            "/etc/machine-id",
            "/etc/pacman.d/gnupg/openpgp-revocs.d/*",
            "/etc/pacman.d/gnupg/private-keys-v1.d/*",
            "/etc/pacman.d/gnupg/pubring.gpg~",
            "/etc/pacman.d/gnupg/S.*",
            # We'll be killing gpg-agent, so remove the lock file incase its left behind
            "/etc/pacman.d/gnupg/*.lock",
            "/root/*",
            "/tmp/*",  # noqa: S108 # nosec
            "/var/cache/pacman/pkg/*",
            "/var/lib/pacman/sync/*",
            "/var/tmp/*",  # noqa: S108 # nosec
        ]

        # Initialize the target
        self.target_init()

    def target_init(self) -> None:
        """
        Target initialization.

        This method is called during __init__().

        The purpose of this method is to set sub-class properties, set property defaults and add configuration stages.
        """

        # Prepare user input required
        self.add_stage(1000, "Prepare configuration", self.prepare_configure)

        self.add_stage(1100, "Configure installer", self.configure)
        self.add_stage(1300, "Configure installer (user)", self.configure_userinput)
        self.add_stage(1400, "Post configuration", self.post_configure)
        self.add_stage(1900, "Validate configuration", self.validate_configuration)

        self.add_stage(2100, "Prepare target", self.prepare_target)
        # 2110 is in image_generic and creates the filesystems
        self.add_stage(2500, "Mount target", self.mount_target)

        self.add_stage(3100, "Pacstrap target", self.pacstrap_target)
        self.add_stage(3200, "Setup up locale", self.setup_target_locale)
        self.add_stage(3300, "Setup up timezone", self.setup_target_timezone)
        self.add_stage(3400, "Setup up resolv.conf", self.setup_target_resolv_conf)
        self.add_stage(
            3500,
            "Setup up pacman keyring init service",
            self.setup_target_pacman_init_service,
        )

        # 4100 network configuration

        self.add_stage(6100, "Setup hostname", self.setup_target_hostname)
        self.add_stage(6200, "Setup users", self.setup_target_users)
        self.add_stage(6300, "Setup root autologin", self.setup_target_root_autologin)
        self.add_stage(6400, "Setup networking", self.setup_target_networking)

        # For custom targets, hook in here

        self.add_stage(8100, "Sanitize target", self.sanitize_target)
        self.add_stage(8200, "Add overlay directories", self.setup_target_overlay_directories)
        self.add_stage(8300, "Enable target services", self.setup_target_systemd_services)
        self.add_stage(8400, "Cleanup mounts", self.cleanup_mounts)
        self.add_stage(8410, "Cleanup LVM logical volumes", self.cleanup_lvm_logical_volumes)
        self.add_stage(8415, "Cleanup LVM volume groups", self.cleanup_lvm_volume_groups)
        self.add_stage(8420, "Cleanup encrypted devices", self.cleanup_luks_volumes)
        self.add_stage(8430, "Cleanup kpartx", self.cleanup_kpartx)
        self.add_stage(8440, "Cleanup loop devices", self.cleanup_loop_devices)
        self.add_stage(8450, "Cleanup MDRAID devices", self.cleanup_mdraid_devices)

        self.add_stage(8800, "Finalize target", self.finalize_target)

        self.add_stage(9500, "Cleanup", self.cleanup)

        self.add_stage(9999, "Done", self.done)

    def add_stage(self, order: int, name: str, method: Callable[[], None]) -> None:
        """Add a run stage."""
        self.stages[order] = {
            "name": name,
            "method": method,
        }

    def add_required_config(self, order: int, config_item: RequiredConfigItem) -> None:
        """Add a required config item."""
        self.required_config[order] = config_item

    def add_package(self, package_name: str) -> None:
        """Add a package to install list."""

        logging.debug("  * Adding package: %s", package_name)

        if package_name not in self.target_packages:
            self.target_packages.append(package_name)

    def add_service(self, service_name: str) -> None:
        """Add service name to enable."""

        logging.debug("  * Adding enable service: %s", service_name)

        if service_name not in self.target_services:
            self.target_services.append(service_name)

    def add_network_device(self, device_name: str) -> None:
        """Add network device to configure."""

        logging.debug("  * Adding network device: %s", device_name)

        if device_name not in self.target_network_devices:
            self.target_network_devices.append(device_name)

    def add_temp_path(self, path: IliPath) -> None:
        """Add a temp path we created that needs to be removed."""

        logging.debug("  * Adding temp path for cleanup: %s", path)

        self.temp_paths.append(path)

    def add_rootfs_remove_glob(self, pathglob: str) -> None:
        """Add a glob to be removed from the rootfs."""

        logging.debug("  * Adding rootfs remove glob: %s", pathglob)

        # Sanity checks
        if not pathglob.startswith("/"):
            raise IliRuntimeError(f"Glob for rootfs remove '{pathglob}' does not start with '/'")
        if ".." in pathglob:
            raise IliRuntimeError(f"Glob for rootfs remove '{pathglob}' must not contain '..'")
        self.rootfs_remove_globs.append(pathglob)

    def add_loop_device(self, image_file: IliPath) -> IliPath:
        """Add a loop device."""

        logging.debug("  * Adding loop device for image file: %s", image_file)

        # Add loop device
        losetup = util.run_host(["/usr/bin/losetup", "--find", "--show", str(image_file)])

        # Work out what the device is and if its valid
        loop_device = IliPath(losetup[0])
        if not re.match("^/dev/loop[0-9]+", str(loop_device)):
            raise IliRuntimeError(f"Output from losetup not understood: {losetup}")
        # Add loop device
        self.loop_devices.append(loop_device)

        logging.debug("  * Loop device created: %s", loop_device)

        return loop_device

    def remove_loop_device(self, loop_device: IliPath) -> None:
        """Remove a loop device."""

        logging.debug("  * Removing loop device: %s", loop_device)

        # Remove loop device
        util.run_host(["/usr/bin/losetup", "-d", str(loop_device)])

        # Grab index, and remove it below
        loop_device_index = self.loop_devices.index(loop_device)
        del self.loop_devices[loop_device_index]

    def add_kpartx_device(self, loop_device: IliPath) -> None:
        """Add kpartx to loop device."""

        logging.debug("  * Adding device mapper devices for: %s", loop_device)

        kpartx_output = util.run_host(["/usr/bin/kpartx", "-av", str(loop_device)])
        # Pull out column 3, which is the device created
        devices_created = []
        for line in kpartx_output:
            # Check output line is in right format
            if not re.match("^add map", line):
                raise IliRuntimeError("Did not understand the output of kpartx")
            # Add map to list
            devices_created.append(line.split(" ")[2])

        logging.debug("  * Device mapper devices added: %s", ", ".join(devices_created))

        # Add loop devices and devices created to our kpartx_devices
        self.kpartx_devices.append([loop_device, devices_created])

    def create_encrypted_device(  # pylint: disable=too-many-arguments
        self, name: str, device: IliPath, uuid: str, password: str, luks_version: int = 2
    ) -> None:
        """Create an encrypted device."""

        logging.debug("  - Creating encrypted device '%s' on '%s'", name, device)

        # Create temporary file
        with tempfile.NamedTemporaryFile() as password_file:
            password_file.write(password.encode("UTF-8"))
            password_file.flush()

            # Format the device
            util.run_host(
                [
                    "/usr/bin/cryptsetup",
                    "luksFormat",
                    "--verbose",
                    "--batch-mode",
                    "--type",
                    f"luks{luks_version}",
                    "--uuid",
                    uuid,
                    str(device),
                    password_file.name,
                ]
            )

            # Open the device
            util.run_host(
                [
                    "/usr/bin/cryptsetup",
                    "open",
                    "--verbose",
                    "--batch-mode",
                    "--type",
                    "luks",
                    "--key-file",
                    password_file.name,
                    str(device),
                    name,
                ]
            )

        # Add to list of created LUKS devices
        self.luks_volumes.append(name)

    def create_lvm_physical_volume(self, device: IliPath) -> None:  # pylint: disable=useless-option-value
        """Create a LVM physical volume."""

        logging.debug("  - Creating LVM physical volume: %s", device)

        util.run_host(["/usr/bin/pvcreate", str(device)])
        # We don't track what PV's we create as they don't need cleanup

    def create_lvm_volume_group(self, vg_name: str, pv_device: IliPath) -> None:
        """Create a LVM volume group."""

        logging.debug(
            "  - Creating LVM volume group '%s' on physical volume '%s'",
            vg_name,
            pv_device,
        )

        util.run_host(["/usr/bin/vgcreate", "--physicalextentsize", "32M", vg_name, str(pv_device)])
        # Add to list of created volume groups
        self.lvm_vgs.append(vg_name)

    def create_lvm_logical_volume(self, lv_name: str, vg_name: str, size: str) -> None:
        """Create a LVM logical volume."""

        logging.debug(
            "  - Creating LVM logical volume '%s' on volume group '%s' with size '%s'",
            lv_name,
            vg_name,
            size,
        )

        # Work out the size option we're going to use
        size_opt = "--extents" if size.find("%") > 0 else "--size"

        # Creat the logical volume
        util.run_host(["/usr/bin/lvcreate", "--wipesignatures", "y", "--yes", size_opt, size, "--name", lv_name, vg_name])

        # Add to list of created logical volumes
        self.lvm_lvs.append(f"{vg_name}/{lv_name}")

    def create_mdraid(self, md_device: IliPath, uuid: str, level: int, devices: list[IliPath]) -> None:
        """Create software raid array."""

        # Work out how many devices we have
        num_devices = len(devices)

        logging.debug(
            "  - Creating MDRAID device '%s' [UUID: %s] with level %s on: %s",
            md_device,
            uuid,
            level,
            ", ".join([str(x) for x in devices]),
        )

        # Run mdadm to create the RAID device
        mdadm_hostname = self.target_hostname or "localhost.localdomain"
        devices_str = [str(x) for x in devices]
        util.run_host(
            [
                "/usr/bin/mdadm",
                "--create",
                str(md_device),
                "--uuid",
                uuid,
                "--level",
                str(level),
                "--raid-devices",
                str(num_devices),
                "--metadata",
                "default",
                "--homehost",
                mdadm_hostname,
                *devices_str,
            ]
        )

        # Add to to list of created software raids
        self.mdraid_devices.append(md_device)

        # Wipe filesystem
        self.wipe_fs(md_device)

    def run(self) -> None:
        """Run the installer."""

        # Grab number of stages
        num_stages = len(self.stages.items())

        # Loop with stages and launch
        stage_num = 1
        for prio, stage in sorted(self.stages.items()):
            logging.info(
                "===> Launching stage %2i/%2i: %s [%s]",
                stage_num,
                num_stages,
                stage["name"],
                prio,
            )
            stage["method"]()
            logging.debug("===> Stage completed: %s", stage["name"])
            # Bump stage number
            stage_num += 1

    def cleanup(self) -> None:
        """Cleanup after install or exception."""

        self.cleanup_mounts()

        self.cleanup_lvm_logical_volumes()
        self.cleanup_lvm_volume_groups()

        self.cleanup_luks_volumes()

        self.cleanup_kpartx()
        self.cleanup_loop_devices()

        self.cleanup_mdraid_devices()

        self.cleanup_temp_paths()

    def done(self) -> None:  # pylint: disable=useless-option-value
        """Run last to indicate we completed successfully."""
        logging.info("All tasks completed!")

    def prepare_configure(self) -> None:
        """Set up the configuration we need."""

        self.add_required_config(
            3100,
            {
                "name": "hostname",
                "type": "hostname",
                "property": "target_hostname",
                "required": True,
                "title": "Hostname",
                "prompt": "Hostname",
            },
        )

        self.add_required_config(
            6200,
            {
                "name": "root_password",
                "type": "password",
                "property": "target_root_password",
                "title": "Root password",
                "prompt": "Root password (blank = lock, space = none)",
            },
        )

        self.add_required_config(
            6300,
            {
                "name": "sudo_username",
                "type": "username",
                "property": "target_sudo_username",
                "title": "Sudo username",
                "prompt": "Sudo username (blank = no sudo user)",
                "dynamic_config_items": self._get_sudo_config_items,
            },
        )

    def configure(  # noqa: C901,PLR0912,PLR0915 # pylint: disable=too-many-statements,too-many-locals,too-many-branches
        self,
    ) -> None:
        """
        Configure items we require.

        This should be the last in the call chain and will ask the user for input where required.
        """

        # Work out if we're using staging repositories or not
        staging = self.get_config_from_args_profile({"name": "staging"})
        if staging is not None:
            logging.info(
                "  - Using commandline/profile for value 'staging': %s [default: %s]",
                staging,
                self.target_staging,
            )
            self.target_staging = staging
        else:
            logging.info("  - Using default value for 'staging': %s", self.target_staging)
        # Set the default pacman configuration file to use
        if self.target_staging:
            self.target_pacman_conf = IliPath("/usr/share/devtools/pacman.conf.d/extra-staging.conf")
        else:
            self.target_pacman_conf = IliPath("/usr/share/devtools/pacman.conf.d/extra.conf")

        # Check if the user has specified a pacman configuration file to use
        pacman_conf = self.get_config_from_args_profile({"name": "pacman_config"})
        if pacman_conf is not None:
            logging.info(
                "  - Using commandline/profile for value 'pacman_config': %s [default: %s]",
                pacman_conf,
                self.target_pacman_conf,
            )
            self.target_pacman_conf = pacman_conf
        else:
            logging.info(
                "  - Using default value for 'pacman_config': %s",
                self.target_pacman_conf,
            )

        # Check if we have packages to add to the install
        packages = self.get_config_from_args_profile({"name": "packages", "multivalue": True})
        if packages is not None:
            logging.info("  - Adding package list: %s", packages)
            for pkgs in packages:
                # Allow comma separated lists
                for pkg in pkgs.split(","):
                    self.add_package(pkg)

        # Check if we have services to enable
        services = self.get_config_from_args_profile({"name": "services", "multivalue": True})
        if services is not None:
            for service_list in services:
                logging.info("  - Adding service list: %s", service_list)
                # Allow comma separated lists
                for service in service_list.split(","):
                    self.add_service(service)

        # hostname comes from user prompt

        # root_password comes from user prompt

        # sudo_username comes from user prompt

        # sudo_password comes from user prompt

        # Check if we are setting autologin
        root_autologin = self.get_config_from_args_profile({"name": "root_autologin"})
        if root_autologin is not None:
            logging.info(
                "  - Using commandline/profile for value 'root_autologin': %s [default: %s]",
                root_autologin,
                self.target_root_autologin,
            )
            self.target_root_autologin = root_autologin
        else:
            logging.info(
                "  - Using default value for 'root_autologin': %s",
                self.target_root_autologin,
            )

        # Check if we have files to add to the target
        overlay_directory = self.get_config_from_args_profile({"name": "overlay_directory", "multivalue": True})
        if overlay_directory is not None:
            logging.info("  - Using overlay directory: %s", overlay_directory)
            for overlays in overlay_directory:
                # Allow comma separated lists
                for overlay in overlays.split(","):
                    # Split the overlay src and dst based on :
                    src, dst = overlay.split(":")
                    # If all checks out, add
                    self.overlay_directories.append((IliPath(src), IliPath(dst)))

        # Check if we are installing microcode updates
        install_microcode = self.get_config_from_args_profile({"name": "install_microcode"})
        if install_microcode is not None:
            logging.info(
                "  - Using commandline/profile for value 'install_microcode': %s [default: %s]",
                install_microcode,
                self.target_install_microcode,
            )
            self.target_install_microcode = install_microcode
        else:
            logging.info(
                "  - Using default value for 'install_microcode': %s",
                self.target_install_microcode,
            )
        # Output if its enabled or not
        if self.target_install_microcode:
            self.add_package("amd-ucode")
            self.add_package("intel-ucode")

        # Check if we're going to automatically resize the rootfs
        networking = self.get_config_from_args_profile({"name": "networking"})
        if networking is not None:
            logging.info(
                "  - Using commandline/profile for value 'networking': %s [default: %s]",
                networking,
                self.target_networking,
            )
            self.target_networking = networking
        else:
            logging.info("  - Using default value for 'networking': %s", self.target_networking)

        # Output if networking is enabled or disabled
        if self.target_networking:
            # Check if we have a list of network devices
            network_devices = self.get_config_from_args_profile({"name": "network_devices", "multivalue": True})
            if network_devices is not None:
                new_network_devices = []
                # Loop with the network devices we got from the commandline or profile config
                for device_list in network_devices:
                    # Split device list on , and add to new_network_devices if not empty
                    new_network_devices.extend([device_name for device_name in device_list.split(",") if device_name])
                # If we got a list of network devices, then add them
                if new_network_devices:
                    logging.info(
                        "  - Using commandline/profile for value 'network_devices': %s [default: %s]",
                        new_network_devices,
                        self.target_network_devices,
                    )
                    self.target_network_devices = new_network_devices
                    # Generate network configuration items for the above and add to the required configuration
                    self.add_required_config(4100, self._get_network_device_config_items(""))
        else:
            logging.info("  - Networking disabled")

    def post_configure(self) -> None:
        """Post configuration."""

    def configure_userinput(self) -> None:  # noqa: C901,PLR0912 # pylint: disable=too-many-branches
        """
        Configure items we require by getting them from the user.

        This should be the last in the call chain and will ask the user for input where required.
        """

        required_config_list = [x for _, x in sorted(self.required_config.items())]
        while True:
            # If we have no more items to configure, break
            if not required_config_list:
                break
            # Grab the first item in the list
            config_item = required_config_list.pop(0)

            # Check if this item is a list, if it is, then add the contents, to the beginning of required_config_list
            # and continue
            if isinstance(config_item, list):
                required_config_list = [*config_item, *required_config_list]
                continue

            # Grab default value
            default_value = None
            if "property" in config_item:
                default_value = getattr(self, config_item["property"])
                # If default_value is a dict, then grab the config item named value
                if default_value is not None and isinstance(default_value, dict):
                    # Make sure the key is in the dict first
                    # If its not, it means we didn't get a default value
                    default_value = default_value.get(config_item["name"])

            # First step is to see if the value comes from the command line or profile
            value = self.get_config_from_args_profile(config_item)
            if value is not None:
                logging.info(
                    "  - Using commandline/profile for value '%s': %s [default: %s]",
                    config_item["name"],
                    util.safe_key_value_output(config_item["name"], value),
                    util.safe_key_value_output(config_item["name"], default_value),
                )

            # If we don't have a value, check if this config_item is linked to a property
            if value is None and "property" in config_item:
                # If it is, try get the  value
                value = getattr(self, config_item["property"])
                # If this is a dict, pull the key value out
                if isinstance(value, dict):
                    # If it didn't exist, then obviously the value is None
                    value = value.get(config_item["name"])
                # If the value is defined we can use it as a default
                if value is not None:
                    logging.info(
                        "  - Using default value for '%s': %s",
                        config_item["name"],
                        util.safe_key_value_output(config_item["name"], value),
                    )

            # If we still don't have a value, prompt the user for it
            if value is None:
                value = self.get_config_from_user(config_item)
                # Check if value is specified by user
                if value is not None:
                    logging.info(
                        "  - User specified value for '%s': %s",
                        config_item["name"],
                        util.safe_key_value_output(config_item["name"], value),
                    )
                # If value is None it means that the prompt was skipped by one of our dynamic value methods
                # ie. the dynamic_values method returned an empty list to display to the user, so no selection was possible
                else:
                    continue

            # Check if we need to set a property
            if "property" in config_item:
                # Grab the property so we can check if its a dict
                property_value = getattr(self, config_item["property"])
                # If this is a dict, all we need to do is set the key to the value as its a reference
                if isinstance(property_value, dict):
                    property_value[config_item["name"]] = value
                # Else we need to set the property itself
                else:
                    setattr(self, config_item["property"], value)

            # Check if there is a method to call after we have retrieved the value
            if "dynamic_config_items" in config_item:
                args = [value]
                # Check if we have an additional data argument to pass to the callback
                if "dynamic_config_items_data" in config_item:
                    args.append(config_item["dynamic_config_items_data"])
                # Grab additional config_items we need
                more_config_items: list[Any] = config_item["dynamic_config_items"](*args)
                # Add the additional config items to the list
                required_config_list = more_config_items + required_config_list

        # If we have a sudo username, we need to add sudo to the package install list
        if self.target_sudo_username:
            self.add_package("sudo")

    def validate_configuration(self) -> None:
        """Validate configuration."""

        if self.target_pacman_conf is None:
            raise IliRuntimeError("target_pacman_conf must be set")

        # Check the pacman config specified exists
        if not self.target_pacman_conf.exists():
            raise IliRuntimeError(f"Pacman config '{self.target_pacman_conf}' does not exist")
        if not self.target_pacman_conf.is_file():
            raise IliRuntimeError(f"Pacman config '{self.target_pacman_conf}' is not a file")

        # Check that the overlay configuration is ok
        for src, dst in self.overlay_directories:
            # Make sure DST starts with /
            if not dst.parts[0] == "/":
                raise IliRuntimeError(f"Overlay destination '{dst}' dst must start with '/'")
            # Check the source exists and is a directory
            if not src.exists():
                raise IliRuntimeError(f"Overlay source '{src}' does not exist")

        # Check network devices are valid'ish
        for network_device in self.target_network_devices:
            # Make sure its not blank
            if not network_device:
                raise IliRuntimeError(f"Network device '{network_device}' is invalid")

    def mount(self, filesystem: IliFilesystem) -> None:
        """Mount a filesystem."""

        logging.debug(
            "    - Mounting filesystem '%s' [%s] on '%s'",
            filesystem.source,
            filesystem.fstype,
            filesystem.mount_point,
        )

        filesystem.mount()
        self.mounts.append(filesystem)

    def unmount(self, filesystem: IliFilesystem) -> None:
        """Unmount a filesystem."""

        logging.debug("  - Unmounting filesystem '%s'", filesystem.mount_point)

        filesystem_index = self.mounts.index(filesystem)
        filesystem.unmount()
        del self.mounts[filesystem_index]

    def prepare_target(self) -> None:
        """
        Prepare target.

        This method should be preparing the disk/image/directory for mount.

        This method should be setting self.rootfs which is required for the next step.

        This method must be overridden in child classes.
        """
        raise NotImplementedError("mount_target must be implemented per target")

    def mount_target(self) -> None:
        """
        Mount target.

        This method must be overridden in child classes.
        """

        logging.debug("  - Mounting rootfs")

        # Sanity checks
        if not self.rootfs:
            raise IliRuntimeError("No rootfs")

        # Mount filesystems
        self.mount(self.rootfs)

    def pacstrap_target(self) -> None:
        """Prepare target for installation."""

        if self.rootfs is None:
            raise IliRuntimeError("rootfs must be set")
        if self.rootfs.mount_point is None:
            raise IliRuntimeError("rootfs mount_point must be set")
        if self.target_pacman_conf is None:
            raise IliRuntimeError("target_pacman_conf must be set")

        # Pacstrap the target
        util.run_host(
            [
                "/usr/bin/pacstrap",
                "-C",
                str(self.target_pacman_conf),
                "-c",
                "-G",
                "-M",
                str(self.rootfs.mount_point),
                *self.target_packages,
            ]
        )

        # Copy in pacman configuration file if its not the same one
        pacman_conf_path = self.rootfs.mount_point.concatpath("etc", "pacman.conf")
        if self.target_pacman_conf != pacman_conf_path:
            logging.debug("  - Creating /etc/pacman.conf")
            try:
                shutil.copy(self.target_pacman_conf, pacman_conf_path)
            except OSError as err:
                raise IliRuntimeError(f"Unable to copy file '{self.target_pacman_conf}' to '{pacman_conf_path}': {err}") from None
        else:
            logging.debug("  - Using custom /etc/pacman.conf")

    def setup_target_locale(self) -> None:
        """Set up target locale."""

        if self.rootfs is None:
            raise IliRuntimeError("rootfs must be set")
        if self.rootfs.mount_point is None:
            raise IliRuntimeError("rootfs mount_point must be set")

        locale = "en_US.UTF-8"

        logging.debug("  - Setting locale: %s", locale)

        # Write out locale.gen
        locale_gen_lines = []
        locale_gen_path = self.rootfs.mount_point.concatpath("etc", "locale.gen")
        try:
            with locale_gen_path.open(encoding="UTF-8") as locale_gen_file:
                locale_gen_lines = locale_gen_file.readlines()
        except OSError as err:
            raise IliRuntimeError(f"Unable to open file '{locale_gen_path}' for reading: {err}") from None

        try:
            with locale_gen_path.open(mode="w", encoding="UTF-8") as locale_gen_file:
                # Loop with lines
                for line in locale_gen_lines:
                    # Uncomment the locale we want
                    file_line = line
                    if file_line.startswith(f"#{locale}"):
                        file_line = file_line[1:].strip() + "\n"
                    # Write out the rest
                    locale_gen_file.write(file_line)
        except OSError as err:
            raise IliRuntimeError(f"Unable to open file '{locale_gen_path}' for writing: {err}") from None

        # Write out locale.conf
        locale_conf_path = self.rootfs.mount_point.concatpath("etc", "locale.conf")
        try:
            with locale_conf_path.open(mode="w", encoding="UTF-8") as conffile:
                conffile.write("LANG=en_US.UTF-8\n")
        except OSError as err:
            raise IliRuntimeError(f"Unable to open file '{locale_conf_path}' for writing: {err}") from None

        # Generate locales
        self.run_target(["/usr/bin/locale-gen"])

    def setup_target_timezone(self) -> None:
        """Set up target timezone."""

        if self.rootfs is None:
            raise IliRuntimeError("rootfs must be set")
        if self.rootfs.mount_point is None:
            raise IliRuntimeError("rootfs mount_point must be set")

        timezone = "UTC"
        timezone_path = self.rootfs.mount_point.concatpath("usr", "share", "zoneinfo", timezone)
        timezone_relpath = f"../usr/share/zoneinfo/{timezone}"

        logging.debug("  - Setting timezone: %s [%s]", timezone, timezone_relpath)

        # Check the zoneinfo file exists before symlinking
        if not timezone_path.is_file():
            raise IliRuntimeError("Invalid timezone: %s")

        # Setup symlink to timezone
        rootfs_timezone_path = self.rootfs.mount_point.concatpath("etc", "localtime")
        try:
            rootfs_timezone_path.symlink_to(timezone_relpath)
        except OSError as err:
            raise IliRuntimeError(f"Failed to symlink '{timezone_relpath}' to '{rootfs_timezone_path}': {err}") from None

    def setup_target_resolv_conf(self) -> None:
        """Set up target resolv.conf."""

        if self.rootfs is None:
            raise IliRuntimeError("rootfs must be set")
        if self.rootfs.mount_point is None:
            raise IliRuntimeError("rootfs mount_point must be set")

        resolv_conf_path = self.rootfs.mount_point.concatpath("etc", "resolv.conf")
        resolv_conf_relpath = "../run/systemd/resolve/stub-resolv.conf"

        logging.debug("  - Setting resolv.conf symlink: %s", resolv_conf_relpath)

        # If path exists, remove it
        if resolv_conf_path.exists():
            resolv_conf_path.unlink()
        # Then link systemd's networkd resolv.conf
        try:
            resolv_conf_path.symlink_to(resolv_conf_relpath)
        except OSError as err:
            raise IliRuntimeError(f"Failed to symlink '{resolv_conf_relpath}' to '{resolv_conf_path}': {err}") from None

    def setup_target_pacman_init_service(self) -> None:
        """Set up target pacman keyring."""

        if self.rootfs is None:
            raise IliRuntimeError("rootfs must be set")
        if self.rootfs.mount_point is None:
            raise IliRuntimeError("rootfs mount_point must be set")

        logging.debug("  - Creating pacman init service")

        # Create systemd dir
        rootfs_systemd_system_path = self.rootfs.mount_point.concatpath("usr", "local", "lib", "systemd", "system")
        try:
            rootfs_systemd_system_path.mkdir(mode=0o0755, parents=True, exist_ok=True)
        except OSError as err:
            raise IliRuntimeError(f"Failed to create systemd system directory '{rootfs_systemd_system_path}': {err}") from None

        # Create systemd target dir
        rootfs_systemd_system_wants_path = rootfs_systemd_system_path.concatpath("multi-user.target.wants")
        try:
            rootfs_systemd_system_wants_path.mkdir(mode=0o0755, parents=True, exist_ok=True)
        except OSError as err:
            raise IliRuntimeError(
                f"Failed to create systemd multi-user.target.wants directory '{rootfs_systemd_system_wants_path}': {err}"
            ) from None

        # Create pacman init service
        pacman_init_service_path = rootfs_systemd_system_path.concatpath("pacman-init.service")
        try:
            with pacman_init_service_path.open(mode="w", encoding="UTF-8") as pacman_init_service_file:
                pacman_init_service_file.write("[Unit]\n")
                pacman_init_service_file.write("Description=Initializes Pacman keyring\n")
                pacman_init_service_file.write("After=etc-pacman.d-gnupg.mount\n")
                pacman_init_service_file.write("ConditionPathExistsGlob=!/etc/pacman.d/gnupg/private-keys-v1.d/*.key\n")
                pacman_init_service_file.write("\n")
                pacman_init_service_file.write("[Service]\n")
                pacman_init_service_file.write("Type=oneshot\n")
                pacman_init_service_file.write("RemainAfterExit=yes\n")
                pacman_init_service_file.write("ExecStart=/usr/bin/pacman-key --init\n")
                pacman_init_service_file.write("ExecStart=/usr/bin/pacman-key --populate idmslinux\n")
                pacman_init_service_file.write("\n")
                pacman_init_service_file.write("[Install]\n")
                pacman_init_service_file.write("WantedBy=multi-user.target\n")
        except OSError as err:
            raise IliRuntimeError(
                f"Failed to write pacman init service '%s': {err}",
                pacman_init_service_path,
            ) from None

        logging.debug("  - Enabling pacman init service")

        # Add service to enable list
        rootfs_systemd_system_wants_path_pacmaninit = rootfs_systemd_system_wants_path.concatpath("pacman-init.service")
        try:
            rootfs_systemd_system_wants_path_pacmaninit.symlink_to("../pacman-init.service")
        except OSError as err:
            raise IliRuntimeError(
                "Failed to symlink pacman-init.service to systemd multi-user.target.wants directory "
                f"'{rootfs_systemd_system_wants_path}': {err}"
            ) from None

    def setup_target_hostname(self) -> None:
        """Set up target hostname."""

        if self.target_hostname is None:
            raise IliRuntimeError("target_hostname must be set")
        if self.rootfs is None:
            raise IliRuntimeError("rootfs must be set")
        if self.rootfs.mount_point is None:
            raise IliRuntimeError("rootfs mount_point must be set")

        logging.debug("  - Setting hostname: %s", self.target_hostname)

        # Write out hostname
        hostname_path = self.rootfs.mount_point.concatpath("etc", "hostname")
        try:
            with hostname_path.open(mode="w", encoding="UTF-8") as hostname_file:
                hostname_file.write(f"{self.target_hostname}\n")
        except OSError as err:
            raise IliRuntimeError(f"Unable to open file '{hostname_path}' for writing: {err}") from None

        # If we have multiple hostname components, add the first component to the hostnames list
        hostnames = self.target_hostname
        hostname_components = self.target_hostname.split(".", maxsplit=1)
        if len(hostname_components) > 1:
            hostnames += f" {hostname_components[0]}"

        logging.debug("  - Setting up /etc/hosts")

        # Write out hosts file
        hosts_path = self.rootfs.mount_point.concatpath("etc", "hosts")
        try:
            with hosts_path.open(mode="w", encoding="UTF-8") as hosts_file:
                hosts_file.write(f"127.0.1.1\t{hostnames}\n")
                hosts_file.write(f"::1\t{hostnames}\n")
        except OSError as err:
            raise IliRuntimeError(f"Unable to open file '{hosts_path}' for writing: {err}") from None

    def setup_target_users(self) -> None:  # noqa: C901,PLR0912 # pylint: disable=too-many-branches
        """Set up target system users."""

        if self.rootfs is None:
            raise IliRuntimeError("rootfs must be set")
        if self.rootfs.mount_point is None:
            raise IliRuntimeError("rootfs mount_point must be set")

        logging.debug("  - Configuring root user")

        # Check if we have the root password set
        if self.target_root_password is not None:
            # If blank we're locking the account
            if self.target_root_password == "":  # nosec
                self.run_target(["/usr/bin/passwd", "-l", "root"])
            # Or no password...
            elif self.target_root_password == " ":  # noqa: S105 # nosec
                self.run_target(["/usr/bin/passwd", "-d", "root"])
            # Or setting it...
            else:
                self.run_target(["/usr/bin/chpasswd"], stdin=f"root:{self.target_root_password}\n")

        # Check if we're creating a sudo user
        if self.target_sudo_username:
            logging.debug("  - Configuring sudo user: %s", self.target_sudo_username)

            self.run_target(["/usr/bin/useradd", "--create-home", self.target_sudo_username])

            # Check if the sudo password was specified
            if self.target_sudo_password is not None:
                # Check if we're locking the account
                if self.target_sudo_password == "":  # nosec
                    self.run_target(["/usr/bin/passwd", "-l", self.target_sudo_username])
                # Disabling the password
                elif self.target_sudo_password == " ":  # noqa: S105 # nosec
                    self.run_target(["/usr/bin/passwd", "-d", self.target_sudo_username])
                # Or setting the passord...
                else:
                    self.run_target(
                        ["/usr/bin/chpasswd"],
                        stdin=f"{self.target_sudo_username}:{self.target_sudo_password}\n",
                    )

            # Write out the sudoers file
            sudoers_path = self.rootfs.mount_point.concatpath("etc", "sudoers.d", self.target_sudo_username)
            umask_orig = os.umask(0o027)
            try:
                with sudoers_path.open(mode="w", encoding="UTF-8") as sudoers_file:
                    sudoers_file.write(f"{self.target_sudo_username} ALL=(ALL) ALL\n")
                    # NK: add in option to require no password when using sudo
                    # sudoers_file.write(f"{self.target_sudo_username} ALL=(ALL) NOPASSWD: ALL\n")  # noqa: ERA001
            except OSError as err:
                raise IliRuntimeError(f"Unable to open file '{sudoers_path}' for writing: {err}") from None
            finally:
                os.umask(umask_orig)

        # If we installed grml-zsh-config, change the shell of root to zsh
        if "grml-zsh-config" in self.target_packages:
            logging.debug("  - Setting shell for root user")
            self.run_target(["/usr/bin/usermod", "-s", "/bin/zsh", "root"])
            if self.target_sudo_username:
                logging.debug("  - Setting shell for sudo user")
                self.run_target(["/usr/bin/usermod", "-s", "/bin/zsh", self.target_sudo_username])

    def setup_target_overlay_directories(self) -> None:  # noqa: C901,PLR0912
        """Overlay files from the host system into the target."""

        if self.rootfs is None:
            raise IliRuntimeError("rootfs must be set")
        if self.rootfs.mount_point is None:
            raise IliRuntimeError("rootfs mount_point must be set")

        # Loop with overlay directories and copy into the target
        for src, dst in self.overlay_directories:
            # Grab rootfs destination
            rootfs_dst = self.rootfs.mount_point.concatpath(dst)
            logging.debug("  - Copying overlay '%s' to '%s'", src, dst)
            # Check if we're copying a file or a directory
            if src.is_file():
                try:
                    # If the dir exists, then we're copying into it
                    if str(rootfs_dst).endswith("/"):
                        # Make directory if it does not exist
                        if not rootfs_dst.exists():
                            logging.debug("    - Creating directory: %s", rootfs_dst)
                            rootfs_dst.mkdir(parents=True, exist_ok=True)
                        # Work out the new destination file name
                        dst_file = rootfs_dst.concatpath(src.name)
                    # If the destination doesn't exist, assume its a full path
                    else:
                        # Check parent exists, if not create it
                        if not rootfs_dst.parent.exists():
                            logging.debug("    - Creating directory: %s", rootfs_dst)
                            rootfs_dst.parent.mkdir(parents=True, exist_ok=True)
                        # Dest file in this case is the rootfs_dst
                        dst_file = rootfs_dst
                    # Copy file
                    logging.debug("    - Copying file: %s to %s", src, dst_file)
                    shutil.copy(src, dst_file)
                except OSError as err:
                    raise IliRuntimeError(f"Failed to copy overlay '{src}' to '{rootfs_dst}': {err}") from None
            else:
                try:
                    for src_dir, _, files in os.walk(src):
                        dst_dir = rootfs_dst.concatpath(IliPath(src_dir).relative_to(src))
                        if not dst_dir.exists():
                            logging.debug("    - Creating directory: %s", dst_dir)
                            dst_dir.mkdir(parents=True, exist_ok=True)
                        for file in files:
                            src_file = IliPath(src_dir).concatpath(file)
                            dst_file = dst_dir.concatpath(file)
                            logging.debug("    - Copying file: %s to %s", src_file, dst_file)
                            shutil.copy(src_file, dst_file)
                except (OSError, shutil.Error) as err:
                    raise IliRuntimeError(f"Failed to copy overlay '{src}' to '{rootfs_dst}' [{dst}]: {err}") from None

    def setup_target_root_autologin(self) -> None:
        """
        Set up target root autologin.

        This is overridden in containers.
        """

        if self.rootfs is None:
            raise IliRuntimeError("rootfs must be set")
        if self.rootfs.mount_point is None:
            raise IliRuntimeError("rootfs mount_point must be set")

        # Short circuit if we're not setting up autologin
        if not self.target_root_autologin:
            logging.info("  - Root autologin not enabled")
            return

        logging.debug("  - Enabling root autologin")

        # Create service override for tty1
        getty_override_dir = self.rootfs.mount_point.concatpath("etc", "systemd", "system", "getty@tty1.service.d")
        try:
            getty_override_dir.mkdir(mode=0o0755)
        except OSError as err:
            raise IliRuntimeError(f"Failed to create getty override dir '{getty_override_dir}': {err}") from None

        getty_override_conf_path = getty_override_dir.concatpath("autologin.conf")
        try:
            with getty_override_conf_path.open(mode="w", encoding="UTF-8") as getty_override_conf_file:
                getty_override_conf_file.write("[Service]\n")
                getty_override_conf_file.write("ExecStart=\n")
                getty_override_conf_file.write(r"ExecStart=-/sbin/agetty --autologin root --noclear %I $TERM\n")
        except OSError as err:
            raise IliRuntimeError(f"Failed to open service override '{getty_override_conf_path}' for writing: {err}") from None

        # Create service override for ttyS0
        getty_serial_override_dir = self.rootfs.mount_point.concatpath("etc", "systemd", "system", "serial-getty@ttyS0.service.d")
        try:
            getty_serial_override_dir.mkdir(mode=0o0755)
        except OSError as err:
            raise IliRuntimeError(
                f"Failed to create getty serial override directory '{getty_serial_override_dir}': {err}"
            ) from None

        getty_serial_override_conf_path = getty_serial_override_dir.concatpath("autologin.conf")
        with getty_serial_override_conf_path.open(mode="w", encoding="UTF-8") as getty_serial_override_conf_file:
            getty_serial_override_conf_file.write("[Service]\n")
            getty_serial_override_conf_file.write("ExecStart=\n")
            getty_serial_override_conf_file.write(
                r"ExecStart=-/sbin/agetty --autologin root --keep-baud 115200,38400,9600 %I $TERM\n"
            )

    def setup_target_networking(self) -> None:
        """Set up networking."""

        # Short circuit if we're not setting up networking
        if not self.target_networking:
            logging.info("  - Networking not enabled")
            return

        for device_name in self.target_network_devices:
            logging.debug("  - Setting up networking: %s", device_name)
            # Carry on to the next device if we're not touching the configuration for this one
            if self.target_network_config[f"network_device_{device_name}"] == "NONE":
                continue
            # Setup network configuration
            self._setup_target_networking_link_file(device_name)
            self._setup_target_networking_network_file(device_name)

    def setup_target_systemd_services(self) -> None:
        """Set up target systemd services."""

        # Enable services
        if self.target_services:
            for service in self.target_services:
                logging.debug("  - Enabling service: %s", service)
                self.run_target(["/usr/bin/systemctl", "enable", service])

    def sanitize_target(self) -> None:
        """Sanitize target."""

        if self.rootfs is None:
            raise IliRuntimeError("rootfs must be set")
        if self.rootfs.mount_point is None:
            raise IliRuntimeError("rootfs mount_point must be set")

        # Create list of paths to remove
        path_list: list[IliPath] = []
        for fglob in self._rootfs_remove_globs:
            # Work out the path with the rootfs mount_point prepended
            rootfs_glob = f"{self.rootfs.mount_point}{fglob}"
            # Split components
            rootfs_glob_components: list[str] = rootfs_glob.split("/")
            rootfs_glob_path = IliPath("/").concatpath(*rootfs_glob_components[:-1])
            rootfs_glob_glob = rootfs_glob_components[-1]
            # Append list of glob results
            path_list.extend(rootfs_glob_path.glob(rootfs_glob_glob))

        # Loop and remove
        for path in sorted(path_list):
            try:
                if path.is_dir():
                    logging.debug("  - Removing directory: %s", path)
                    shutil.rmtree(path)
                else:
                    logging.debug("  - Removing file: %s", path)
                    path.unlink()
            except OSError as err:
                raise IliRuntimeError(f"Failed to remove '{path}': {err}") from None

    def wait_block_settle(self) -> None:  # pylint: disable=useless-option-value
        """Wait for block devices to settle."""
        logging.debug("  - Waiting for block devices to settle")
        util.run_host(["/usr/bin/udevadm", "settle"])

    def wipe_fs(self, device: IliPath) -> None:  # pylint: disable=useless-option-value
        """Wipe block device filesystems."""
        logging.debug("  - Wiping block device: %s", device)
        util.run_host(["/usr/bin/wipefs", "--all", str(device)])

    def cleanup_temp_paths(self) -> None:
        """
        Cleanup temporary paths creating during install.

        This happens after the finalize_target. But also happens during cleanup.
        """

        # Just return if we have none
        if not self.temp_paths:
            logging.info("  - No temp paths to cleanup")
            return

        logging.debug("  - Cleaning up temp paths")

        # Loop while we are removing temp paths
        while True:
            # If we don't have temp paths left, abort while loop
            if not self.temp_paths:
                break
            # Grab last temp path in the list
            path = self.temp_paths.pop()
            if not path.exists():
                logging.error("  - Temp path does not exist: %s", path)
                continue
            if path.is_dir():
                logging.debug("  - Cleaning up temp paths - removing directory: %s", path)
                shutil.rmtree(path)
            else:
                logging.debug("  - Cleaning up temp paths - removing file: %s", path)
                path.unlink()

    def cleanup_mounts(self) -> None:
        """
        Unmount target.

        The purpose of this method is to unmount the rootfs mounts.

        This happens before finalize_target. This also happens during cleanup.
        """

        # Just return if we don't have any
        if not self.mounts:
            logging.info("  - No mounts to cleanup")
            return

        logging.debug("  - Cleaning up mounts")

        # Loop while we are unmounting
        while True:
            # If we don't have mounts left, abort while loop
            if not self.mounts:
                break
            # Grab last mount in the list and unmount
            filesystem = self.mounts[-1]
            logging.debug("  - Cleaning up mounts - unmounting: %s", filesystem.mount_point)
            self.unmount(filesystem)

    def cleanup_lvm_logical_volumes(self) -> None:
        """
        Deactivate LVM logical volume devices.

        This happens before finalize_target. This also happens during cleanup.
        """

        # Just return if we have none
        if not self.lvm_lvs:
            logging.info("  - No LVM logical volumes to cleanup")
            return

        logging.debug("  - Deactivating LVM logical volumes")

        # Loop while we are removing devices
        while True:
            # If we don't have LVM logical devices left, abort while loop
            if not self.lvm_lvs:
                break
            # Grab last LVM logical volume
            lvm_lv = self.lvm_lvs.pop()
            logging.debug("  - Deactivating LVM logical volumes: %s", lvm_lv)
            util.run_host(["/usr/bin/lvchange", "--activate", "n", lvm_lv])

    def cleanup_lvm_volume_groups(self) -> None:
        """
        Deactivate LVM volume group devices.

        This happens before finalize_target. This also happens during cleanup.
        """

        # Just return if we don't have any
        if not self.lvm_vgs:
            logging.info("  - No LVM volume groups to cleanup")
            return

        logging.debug("  - Deactivating LVM volume groups")

        # Loop while we are removing devices
        while True:
            # If we don't have LVM volume groups left, abort while loop
            if not self.lvm_vgs:
                break
            # Grab last LVM volume group
            lvm_vg = self.lvm_vgs.pop()
            logging.debug("  - Deactivating LVM volume groups: %s", lvm_vg)
            util.run_host(["/usr/bin/vgchange", "--activate", "n", lvm_vg])

    def cleanup_luks_volumes(self) -> None:
        """
        Close encrypted LUKS devices.

        This happens before finalize_target. This also happens during cleanup.
        """

        # Just return if we don't have any
        if not self.luks_volumes:
            logging.info("  - No LUKS devices to cleanup")
            return

        logging.debug("  - Closing LUKS devices")

        # Loop while we are removing devices
        while True:
            # If we don't have LUKS volumes left, abort while loop
            if not self.luks_volumes:
                break
            # Grab last LUKS device
            luks_volume = self.luks_volumes.pop()
            logging.debug("  - Closing LUKS volume: %s", luks_volume)
            util.run_host(["/usr/bin/cryptsetup", "close", "--batch-mode", str(luks_volume)])

    def cleanup_kpartx(self) -> None:
        """
        Remove kpartx devices.

        The purpose of this method is to remove kpartx devices.

        This happens before finalize_target. This also happens during cleanup.
        """

        # Just return if we don't have any to cleanup
        if not self.kpartx_devices:
            logging.info("  - No kpartx devices to cleanup")
            return

        logging.debug("  - Cleaning up kpartx devices")

        # Loop while we are removing devices
        while True:
            # If we don't have kpartx devices left, abort while loop
            if not self.kpartx_devices:
                break
            # Grab last kpartx entry in list
            (kpartx_device, kpartx_maps) = self.kpartx_devices.pop()

            logging.debug("  - Cleaning up kpartx devices: %s => %s", kpartx_device, kpartx_maps)

            tries = 0
            while tries < 10:  # noqa: PLR2004
                try:
                    util.run_host(["/usr/bin/kpartx", "-d", kpartx_device])
                    break
                except IliRuntimeError as exc:
                    logging.error("  - Exception during cleaning up kpartx devices: %s", exc)  # noqa: TRY400
                    if exc.data:
                        logging.error("  - Exception data: %s", exc.data)  # noqa: TRY400
                # Bump tries and wait a bit
                tries += 1
                time.sleep(0.5)

    def cleanup_loop_devices(self) -> None:
        """
        Remove loop devices.

        The purpose of this method is to remove loop devices.

        This happens before finalize_target. This also happens during cleanup.
        """

        # Just return if we don't have any loop devices
        if not self.loop_devices:
            logging.info("  - No loop devices to cleanup")
            return

        logging.debug("  - Cleaning up loop devices")

        # Loop while we are removing devices
        while True:
            # If we don't have loop devices left, abort while loop
            if not self.loop_devices:
                break
            # Grab last loop device in the list
            loop_device = self.loop_devices[-1]
            logging.debug("  - Cleaning up loop devices: %s", loop_device)
            self.remove_loop_device(loop_device)

    def cleanup_mdraid_devices(self) -> None:
        """
        Stop MDRAID devices.

        This happens before finalize_target. This also happens during cleanup.
        """

        # Just return if we don't have any mdraid devices active
        if not self.mdraid_devices:
            logging.info("  - No MDRAID devices to cleanup")
            return

        logging.debug("  - Stopping MDRAID devices")

        # Loop while we are removing devices
        while True:
            # If we don't have MDRAID devices left, abort while loop
            if not self.mdraid_devices:
                break
            # Grab last MDRAID device
            mdraid_device = self.mdraid_devices.pop()
            logging.debug("  - Stopping MDRAID devices: %s", mdraid_device)
            util.run_host(["/usr/bin/mdadm", "--stop", str(mdraid_device)])

    def finalize_target(self) -> None:
        """
        Finalize target.

        The purpose of this method is to do things like tar the rootfs.
        """

    def run_target(self, args: list[str], **kwargs: Any) -> list[str]:  # noqa: ANN401
        """Run a command inside the target environment."""

        # Check we have a rootfs first
        if not self.rootfs:
            raise IliRuntimeError("No rootfs when calling run_target")

        # Run within rootfs mount point
        new_args = ["arch-chroot", str(self.rootfs.mount_point), *args]
        return util.run_host(new_args, **kwargs)

    def get_config_from_user(self, config_item: dict[str, Any]) -> str:  # noqa: C901,PLR0912 # pylint: disable=too-many-branches
        """
        Get a single item, from config or user.

        Valid keys for `config_item` are...
          - 'name': Name of item stored in the configuration
          - 'property': Name of the property to set if we get a value
          - 'type': Type of the config item, which is one of the following 'text', 'username', 'hostname', 'password',
          -         'checkboxlist', 'radiolist'
          - 'title': Dialog box title
          - 'prompt': Prompt inside the dialog box
          - 'required': Is this item required to have a value
          - 'values': Used for the 'checkboxlist' and 'radiolist' types
          - 'dynamic_values': This is a callable used to produce dynamic values for 'checkboxlist' and 'radiolist' types
          - 'dynamic_config_items': This is callable to return additional config items to process, it should accept two arguments.
                                    The first argument would be the value chosen for this config_item, the second is a 'data'
                                    argument which we can use to pass state down the chain.
          - 'multivalue': Does this item support multiple values?

        :returns: str
        """
        # Sanity check
        for key in ("name", "type", "title", "prompt"):
            if key not in config_item:
                raise IliRuntimeError(f'Internal error - config_item must have a "{key}" key')

        # Text is default
        if config_item["type"] == "text":
            value = self._get_user_input_text(config_item)
        elif config_item["type"] == "password":
            value = self._get_user_input_password(config_item)
        elif config_item["type"] == "hostname":
            value = self._get_user_input_hostname(config_item)
        elif config_item["type"] == "username":
            value = self._get_user_input_username(config_item)
        elif config_item["type"] == "checkboxlist":
            value = self._get_user_input_checkboxlist(config_item)
        elif config_item["type"] == "radiolist":
            value = self._get_user_input_radiolist(config_item)
        elif config_item["type"] == "ipv4address":
            value = self._get_user_input_ipv4address(config_item)
        elif config_item["type"] == "ipv4address.comma-separated":
            value = self._get_user_input_ipv4address_multiple(config_item)
        elif config_item["type"] == "ipv6address":
            value = self._get_user_input_ipv6address(config_item)
        elif config_item["type"] == "ipv6address.comma-separated":
            value = self._get_user_input_ipv6address_multiple(config_item)
        elif config_item["type"] == "ipv4cidr":
            value = self._get_user_input_ipv4cidr(config_item)
        elif config_item["type"] == "ipv6cidr":
            value = self._get_user_input_ipv6cidr(config_item)
        else:
            raise IliRuntimeError(f"Internal error - config_item type not undertood {config_item['type']}")

        return value

    def get_config_from_args_profile(self, config_item: dict[str, Any]) -> Any:  # noqa: ANN401
        """
        Get a single configuration item.

        The mandatory keys for `config_item` for this method are:
            - `name`: The name of the configuration item to retrieve

        :param config_item: Configuration item definition
        :returns: Any
        """

        # Command line options are checked first
        if config_item["name"] in self.args:
            value = getattr(self.args, config_item["name"])
            # Make sure it is not None
            if value is not None:
                # If this is a list
                if isinstance(value, list):  # noqa: SIM102
                    # Check if it is not a multivalue item, if not, return the first item instead
                    if ("multivalue" not in config_item) or not config_item["multivalue"]:
                        value = value[0]
                return value

        # Check if the config item was specified in the profile
        if self.profile:  # noqa: SIM102
            if config_item["name"] in self.profile:
                return self.profile[config_item["name"]]

        return None

    def resolve_data_file(self, filename: str) -> pathlib.Path:
        """Resolve the path of one of our data files."""
        search_paths = util.search_paths(self.args.data_search_paths)

        # Loop with each path
        for path in search_paths:
            # Test if the file is within this search
            test_path = path.joinpath(filename)
            if test_path.exists():
                return test_path

        raise IliRuntimeError(f"Cannot find data file '{filename}' in any search path: {search_paths}")

    def _get_user_input_text(  # pylint: disable=useless-option-value
        self,
        config_item: dict[str, Any],
        title_extra: str | None = None,
        **kwargs: Any,  # noqa: ANN401
    ) -> str:
        """Get a text configuration item."""

        title = config_item["title"] + title_extra if title_extra else config_item["title"]

        # Loop until we have the input we need
        while True:
            value = input_dialog(title=title, text=config_item["prompt"], **kwargs).run()
            # This means that we got a cancel button press
            if value is None:
                raise IliRuntimeError("User aborted via dialog")
            # If we have value, we can exit the loop
            if value:
                break
            # If this is not a required item, we can also exit the loop
            if ("required" not in config_item) or not config_item["required"]:
                break
            # Display a message that the input is required
            message_dialog(title=config_item["title"], text="This input is required!").run()

        return value

    def _get_user_input_checkboxlist(  # noqa: RUF100 # pylint: disable=useless-option-value
        self,
        config_item: dict[str, Any],
        title_extra: str | None = None,
        **kwargs: Any,  # noqa: ANN401
    ) -> Any:  # noqa: ANN401
        """Get a checkboxlist configuration item."""

        title = config_item["title"] + title_extra if title_extra else config_item["title"]

        # Check for dynamic values
        if "dynamic_values" in config_item:
            values = config_item["dynamic_values"]()
            # If no values were provided, we're skipping prompting the user
            if not values:
                return None
        else:
            values = config_item["values"]

        # Make sure we have values for the checkboxlist
        if not values:
            raise IliRuntimeError(f"No values to display for: {title}")

        # Loop until we have the input we need
        while True:
            value: Any = checkboxlist_dialog(title=title, text=config_item["prompt"], values=values, **kwargs).run()
            # This means that we got a cancel button press
            if value is None:
                raise IliRuntimeError("User aborted via dialog")
            # If we have value, we can exit the loop
            if value:
                break
            # If this is not a required item, we can also exit the loop
            if ("required" not in config_item) or not config_item["required"]:
                break
            # Display a message that the input is required
            message_dialog(title=config_item["title"], text="This input is required!").run()

        return value

    def _get_user_input_radiolist(  # pylint: disable=useless-option-value
        self,
        config_item: dict[str, Any],
        title_extra: str | None = None,
        **kwargs: Any,  # noqa: ANN401
    ) -> Any:  # noqa: ANN401
        """Get a radiolist configuration item."""

        title = config_item["title"] + title_extra if title_extra else config_item["title"]

        # Check for dynamic values
        if "dynamic_values" in config_item:
            values = config_item["dynamic_values"]()
            # If not values were provided, we're skipping prompting the user
            if not values:
                return None
        else:
            values = config_item["values"]

        # Make sure we have values for the checkboxlist
        if not values:
            raise IliRuntimeError(f"No values to display for: {title}")

        # Loop until we have the input we need
        while True:
            value: Any = radiolist_dialog(title=title, text=config_item["prompt"], values=values, **kwargs).run()
            # This means that we got a cancel button press
            if value is None:
                raise IliRuntimeError("User aborted via dialog")
            # If we have value, we can exit the loop
            if value:
                break
            # If this is not a required item, we can also exit the loop
            if ("required" not in config_item) or not config_item["required"]:
                break
            # Display a message that the input is required
            message_dialog(title=config_item["title"], text="This input is required!").run()

        return value

    def _get_user_input_password(self, config_item: dict[str, Any]) -> str:
        """Get password from user."""

        # We need 2 passwords to compare
        password2: str | None = None
        while True:
            # Check if we're getting input for the 2nd or first password
            if password2:
                password = self._get_user_input_text(config_item, password=True, title_extra=" AGAIN")
            else:
                password = self._get_user_input_text(config_item, password=True)
            # Check if the password is allowed to be blank
            if (password == "") and (("required" not in config_item) or not config_item["required"]):  # nosec
                break
            # If we have a password2, check it matches
            if password2:
                # If not, we need to display a dialog
                if password != password2:
                    message_dialog(title=config_item["title"], text="Passwords do not match").run()
                    password2 = None
                    continue
                break
            # Set password2 to password so we can match it
            password2 = password

        return password

    def _get_user_input_hostname(self, config_item: dict[str, Any]) -> str:
        """Get hostname from user."""

        # Loop until we get something valid
        while True:
            value = self._get_user_input_text(config_item)
            # Check if the hostname can be blank
            if (value == "") and (("required" not in config_item) or not config_item["required"]):
                break
            # Check if its valid
            if util.is_valid_hostname(value):
                break
            message_dialog(title=config_item["title"], text="Field contains an invalid hostname").run()

        return value

    def _get_user_input_username(self, config_item: dict[str, Any]) -> str:
        """Get username from user."""

        # Loop until we get something valid
        while True:
            value = self._get_user_input_text(config_item)
            # Check if the username can be blank
            if (value == "") and (("required" not in config_item) or not config_item["required"]):
                break
            # Check if the username is valid
            if util.is_valid_username(value):
                break
            message_dialog(title=config_item["title"], text="Field contains an invalid username").run()

        return value

    def _get_user_input_ipv4address(self, config_item: dict[str, Any]) -> str:
        """Get IPv4 address from user."""

        # Loop until we get something valid
        while True:
            value = self._get_user_input_text(config_item)
            # Check if the ipv4address can be blank
            if (value == "") and (("required" not in config_item) or not config_item["required"]):
                break
            # Check if its valid
            if util.is_valid_ipv4address(value):
                break
            message_dialog(
                title=config_item["title"],
                text="Field contains an invalid IPv4 address",
            ).run()

        return value

    def _get_user_input_ipv4address_multiple(self, config_item: dict[str, Any]) -> str:
        """Get multiple IPv4 addresses from user, comma separated."""

        # Loop until we get something valid
        while True:
            value = self._get_user_input_text(config_item)
            # Check if the ipv4address can be blank
            if (value == "") and (("required" not in config_item) or not config_item["required"]):
                break
            # Check if its valid
            all_valid = True
            for ipv4address in value.split(","):
                if not util.is_valid_ipv4address(ipv4address):
                    all_valid = False
            if all_valid:
                break
            message_dialog(
                title=config_item["title"],
                text="Field contains an invalid IPv4 address",
            ).run()

        return value

    def _get_user_input_ipv6address(self, config_item: dict[str, Any]) -> str:
        """Get IPv6 address from user."""

        # Loop until we get something valid
        while True:
            value = self._get_user_input_text(config_item)
            # Check if the ipv4address can be blank
            if (value == "") and (("required" not in config_item) or not config_item["required"]):
                break
            # Check if its valid
            if util.is_valid_ipv6address(value):
                break
            message_dialog(
                title=config_item["title"],
                text="Field contains an invalid IPv6 address",
            ).run()

        return value

    def _get_user_input_ipv6address_multiple(self, config_item: dict[str, Any]) -> str:
        """Get multiple IPv6 addresses from user, comma separated."""

        # Loop until we get something valid
        while True:
            value = self._get_user_input_text(config_item)
            # Check if the ipv6address can be blank
            if (value == "") and (("required" not in config_item) or not config_item["required"]):
                break
            # Check if its valid
            all_valid = True
            for ipv6address in value.split(","):
                if not util.is_valid_ipv6address(ipv6address):
                    all_valid = False
            if all_valid:
                break
            message_dialog(
                title=config_item["title"],
                text="Field contains an invalid IPv6 address",
            ).run()

        return value

    def _get_user_input_ipv4cidr(self, config_item: dict[str, Any]) -> str:
        """Get IPv4 CIDR from user."""

        # Loop until we get something valid
        while True:
            value = self._get_user_input_text(config_item)
            # Check if the ipv4cidr can be blank
            if (value == "") and (("required" not in config_item) or not config_item["required"]):
                break
            # Check if its valid
            if util.is_valid_ipv4cidr(value):
                break
            message_dialog(title=config_item["title"], text="Field contains an invalid IPv4 CIDR").run()

        return value

    def _get_user_input_ipv6cidr(self, config_item: dict[str, Any]) -> str:
        """Get IPv6 CIDR from user."""

        # Loop until we get something valid
        while True:
            value = self._get_user_input_text(config_item)
            # Check if the ipv4cidr can be blank
            if (value == "") and (("required" not in config_item) or not config_item["required"]):
                break
            # Check if its valid
            if util.is_valid_ipv6cidr(value):
                break
            message_dialog(title=config_item["title"], text="Field contains an invalid IPv6 CIDR").run()

        return value

    def _get_sudo_config_items(self, sudo_username: str) -> list[dict[str, Any]]:  # pylint: disable=useless-option-value
        """Return sudo user config items we need if the sudo username was specified."""

        # If we didn't get a sudo username, just return an empty list
        if not sudo_username:
            return []

        sudo_config_items = []

        sudo_config_items.append(
            {
                "name": "sudo_password",
                "type": "password",
                "property": "target_sudo_password",
                "title": "Sudo user password",
                "prompt": "Sudo user password",
                "required": True,
            },
        )
        return sudo_config_items

    def _get_network_device_config_items(self, value: str) -> list[dict[str, Any]]:
        """Return the network device addressing types config item for each device."""

        # Make pylint happy
        if value is None:
            raise IliRuntimeError("value should not be None")

        # Loop through the network devices and add their addressing config options to our config items list
        device_config_items = []
        for device_name_raw in natsorted(self.target_network_devices):
            device_name = str(device_name_raw)
            device_config_items.append(
                {
                    "name": f"network_device_{device_name}",
                    "type": "radiolist",
                    "property": "target_network_config",
                    "required": True,
                    "title": f"Network Device Addressing: {device_name}",
                    "prompt": f"Select network device addressing for {device_name}",
                    "values": [("NONE", "Do not configure"), ("DHCP", "Configure DHCP"), ("STATIC", "Static IP addressing")],
                    "dynamic_config_items": self._get_network_device_address_config_items,
                    "dynamic_config_items_data": {"device": device_name},
                }
            )

        return device_config_items

    def _get_network_device_address_config_items(self, addressing_type: str, data: dict[Any, Any]) -> list[dict[str, Any]]:
        """Return configuration for a network interface depending on its addressing type."""

        address_config_items = []

        device_name = data["device"]

        # If this is static networking we need to ask for the gateway
        if addressing_type == "STATIC":
            address_config_items.append(
                {
                    "name": f"network_device_{device_name}_ipv4address",
                    "type": "ipv4cidr",
                    "property": "target_network_config",
                    "title": f"Network IPv4 Address: {device_name}",
                    "prompt": f"Enter IPv4 address for {device_name} (blank = none)",
                    "dynamic_config_items": self._get_network_device_ip4_config_items,
                    "dynamic_config_items_data": data,
                }
            )
            address_config_items.append(
                {
                    "name": f"network_device_{device_name}_ipv6address",
                    "type": "ipv6cidr",
                    "property": "target_network_config",
                    "title": f"Network IPv6 Address: {device_name}",
                    "prompt": f"Enter IPv6 address for {device_name} (blank = none)",
                    "dynamic_config_items": self._get_network_device_ip6_config_items,
                    "dynamic_config_items_data": data,
                }
            )

        return address_config_items

    def _get_network_device_ip4_config_items(  # pylint: disable=useless-option-value
        self, ipaddress: str, data: dict[Any, Any]
    ) -> list[dict[str, Any]]:
        """Return IPv4 config items if we have an address set."""

        # If we don't have any IP, just return no additional options
        if not ipaddress:
            return []

        # Use device variable for convenience below
        device_name = data["device"]

        # Additional IP config items we need
        ip_config_items = []

        # Add IPv4 gateway
        ip_config_items.append(
            {
                "name": f"network_device_{device_name}_ipv4gateway",
                "type": "ipv4address",
                "property": "target_network_config",
                "title": f"Network IPv4 Gateway: {device_name}",
                "prompt": f"Enter IPv4 gateway for {device_name} (blank = none)",
            }
        )
        # Add IPv4 DNS
        ip_config_items.append(
            {
                "name": f"network_device_{device_name}_ipv4dns",
                "type": "ipv4address.comma-separated",
                "property": "target_network_config",
                "title": f"Network IPv4 DNS Server(s): {device_name}",
                "prompt": f"Enter IPv4 DNS server(s) for {device_name} (comma separated, blank = none)",
            }
        )

        return ip_config_items

    def _get_network_device_ip6_config_items(  # pylint: disable=useless-option-value
        self, ipaddress: str, data: dict[Any, Any]
    ) -> list[dict[str, Any]]:
        """Return configuration for a network interface depending on its addressing type."""

        # If we don't have any IP, just return no additional options
        if not ipaddress:
            return []

        # Use device variable below for convenience
        device_name = data["device"]

        # Additional IP configuration items required
        ip_config_items = []

        # IPv6 gateway
        ip_config_items.append(
            {
                "name": f"network_device_{device_name}_ipv6gateway",
                "type": "ipv6address",
                "property": "target_network_config",
                "title": f"Network IPv6 Gateway: {device_name}",
                "prompt": f"Enter IPv6 gateway for {device_name} (blank = none)",
            }
        )

        # IPv6 DNS
        ip_config_items.append(
            {
                "name": f"network_device_{device_name}_ipv6dns",
                "type": "ipv6address.comma-separated",
                "property": "target_network_config",
                "title": f"Network IPv6 DNS Server(s): {device_name}",
                "prompt": f"Enter IPv6 DNS server(s) for {device_name} (comma separated, blank = none)",
            }
        )

        return ip_config_items

    def _setup_target_networking_link_file(self, device_name: str) -> None:
        """Write out systemd networkd link configuration file."""

        # If this is a wildcard device name, just return, we don't need to setup a link file
        if device_name.endswith("*"):
            return

        # Make pylint happy :), even though rootfs is ALWAYS set
        if not self.rootfs:
            raise RuntimeError("rootfs not set")

        # Write out systemd networkd link file
        link_file_path = self.rootfs.mount_point.concatpath("etc", "systemd", "network", f"25-{device_name}.link")
        try:
            with link_file_path.open(mode="w", encoding="UTF-8") as link_file:
                # Setup match
                link_file.write("# Created during install\n")
                link_file.write("[Match]\n")
                # If we have a MAC we're going to be renaming the device
                mac_config_attr = f"network_device_{device_name}_mac"
                if mac_config_attr in self.target_network_config:
                    link_file.write(f"MACAddress={self.target_network_config[mac_config_attr]}\n")
                    link_file.write("Path=pci-*\n")
                # If we don't have a MAC we'll be using the original name
                else:
                    link_file.write(f"OriginalName={device_name}\n")
                link_file.write("\n")

                # Setup link properties
                link_file.write("[Link]\n")
                # If we have a MAC it means we're renaming the device
                if mac_config_attr in self.target_network_config:
                    link_file.write(f"Name={device_name}\n")
                # Check if we can add an alias
                alias_config_attr = f"network_device_{device_name}_alias"
                if alias_config_attr in self.target_network_config:
                    link_file.write(f"Alias={self.target_network_config[alias_config_attr]}\n")
                # Check if we can add a description
                description_config_attr = f"network_device_{device_name}_description"
                if description_config_attr in self.target_network_config:
                    link_file.write(f"Description={self.target_network_config[description_config_attr]}\n")

                link_file.close()
        except OSError as err:
            raise IliRuntimeError(f"Failed to open '{link_file_path}' for reading: {err}") from None

    def _setup_target_networking_network_file(self, device_name: str) -> None:  # noqa: C901,PLR0912,PLR0915
        """Write out systemd networkd network configuration file."""

        # If this is a wildcard device name, we need to change a few things
        if device_name.endswith("*"):
            file_priority = "99"
            file_devname = "default"
        else:
            file_priority = "50"
            file_devname = device_name

        if not self.rootfs:
            raise RuntimeError("rootfs not set")

        # Open the hardware address file
        network_file_path = self.rootfs.mount_point.concatpath(
            "etc", "systemd", "network", f"{file_priority}-{file_devname}.network"
        )
        try:
            with network_file_path.open(mode="w", encoding="UTF-8") as network_file:
                network_file.write("# Created during install\n")
                network_file.write("[Match]\n")
                network_file.write(f"Name={device_name}\n")
                network_file.write("\n")

                network_file.write("[Network]\n")
                # Check if we can add a description
                description_config_attr = f"network_device_{device_name}_description"
                if description_config_attr in self.target_network_config:
                    network_file.write(f"Description={self.target_network_config[description_config_attr]}\n")
                # Check if this interface uses DHCP
                if self.target_network_config[f"network_device_{device_name}"] == "DHCP":
                    network_file.write("# The below is required to enable IPv6 address assignment with IPForward=yes\n")
                    network_file.write("IPv6AcceptRA=yes\n")
                    network_file.write("\n")
                    network_file.write("DHCP=yes\n")
                    # Short circuit as everything below is for STATIC configuration
                    return

                # Check if we have IP addressing
                ipv4address_config_attr = f"network_device_{device_name}_ipv4address"
                ipv4address = self.target_network_config.get(ipv4address_config_attr)

                ipv6address_config_attr = f"network_device_{device_name}_ipv6address"
                ipv6address = self.target_network_config.get(ipv6address_config_attr)

                network_file.write("ConfigureWithoutCarrier=yes\n")
                network_file.write("IgnoreCarrierLoss=yes\n")
                network_file.write("# Ignore IPv6 RA routes as we are not doing DHCP\n")
                network_file.write("IPv6AcceptRA=no\n")
                if ipv6address:
                    network_file.write("# Static IPv6 is configured, ignore duplicate addresses\n")
                    network_file.write("IPv6DuplicateAddressDetection=0\n")
                else:
                    network_file.write("# Disable link local addresssing as IPv6 is not enabled for this interface\n")
                    network_file.write("LinkLocalAddressing=no\n")

                if ipv4address:
                    network_file.write(f"Address={ipv4address}\n")
                if ipv6address:
                    network_file.write(f"Address={ipv6address}\n")

                # Check if we have DNS servers
                ipv4dns_config_attr = f"network_device_{device_name}_ipv4dns"
                if ipv4dns_config_attr in self.target_network_config:
                    for ipv4dns in self.target_network_config[ipv4dns_config_attr].split(","):
                        if ipv4dns:
                            network_file.write(f"DNS={ipv4dns}\n")
                ipv6dns_config_attr = f"network_device_{device_name}_ipv6dns"
                if ipv6dns_config_attr in self.target_network_config:
                    for ipv6dns in self.target_network_config[ipv6dns_config_attr].split(","):
                        if ipv6dns:
                            network_file.write(f"DNS={ipv6dns}\n")

                # Setup the gateways
                ipv4gateway_config_attr = f"network_device_{device_name}_ipv4gateway"
                ipv4gateway = self.target_network_config.get(ipv4gateway_config_attr)

                ipv6gateway_config_attr = f"network_device_{device_name}_ipv6gateway"
                ipv6gateway = self.target_network_config.get(ipv6gateway_config_attr)

                # If ipv4gateway is not within ipv4address's range, then add a route section
                if ipv4address and ipv4gateway and util.outside_network4(ipv4address, ipv4gateway):
                    network_file.write("\n[Route]\n")
                    network_file.write(f"Destination={ipv4gateway}\n")
                    network_file.write("Metric=256\n")
                    network_file.write("Scope=link\n")
                # Do the same check for IPV6
                if ipv6address and ipv6gateway and util.outside_network6(ipv6address, ipv6gateway):
                    network_file.write("\n[Route]\n")
                    network_file.write(f"Destination={ipv6gateway}\n")
                    network_file.write("Metric=256\n")
                    network_file.write("Scope=link\n")

                # Check if we have gateways
                if ipv4gateway:
                    network_file.write("\n[Route]\n")
                    network_file.write(f"Gateway={ipv4gateway}\n")
                    network_file.write("Metric=1000\n")

                if ipv6gateway:
                    network_file.write("\n[Route]\n")
                    network_file.write(f"Gateway={ipv6gateway}\n")
                    network_file.write("Metric=1000\n")

                network_file.write("\n")

        except OSError as err:
            raise IliRuntimeError(f"Failed to open '{network_file_path}' for reading: {err}") from None

    @property
    def config(self) -> dict[str, Any]:
        """Return configuration."""
        return self._config

    @property
    def profile(self) -> MutableMapping[str, Any] | None:
        """Return profile configuration."""
        return self._profile

    @property
    def required_config(self) -> RequiredConfigItems:
        """Return required configuration."""
        return self._required_config

    @property
    def stages(self) -> dict[int, dict[str, Any]]:
        """Return install stages."""
        return self._stages

    @property
    def rootfs(self) -> IliFilesystem | None:
        """Return the root filesystem we're going to use."""
        return self._rootfs

    @rootfs.setter
    def rootfs(self, rootfs: IliFilesystem) -> None:
        """Set the root filesystem we're going to use."""
        self._rootfs = rootfs

    @property
    def mounts(self) -> list[IliFilesystem]:
        """Return current list of mounts."""
        return self._mounts

    @property
    def kpartx_devices(self) -> list[list[Any]]:
        """Return kpartx devices we've created."""
        return self._kpartx_devices

    @property
    def loop_devices(self) -> list[IliPath]:
        """Return loop devices we've created."""
        return self._loop_devices

    @property
    def lvm_vgs(self) -> list[str]:
        """Return LVM VG's we've created."""
        return self._lvm_vgs

    @property
    def lvm_lvs(self) -> list[str]:
        """Return LVM LV's we've created."""
        return self._lvm_lvs

    @property
    def luks_volumes(self) -> list[str]:
        """Return LUKS volumes we've created."""
        return self._luks_volumes

    @property
    def mdraid_devices(self) -> list[IliPath]:
        """Return MD RAID devices we've created."""
        return self._mdraid_devices

    @property
    def args(self) -> argparse.Namespace:
        """Return arguments specified on commandline."""
        return self._args

    @property
    def overlay_directories(self) -> list[tuple[IliPath, IliPath]]:
        """Return overlay directories."""
        return self._overlay_directories

    @property
    def target_staging(self) -> bool | None:
        """Return if we're using staging repositories."""
        return self._target_staging

    @target_staging.setter
    def target_staging(self, target_staging: bool) -> None:
        """Set the state of us using staging repositories."""
        self._target_staging = target_staging

    @property
    def target_pacman_conf(self) -> IliPath | None:
        """Return the the pacman configuration file we're going to use."""
        return self._target_pacman_conf

    @target_pacman_conf.setter
    def target_pacman_conf(self, target_pacman_conf: IliPath) -> None:
        """Set the pacman configuration file to use for installing the target."""
        self._target_pacman_conf = target_pacman_conf

    @property
    def target_packages(self) -> list[str]:
        """Return package list for install."""
        return self._target_packages

    @target_packages.setter
    def target_packages(self, packages: list[str]) -> None:
        """Replace the package list we'll install for the target."""
        self._target_packages = packages

    @property
    def target_services(self) -> list[str]:
        """Return list of services to enable."""
        return self._target_services

    @target_services.setter
    def target_services(self, services: list[str]) -> None:
        """Replace the service list that we'll be starting for the target."""
        self._target_services = services

    @property
    def target_hostname(self) -> str | None:
        """Return the target hostname."""
        return self._target_hostname

    @target_hostname.setter
    def target_hostname(self, target_hostname: str) -> None:
        """Set the hostname for the target."""
        self._target_hostname = target_hostname

    @property
    def target_root_password(self) -> str | None:
        """Return the target root password."""
        return self._target_root_password

    @target_root_password.setter
    def target_root_password(self, target_root_password: str) -> None:
        """Set the root password for the target."""
        self._target_root_password = target_root_password

    @property
    def target_sudo_username(self) -> str | None:
        """Return the target sudo username."""
        return self._target_sudo_username

    @target_sudo_username.setter
    def target_sudo_username(self, target_sudo_username: str) -> None:
        """Set the sudo username for the target."""
        self._target_sudo_username = target_sudo_username

    @property
    def target_sudo_password(self) -> str | None:
        """Return the target sudo password."""
        return self._target_sudo_password

    @target_sudo_password.setter
    def target_sudo_password(self, target_sudo_password: str) -> None:
        """Set the sudo password for the target."""
        self._target_sudo_password = target_sudo_password

    @property
    def target_root_autologin(self) -> bool:
        """Return if we have root autologin enabled or not."""
        return self._target_root_autologin

    @target_root_autologin.setter
    def target_root_autologin(self, target_root_autologin: bool) -> None:
        """Return if we have root autologin enabled or not."""
        self._target_root_autologin = target_root_autologin

    @property
    def target_install_microcode(self) -> bool:
        """Return if we're installing CPU microcode for this target."""
        return self._target_install_microcode

    @target_install_microcode.setter
    def target_install_microcode(self, target_install_microcode: bool) -> None:
        """Set the state of the installation of microcode on the target."""
        self._target_install_microcode = target_install_microcode

    @property
    def target_networking(self) -> bool:
        """Return if we're going to automatically enable networking."""
        return self._target_networking

    @target_networking.setter
    def target_networking(self, target_networking: bool) -> None:
        """Set the state of automatic networking."""
        self._target_networking = target_networking

    @property
    def target_network_devices(self) -> list[str]:
        """Return the network devices we're going to use."""
        return self._target_network_devices

    @target_network_devices.setter
    def target_network_devices(self, network_devices: list[str]) -> None:
        """Set the network devices we'll be utilizing."""
        self._target_network_devices = network_devices
        self._target_network_config = {}

    @property
    def rootfs_remove_globs(self) -> list[str]:
        """Return list of rootfs paths to remove."""
        return self._rootfs_remove_globs

    @property
    def target_network_config(self) -> dict[str, str]:
        """Return the network device configuration."""
        return self._target_network_config

    @property
    def temp_paths(self) -> list[IliPath]:
        """Return list of temp paths we created."""
        return self._temp_paths
