#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# Copyright (c) 2019-2024, AllWorldIT
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""IDMS Linux installer filesystem class."""

import pathlib

from . import util
from .exceptions import IliRuntimeError

__all__ = ["IliFilesystem", "IliPath"]


class IliPath(pathlib.Path):
    """
    IDMS Linux Installer path class.

    This class add as concatpath method to the pathlib.Path class.
    """

    def concatpath(self, *args: str | pathlib.PurePath | None) -> "IliPath":
        """Concatenate path segments."""
        components = []
        for item in args:
            # Ignore blank items
            if item is None:
                continue
            # If its a pure path, lets remove "/"
            if isinstance(item, pathlib.PurePath):
                components.extend([x for x in item.parts if x != "/"])
            # If its a string, remove "/", after split it is ""
            elif isinstance(item, str):
                components.extend([x for x in item.split("/") if x != ""])
            else:
                raise TypeError(f"Invalid type '{type(item)}' for path component")

        return IliPath(self, *components)


class IliFilesystem:  # pylint: disable=too-many-instance-attributes
    """IDMS Linux Installer filesystem."""

    _source: IliPath
    _fstype: str
    _mount_root: IliPath  # Root path of the mount
    _mount_path: IliPath  # Mount path within the root path
    _mkfs_options: list[str]
    _mount_options: list[str]
    _uuid: str | None
    _mounted: bool

    def __init__(  # pylint: disable=too-many-arguments,too-many-positional-arguments
        self,
        source: IliPath,
        fstype: str,
        mount_root: IliPath,
        mount_path: IliPath,
        options: list[str] | None = None,
    ) -> None:
        """Initialize object."""
        self._source = source
        self._fstype = fstype
        self._mount_root = mount_root
        self._mount_path = mount_path
        self._mounted = False

        if options:
            self._mkfs_options = options
        else:
            self._mkfs_options = []

        self._mount_options = []

        self._uuid = None

    def mkfs(self) -> None:
        """Make filesystem and return UUID."""

        if "bind" in self.fstype:
            raise RuntimeError("Cannot mkfs on a bind mount")

        # Only add mkfs_options if its set
        args: list[str] = ["/usr/bin/mkfs", "--type", self.fstype]
        if self.mkfs_options:
            args.extend(self.mkfs_options)
        args.append(str(self.source))

        # Make filesystem
        util.run_host(args)

        # Run blkid on the device
        blkid_output = util.run_host(["/usr/bin/blkid", "--output", "export", str(self.source)])
        uuid = ""
        # Loop with the lines returned
        for line in blkid_output:
            # Split on =
            key, value = line.split("=")
            # Check if we found the UUID
            if key == "UUID":
                # If we did set it and break the loop
                uuid = value
                break
        if not uuid:
            raise IliRuntimeError(f"Failed to get UUID from '{self.source}'")
        # Set this filesystem UUID
        self._uuid = uuid

        # For btrfs things are a bit different, we need to temporarily mount the filesystem to create the subvolumes
        if self.fstype == "btrfs":
            # Mount filesystem
            self.mount()
            # Create subvolumes for @ and @home
            root_subvol = self.mount_point.concatpath("@")
            try:
                util.run_host(["btrfs", "subvolume", "create", str(root_subvol)])
            except OSError as err:
                raise IliRuntimeError(f"Failed to create root subvolume '{root_subvol}': {err}") from None

            # Unmount filesystem
            self.unmount()
            # Add subvol options
            self.mount_options.insert(0, "subvol=@")

    def mount(self, extra_mount_options: list[str] | None = None) -> None:
        """Mount src to dst."""

        mount_args = []
        if self.fstype == "bind":
            mount_args.append("--bind")
        else:
            mount_args.extend(["-t", self.fstype])

        # Work out mount options to pass
        mount_options: list[str] = []
        if self.mount_options:
            mount_options.extend(self.mount_options)
        if extra_mount_options:
            mount_options.extend(extra_mount_options)
        if mount_options:
            mount_args.extend(["-o", ",".join(mount_options)])

        # Check if path exists and if it does, if it is a directory
        if self.mount_point.exists() and not self.mount_point.is_dir():
            raise IliRuntimeError(f"Mount point '{self.mount_point}' is not a directory")

        # Create mount point
        try:
            self.mount_point.mkdir(mode=0o0755, parents=True, exist_ok=True)
        except OSError as err:
            raise IliRuntimeError(f"Failed to create boot directory '{self.mount_point}': {err}") from None

        util.run_host(["mount", str(self.source), str(self.mount_point), *mount_args])
        # Flag filesystem as mounted
        self._mounted = True

    def unmount(self) -> None:
        """Unmount a path."""

        if not self.mounted:
            raise RuntimeError(f"Cannot unmount an unmounted filesystem {self.source}")

        # Unmount path
        util.run_host(["/usr/bin/umount", str(self.mount_point)])
        # Flag filesystem as no longer mounted
        self._mounted = False

    @property
    def source(self) -> IliPath:
        """Return source."""
        return self._source

    @property
    def fstype(self) -> str:
        """Return filesystem type."""
        return self._fstype

    @property
    def mount_root(self) -> IliPath:
        """Return mount root."""
        return self._mount_root

    @property
    def mount_path(self) -> IliPath:
        """Return mount path."""
        return self._mount_path

    @property
    def mount_point(self) -> IliPath:
        """Return mount point which is based on the root + path."""
        return self.mount_root.concatpath(self.mount_path)

    @property
    def mkfs_options(self) -> list[str]:
        """Return mkfs options."""
        return self._mkfs_options

    @property
    def mount_options(self) -> list[str]:
        """Return mount options."""
        return self._mount_options

    @property
    def mounted(self) -> bool:
        """Return a bool indicating if this filesystem is mounted or not."""
        return self._mounted

    @property
    def uuid(self) -> str | None:
        """Return UUID of filesystem."""
        return self._uuid
