#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# Copyright (c) 2019-2024, AllWorldIT
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Container directory target."""

import logging

from .. import IliBase
from ..exceptions import IliRuntimeError
from ..filesystem import IliFilesystem, IliPath

__all__ = ["IliTargetContainerDir"]


class IliTargetContainerDir(IliBase):  # pylint: disable=too-many-instance-attributes
    """Ili container directory target."""

    # Internals
    _target: IliPath | None

    def target_init(self) -> None:
        """Initialize object. This is called during __init__()."""

        # Hand over to parent
        super().target_init()

        # Setup internals
        self._target = None

        # Setup our defaults
        self.target_hostname = "localhost.localdomain"
        # We allow the root user to login without a password
        self.target_root_password = " "  # noqa: S105 # nosec
        self.target_sudo_username = ""  # Disable sudo username by default
        # Auto-login by default
        self.target_root_autologin = True
        # Setup default networking
        self.target_network_devices = ["e*"]
        self.target_network_config["network_device_e*"] = "DHCP"

    def configure_target(self) -> None:
        """Configure this specific target."""

        # Grab target directory for the container
        target_dir = self.get_config_from_args_profile({"name": "target_dir"})
        if not target_dir:
            raise IliRuntimeError("No target directory specified, use --target-dir or the configuration file")
        self.target = IliPath(target_dir)

    def configure(self) -> None:
        """Set up configuration."""

        super().configure()

        # Split the configure, so it can be completely override in child classes
        self.configure_target()

    def validate_configuration_target(self) -> None:
        """Validate this specific targets configuration."""

        if self.target is None:
            raise IliRuntimeError("target must be set")

        # Make sure directory exists
        if not self.target.exists():
            raise IliRuntimeError(f"Target directory '{self.target}' does not exist")
        # Make sure target directory is a directory
        if not self.target.is_dir():
            raise IliRuntimeError(f"Target directory '{self.target}'' is not a directory")
        # Make sure the target directory is empty
        if self.target.concatpath("etc").exists():
            raise IliRuntimeError(f"Target directory '{self.target}' is not clean")

    def validate_configuration(self) -> None:
        """Validate configuration."""
        # Split the validation, so it can be completely override in child classes
        super().validate_configuration()
        self.validate_configuration_target()

    def prepare_target(self) -> None:
        """
        Prepare target.

        This must be overridden for each target.
        """

        if self.target is None:
            raise IliRuntimeError("target must be set")

        logging.debug("  - Preparing target: %s", self.target)

        # Add rootfs filesystem
        self.rootfs = IliFilesystem(source=self.target, fstype="bind", mount_root=self.target, mount_path=IliPath("/"))

    def setup_target_root_autologin(self) -> None:
        """
        Set up target root autologin.

        This is specific to containers as we do not have a tty1 or serial device.
        """

        if self.rootfs is None:
            raise IliRuntimeError("rootfs must be set")
        if self.rootfs.mount_point is None:
            raise IliRuntimeError("rootfs mount_point must be set")

        # Short circuit if we're not setting up root autologin
        if not self.target_root_autologin:
            logging.info("  - Root autologin not enabled")
            return

        logging.debug("  - Enabling root autologin")

        # Create service override for the console
        console_getty_override_dir = self.rootfs.mount_point.concatpath("etc", "systemd", "system", "console-getty.service.d")
        try:
            console_getty_override_dir.mkdir(mode=0o0755)
        except OSError as err:
            raise IliRuntimeError(f"Failed to create console getty override dir '{console_getty_override_dir}': {err}") from None

        console_getty_override_conf_path = console_getty_override_dir.concatpath("autologin.conf")
        try:
            with console_getty_override_conf_path.open(mode="w", encoding="UTF-8") as console_getty_override_conf_file:
                console_getty_override_conf_file.write("[Service]\n")
                console_getty_override_conf_file.write("ExecStart =\n")
                console_getty_override_conf_file.write(
                    r"ExecStart=-/sbin/agetty --autologin root --keep-baud 115200,38400,9600 $TERM\n"
                )
        except OSError as err:
            raise IliRuntimeError(f"Failed to open file '{console_getty_override_conf_path}' for writing: {err}") from None

        # NK: For machinectl login <container>
        # Create service override for the console
        container_getty_override_dir = self.rootfs.mount_point.concatpath("etc", "systemd", "system", "container-getty@.service.d")
        try:
            container_getty_override_dir.mkdir(mode=0o0755)
        except OSError as err:
            raise IliRuntimeError(f"Failed to create console getty override dir '{container_getty_override_dir}': {err}") from None

        container_getty_override_conf_path = container_getty_override_dir.concatpath("autologin.conf")
        with container_getty_override_conf_path.open(mode="w", encoding="UTF-8") as container_getty_override_conf_file:
            container_getty_override_conf_file.write("[Service]\n")
            container_getty_override_conf_file.write("ExecStart =\n")
            container_getty_override_conf_file.write(
                r"ExecStart=-/sbin/agetty --autologin root --keep-baud pts/%I 115200,38400,9600 $TERM\n"
            )

    @property
    def target(self) -> IliPath | None:
        """Return the target we're going to use."""
        return self._target

    @target.setter
    def target(self, target: IliPath) -> None:
        """Set the target to use."""
        self._target = target
