#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# Copyright (c) 2019-2024, AllWorldIT
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Container tar docker target."""

import logging

from ..exceptions import IliRuntimeError
from .container_tar import IliTargetContainerTar

__all__ = ["IliTargetContainerTarDocker"]


class IliTargetContainerTarDocker(IliTargetContainerTar):  # pylint: disable=too-many-instance-attributes
    """
    Ili container tar docker target.

    This class inherits from `IliTargetContainerTar`.
    """

    def target_init(self) -> None:
        """
        Initialize the target.

        This is part of the __init__ process.
        """

        # Hand over to parent
        super().target_init()

        # For docker containers we just need base
        self.target_packages = ["base"]
        self.target_services = []

        self.add_stage(7100, "Setup pacman keyring", self.setup_target_pacman_keyring)

    def pacstrap_target(self) -> None:  # noqa: PLR0915 # pylint: disable=too-many-statements
        """
        Pacstrap target.

        We override this to inject our own pacman.conf configuration.
        """

        if self.rootfs is None:
            raise IliRuntimeError("rootfs must be set")
        if self.rootfs.mount_point is None:
            raise IliRuntimeError("rootfs mount_point must be set")
        if self.target_pacman_conf is None:
            raise IliRuntimeError("target_pacman_conf must be set")

        # Create etc directory
        etc_dir = self.rootfs.mount_point.concatpath("etc")
        try:
            etc_dir.mkdir(mode=0o0755)
        except OSError as err:
            raise IliRuntimeError(f"Failed to create etc dir '{etc_dir}': {err}") from None

        # Read in the config file.write we supposed to use
        try:
            with self.target_pacman_conf.open(encoding="UTF-8") as pacman_cfg_host:
                config_lines = pacman_cfg_host.readlines()
        except OSError as err:
            raise IliRuntimeError(f"Failed to open '{self.target_pacman_conf}' for reading: {err}") from None

        # Add configuration for docker to exclude certain dirs
        config_lines.append("\n")
        config_lines.append("# Docker config")
        config_lines.append("[options]")
        config_lines.append("NoExtract = usr/share/help/* !usr/share/help/en*")
        config_lines.append("NoExtract = usr/share/gtk-doc/html/* usr/share/doc/*")
        config_lines.append("NoExtract = usr/share/locale/* usr/share/X11/locale/* usr/share/i18n/*")
        config_lines.append("NoExtract = !*locale*/en*/* !usr/share/i18n/charmaps/UTF-8.gz !usr/share/*locale*/locale.*")
        config_lines.append("NoExtract = !usr/share/*locales/en_?? !usr/share/*locales/i18n* !usr/share/*locales/iso*")
        config_lines.append("NoExtract = !usr/share/*locales/trans*")
        config_lines.append("NoExtract = usr/share/man/* usr/share/info/*")
        config_lines.append("NoExtract = usr/share/vim/vim*/lang/*")

        # Write out new config file.write
        pacman_conf_path = etc_dir.concatpath("pacman.conf")
        try:
            with pacman_conf_path.open(mode="w", encoding="UTF-8") as pacman_conf_file:
                pacman_conf_file.write("\n".join(config_lines))
        except OSError as err:
            raise IliRuntimeError(f"Failed to open '{pacman_conf_path}' for writing: {err}") from None

        # Pivot to new config file.write
        self.target_pacman_conf = pacman_conf_path

        # Create libalpm hooks directory
        libalpm_hookdir_path = self.rootfs.mount_point.concatpath("usr", "share", "libalpm", "hooks")
        try:
            libalpm_hookdir_path.mkdir(mode=0o0755, parents=True, exist_ok=True)
        except OSError as err:
            raise IliRuntimeError(f"Failed to create libalpm hooks directory '{libalpm_hookdir_path}': {err}") from None

        libalpm_pkgcleanup_hook_path = libalpm_hookdir_path.concatpath("package-cleanup.hook")
        try:
            with libalpm_pkgcleanup_hook_path.open(mode="w", encoding="UTF-8") as libalpm_pkgcleanup_hook_file:
                libalpm_pkgcleanup_hook_file.write("[Trigger]\n")
                libalpm_pkgcleanup_hook_file.write("Operation = Install\n")
                libalpm_pkgcleanup_hook_file.write("Operation = Upgrade\n")
                libalpm_pkgcleanup_hook_file.write("Type = Package\n")
                libalpm_pkgcleanup_hook_file.write("Target = *\n")
                libalpm_pkgcleanup_hook_file.write("\n")
                libalpm_pkgcleanup_hook_file.write("[Action]\n")
                libalpm_pkgcleanup_hook_file.write("Description = Cleaning up package cache...\n")
                libalpm_pkgcleanup_hook_file.write("Depends = coreutils\n")
                libalpm_pkgcleanup_hook_file.write("When = PostTransaction\n")
                libalpm_pkgcleanup_hook_file.write("Exec = /usr/bin/find /var/cache/pacman/pkg/ -type f -delete\n")
        except OSError as err:
            raise IliRuntimeError(
                f"Failed to create libalpm package cleanup config file '{libalpm_pkgcleanup_hook_path}': {err}"
            ) from None

        # Pass to parent to do the heavy lifting
        super().pacstrap_target()

    def setup_target_pacman_keyring(self) -> None:
        """Set up target pacman keyring."""

        logging.debug("  - Initializing pacman keyring")

        self.run_target(["/usr/bin/pacman-key", "--init"])
        self.run_target(["/usr/bin/pacman-key", "--populate", "idmslinux"])
