#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# Copyright (c) 2019-2024, AllWorldIT
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Generic disk target."""

# pylint: disable=too-many-lines

import logging
import os
import re
import shutil

from .. import IliBase, util
from ..exceptions import IliRuntimeError
from ..filesystem import IliFilesystem, IliPath

__all__ = ["IliTargetDiskGeneric"]


class IliTargetDiskGeneric(IliBase):  # pylint: disable=too-many-instance-attributes,too-many-public-methods
    """Ili generic image target."""

    # Target properties
    _target_create_efi: bool
    _target_create_bootfs: bool
    _target_rootfs_fstype: str | None
    _target_bootfs_fstype: str | None
    _target_bootfs_size: str
    _target_efifs_fstype: str | None
    _target_efifs_size: str

    _target_autoresize_root: bool

    _target_kernel: str

    # MBR properties
    _target_mbr_devices: list[IliPath]

    # Filesystems
    _bootfs: IliFilesystem | None
    _efifs_list: list[IliFilesystem]

    # Encrypted volume info
    _encrypted_boot_uuid: str | None
    _encrypted_hints: list[str]  # UUID hints of our encrypted volumes

    # Boot filesystem info
    _bootfs_uuid: str | None  # Used to determine the boot filesystem during boot
    _grub_bootfs_hint: str | None  # This is what we use to bootstrap the FS with grub on it
    _grub_bootfs_prefix: IliPath  # Prefix to grub on bootfs

    # mkinitcpio
    _target_mkinitcpio_feature_support: list[str]
    _target_mkinitcpio_autodetect: bool
    _target_mkinitcpio_hooks: list[str]
    _target_mkinitcpio_modules: list[str]

    # Grub
    _target_grub_preload_modules: list[str]
    _target_grub_efi_modules: list[str]

    # Serial terminal options
    _target_serial_terminal: str | None
    _target_serial_terminal_type: str | None

    # Cloud-init options
    _target_use_cloudinit: bool
    _target_cloudinit_enable_root: bool
    _target_cloudinit_admin_username: str | None

    # sshd configuration options
    _target_sshd_client_alive_interval: bool | None

    def add_mbr_device(self, device: IliPath) -> None:
        """Add MBR device."""
        self.target_mbr_devices.append(device)

    def target_init(self) -> None:  # noqa: PLR0915 # pylint: disable=too-many-statements
        """
        Initialize the target.

        This is part of the __init__ process.
        """

        # Hand over to parent
        super().target_init()

        # Target properties - used during prepare_target
        self._target_create_efi = False  # If we will use EFI or not
        self._target_create_bootfs = False  # Use separate boot filesystem

        # Support for root filesystem autoresizing
        self._target_autoresize_root = False

        self._target_kernel = "linux-lts"  # Kernel to install

        # MBR properties
        self._target_mbr_devices = []

        # Rootfs properties
        self._target_rootfs_fstype = "ext4"  # rootfs filesystem type
        self._target_bootfs_size = "500M"  # If we have our own /boot partition, make it this big
        self._target_bootfs_fstype = "ext3"  # Boot filesystem type
        self._target_efifs_size = "64M"  # Size of the EFI partition to use
        self._target_efifs_fstype = "vfat"  # EFI filesystem type, this is always vfat

        # Additional filesystems
        self._bootfs = None
        self._efifs_list = []

        # Encrypted volume info
        self._encrypted_boot_uuid = None
        self._encrypted_hints = []

        # Used to determine the boot filesystem holding the kernel and grub config
        self._bootfs_uuid = None  # Boot filesystem UUID
        self._grub_bootfs_hint = None  # Boot hint for our chain loaders
        self._grub_bootfs_prefix = IliPath("/")  # Prefix to grub on bootfs

        # mkinitcpio
        self._target_mkinitcpio_feature_support = []
        self._target_mkinitcpio_autodetect = True
        self._target_mkinitcpio_modules = ["i8042", "atkbd"]
        self._target_mkinitcpio_hooks = ["base", "udev", "keyboard", "autodetect", "modconf", "block", "filesystems", "fsck"]

        # Grub properties
        self._target_grub_preload_modules = []
        self._target_grub_efi_modules = []

        # Serial terminal options
        self._target_serial_terminal = None
        self._target_serial_terminal_type = None

        # cloud-init options
        self._target_use_cloudinit = False
        self._target_cloudinit_enable_root = True
        self._target_cloudinit_admin_username = None

        # sshd configuration options
        self._target_sshd_client_alive_interval = None

        # Add ZSH shell
        self.add_package("zsh")
        self.add_package("grml-zsh-config")

        # We need the mkinitcpio-systemd-emergency package to set the emergency environment password
        self.add_package("mkinitcpio-systemd-emergency")

        # Add packages we need for images
        self.add_package("inetutils")  # for "hostname"
        self.add_package("linux-firmware")  # Firmware for kernel modules
        self.add_package("vim")
        self.add_package("grub")
        self.add_package("mkinitcpio")
        # We add openssh by default, it is also required by cloud-init
        self.add_package("openssh")
        self.add_service("sshd")

        # Hook into stages
        self.add_stage(2110, "Create filesystems", self.prepare_target_create_filesystems)

        self.add_stage(7100, "Setup fstab", self.setup_target_fstab)

        self.add_stage(7200, "Setup systemd-repart", self.setup_target_systemd_repart)

        self.add_stage(7300, "Setup cloud-init", self.setup_target_cloudinit)
        self.add_stage(7310, "Customize sshd_config", self.setup_target_sshd_config)

        self.add_stage(7400, "Setup serial terminal", self.setup_target_serial_terminal)

        self.add_stage(7500, "Setup initcpio", self.setup_target_initcpio)
        self.add_stage(7800, "Setup grub", self.setup_target_grub)
        self.add_stage(7810, "Setup grub MBR", self.setup_target_grub_mbr)
        self.add_stage(7820, "Setup grub EFI", self.setup_target_grub_efi)
        # 7900 is used by disk_install for writing out /etc/mdadm.conf

    def add_mkinitcpio_feature(self, feature: str) -> None:
        """Add a mkinitcpio feature we need."""

        if feature not in self.target_mkinitcpio_feature_support:
            logging.debug("  - Adding mkinitcpio feature: %s", feature)
            self.target_mkinitcpio_feature_support.append(feature)

    def add_mkinitcpio_module(self, module: str) -> None:
        """Add a mkinitcpio module."""

        if module not in self.target_mkinitcpio_modules:
            logging.debug("  - Adding mkinitcpio module: %s", module)
            self.target_mkinitcpio_modules.append(module)

    def add_grub_preload_module(self, module: str) -> None:
        """Add a grub preload module."""

        if module not in self.target_grub_preload_modules:
            logging.debug("  - Adding grub preload module: %s", module)
            self.target_grub_preload_modules.append(module)

    def add_grub_efi_module(self, module: str) -> None:
        """Add a grub EFI module."""

        if module not in self.target_grub_efi_modules:
            logging.debug("  - Adding grub EFI module: %s", module)
            self.target_grub_efi_modules.append(module)

    def configure(self) -> None:  # noqa: C901,PLR0912,PLR0915
        """
        Configure target.

        This method must be overridden for each child class.
        """

        # Configure parent
        super().configure()

        # Check if we are creating an EFI bootable disk
        create_efi = self.get_config_from_args_profile({"name": "create_efi"})
        if create_efi is not None:
            logging.info(
                "  - Using commandline/profile for value 'create_efi': %s [default: %s]",
                create_efi,
                self.target_create_efi,
            )
            self.target_create_efi = create_efi
        else:
            logging.info("  - Using default value for 'create_efi': %s", self.target_create_efi)

        # Check if we are creating a separate bootfs
        rootfs_fstype = self.get_config_from_args_profile({"name": "rootfs_fstype"})
        if rootfs_fstype is not None:
            logging.info(
                "  - Using commandline/profile for value 'rootfs_fstype': %s [default: %s]",
                rootfs_fstype,
                self.target_rootfs_fstype,
            )
            self.target_rootfs_fstype = rootfs_fstype
        else:
            logging.info(
                "  - Using default value for 'rootfs_fstype': %s",
                self.target_rootfs_fstype,
            )

        # Check if we are creating a separate bootfs
        create_bootfs = self.get_config_from_args_profile({"name": "create_bootfs"})
        if create_bootfs is not None:
            logging.info(
                "  - Using commandline/profile for value 'create_bootfs': %s [default: %s]",
                create_bootfs,
                self.target_create_bootfs,
            )
            self.target_create_bootfs = create_bootfs
        else:
            logging.info(
                "  - Using default value for 'create_bootfs': %s",
                self.target_create_bootfs,
            )

        # Check if we are creating a separate bootfs
        autoresize_root = self.get_config_from_args_profile({"name": "autoresize_root"})
        if autoresize_root is not None:
            logging.info(
                "  - Using commandline/profile for value 'autoresize_root': %s [default: %s]",
                autoresize_root,
                self.target_autoresize_root,
            )
            self.target_autoresize_root = autoresize_root
        else:
            logging.info(
                "  - Using default value for 'autoresize_root': %s",
                self.target_autoresize_root,
            )

        # Grab the kernel to use
        kernel = self.get_config_from_args_profile({"name": "kernel"})
        if kernel is not None:
            logging.info(
                "  - Using commandline/profile for value 'kernel': %s [default: %s]",
                kernel,
                self.target_kernel,
            )
            self.target_kernel = kernel
        else:
            logging.info("  - Using default value for 'kernel': %s", self.target_kernel)
        # Add kernel to package list
        self.add_package(self.target_kernel)

        # Check if we're going to be using a serial console
        serial_terminal = self.get_config_from_args_profile({"name": "serial_terminal"})
        if serial_terminal is not None:
            logging.info(
                "  - Using commandline/profile for value 'serial_terminal': %s [default: %s]",
                serial_terminal,
                self.target_serial_terminal,
            )
            self.target_serial_terminal = serial_terminal
        else:
            logging.info(
                "  - Using default value for 'serial_terminal': %s",
                self.target_serial_terminal,
            )

        # Check if we're going to be using a specific serial terminal type
        serial_terminal_type = self.get_config_from_args_profile({"name": "serial_terminal_type"})
        if serial_terminal_type is not None:
            logging.info(
                "  - Using commandline/profile for value 'serial_terminal_type': %s [default: %s]",
                serial_terminal_type,
                self.target_serial_terminal_type,
            )
            self.target_serial_terminal_type = serial_terminal_type
        else:
            logging.info(
                "  - Using default value for 'serial_terminal_type': %s",
                self.target_serial_terminal_type,
            )

        # Check if we were given modules to add to the initramfs
        mkinitcpio_modules = self.get_config_from_args_profile({"name": "mkinitcpio_modules", "multivalue": True})
        if mkinitcpio_modules is not None:
            for module_list in mkinitcpio_modules:
                for module in module_list.split(","):
                    self.add_mkinitcpio_module(module)

        # Check if we're using kernel module autodetection
        mkinitcpio_autodetect = self.get_config_from_args_profile({"name": "mkinitcpio_autodetect"})
        if mkinitcpio_autodetect is not None:
            logging.info(
                "  - Using commandline/profile for value 'mkinitcpio_autodetect': %s [default: %s]",
                mkinitcpio_autodetect,
                self.target_mkinitcpio_autodetect,
            )
            self.target_mkinitcpio_autodetect = mkinitcpio_autodetect
        else:
            logging.info(
                "  - Using default value for 'mkinitcpio_autodetect': %s",
                self.target_mkinitcpio_autodetect,
            )

        # Check if we're using cloud-init
        use_cloudinit = self.get_config_from_args_profile({"name": "use_cloudinit"})
        if use_cloudinit is not None:
            logging.info(
                "  - Using commandline/profile for value 'use_cloudinit': %s [default: %s]",
                use_cloudinit,
                self.target_use_cloudinit,
            )
            self.target_use_cloudinit = use_cloudinit
        else:
            logging.info(
                "  - Using default value for 'use_cloudinit': %s",
                self.target_use_cloudinit,
            )
        # Notify user if cloud-init is enabled or not
        if self.target_use_cloudinit:
            self.add_package("cloud-init")
            self.add_service("cloud-init-local")
            self.add_service("cloud-init-main")
            self.add_service("cloud-init-network")
            self.add_service("cloud-init-hotplugd")

        # Check if we're enabling the root account when using cloud-init
        cloudinit_enable_root = self.get_config_from_args_profile({"name": "cloudinit_enable_root"})
        if cloudinit_enable_root is not None:
            logging.info(
                "  - Using commandline/profile for value 'cloudinit_enable_root': %s [default: %s]",
                cloudinit_enable_root,
                self.target_cloudinit_enable_root,
            )
            self.target_cloudinit_enable_root = cloudinit_enable_root
        else:
            logging.info(
                "  - Using default value for 'cloudinit_enable_root': %s",
                self.target_cloudinit_enable_root,
            )

        # Check if we're overriding the default cloud-init configuration username
        cloudinit_admin_username = self.get_config_from_args_profile({"name": "cloudinit_admin_username"})
        if cloudinit_admin_username is not None:
            logging.info(
                "  - Using commandline/profile for value 'cloudinit_admin_username': %s [default: %s]",
                cloudinit_admin_username,
                self.target_cloudinit_admin_username,
            )
            self.target_cloudinit_admin_username = cloudinit_admin_username
        else:
            logging.info(
                "  - Using default value for 'cloudinit_admin_username': %s",
                self.target_cloudinit_admin_username,
            )

        # Check if we're setting the sshd_config ClientAliveInterval option
        sshd_client_alive_interval = self.get_config_from_args_profile({"name": "sshd_client_alive_interval"})
        if sshd_client_alive_interval is not None:
            logging.info(
                "  - Using commandline/profile for value 'sshd_client_alive_interval': %s [default: %s]",
                sshd_client_alive_interval,
                self.target_sshd_client_alive_interval,
            )
            self.target_sshd_client_alive_interval = sshd_client_alive_interval
        else:
            logging.info(
                "  - Using default value for 'sshd_client_alive_interval': %s",
                self.target_sshd_client_alive_interval,
            )

    def validate_configuration(self) -> None:
        """Validate the configuration we have."""

        # Propagate upwards
        super().validate_configuration()

        # Check root filesystem type is valid
        if self.target_rootfs_fstype not in ("ext4", "btrfs"):
            raise IliRuntimeError(f"Invalid rootfs filesystem type '{self.target_rootfs_fstype}' specified")

        # Check kernel choice is valid
        if self.target_kernel not in ("linux", "linux-lts", "linux-hardened"):
            raise IliRuntimeError(f"Invalid kernel '{self.target_kernel}' specified")

        # Check kernel choice is valid
        if self.target_serial_terminal:  # noqa: SIM102
            if not re.match(r"^ttyS[0-9],(?:115200|38400|9600)$", self.target_serial_terminal):
                raise IliRuntimeError(f"Invalid serial_terminal '{self.target_serial_terminal}' specified")

        # If we use cloudinit and have no networking raise an exception
        if self.target_use_cloudinit and not self.target_networking:
            raise IliRuntimeError("Cloud-init is enabled, but networking is not, this will not work")

    def prepare_target(self) -> None:
        """
        Prepare target.

        This method must be overridden for each child class.
        """
        raise NotImplementedError("This method must be overridden for each child class")

    def prepare_target_create_filesystems(self) -> None:
        """Create filesystems."""

        if self.rootfs is None:
            raise IliRuntimeError("rootfs must be set")
        if self.rootfs.mount_point is None:
            raise IliRuntimeError("rootfs mount_point must be set")

        # Setup root filesystem
        logging.debug("  - Creating filesystem: rootfs [%s]", self.rootfs.mount_path)
        self.rootfs.mkfs()

        # Setup EFI filesystem(s)
        for efifs in self.efifs_list:
            logging.debug("  - Creating filesystem: efifs [%s]", efifs.mount_path)
            efifs.mkfs()

        # Setup boot filesystem
        if self.bootfs:
            logging.debug("  - Creating filesystem: bootfs [%s]", self.bootfs.mount_path)
            self.bootfs.mkfs()
            self.bootfs_uuid = self.bootfs.uuid
        # If there is no boot device, we set the bootfs uuid to the rootfs uuid
        else:
            logging.debug("  - Setting bootfs UUID to rootfs")
            # Set the bootfs uuid to the root if we don't have a separate filesystem
            self.bootfs_uuid = self.rootfs.uuid

    def mount_target(self) -> None:
        """
        Mount image target target.

        This must be overridden for each target.
        """

        # Call parent, this includes mounting rootfs
        super().mount_target()

        for efifs in self.efifs_list:
            logging.debug("  - Mounting efifs: %s", efifs.mount_path)
            self.mount(efifs)

        if self.bootfs:
            logging.debug("  - Mounting bootfs: %s", self.bootfs.mount_path)
            self.mount(self.bootfs)

    def setup_target_fstab(self) -> None:
        """Set up target fstab."""

        logging.debug("  - Creating fstab")

        if self.rootfs is None:
            raise IliRuntimeError("rootfs must be set")
        if self.rootfs.mount_point is None:
            raise IliRuntimeError("rootfs mount_point must be set")

        fstab_path = self.rootfs.mount_point.concatpath("etc", "fstab")
        try:
            with fstab_path.open(mode="w", encoding="UTF-8") as fstab_file:
                # Check if the root filesystem should be resized
                if self.target_autoresize_root:
                    self.rootfs.mount_options.append("x-systemd.growfs")

                # Work out root mount options
                mount_options = self.rootfs.mount_options or ["defaults"]
                fstab_file.write(
                    f"UUID={self.rootfs.uuid}		/		{self.rootfs.fstype}	{','.join(mount_options)}		0 1\n"
                )

                # Add boot filesystem if we have one, and work out its mount options
                if self.bootfs:
                    mount_options = self.bootfs.mount_options or ["defaults"]
                    fstab_file.write(
                        f"UUID={self.bootfs_uuid}		/boot		{self.bootfs.fstype}	{','.join(mount_options)}		0 2\n"
                    )

                # Add EFI filesystems
                for efifs in self.efifs_list:
                    # Add "nofail" and then the additional mount options
                    mount_options = ["nofail", *efifs.mount_options]
                    fstab_file.write(
                        f"UUID={efifs.uuid}		{efifs.mount_path}		{efifs.fstype}	{','.join(mount_options)}		0 2\n"
                    )

        except OSError as err:
            raise IliRuntimeError(f"Unable to open file '{fstab_path}' for writing: {err}") from None

    def setup_target_systemd_repart(self) -> None:
        """Set up systemd-repart."""

        if self.rootfs is None:
            raise IliRuntimeError("rootfs must be set")
        if self.rootfs.mount_point is None:
            raise IliRuntimeError("rootfs mount_point must be set")

        if not self.target_autoresize_root:
            logging.info("  - Auto-resizing of root not enabled")
            return

        logging.debug("  - Setting up systemd repart for auto-resizing")

        # Create the systemd repart config directory
        repart_dir_path = self.rootfs.mount_point.concatpath("etc", "repart.d")
        try:
            repart_dir_path.mkdir(mode=0o0755)
        except OSError as err:
            raise IliRuntimeError(f"Failed to create systemd-repart config directory '{repart_dir_path}': {err}") from None

        # Create systemd repart config
        repart_conf_path = self.rootfs.mount_point.concatpath("etc", "repart.d", "50-root.conf")
        try:
            with repart_conf_path.open(mode="w", encoding="UTF-8") as repart_conf_file:
                repart_conf_file.write("[Partition]\n")
                repart_conf_file.write("Type=root\n")
        except OSError as err:
            raise IliRuntimeError(f"Unable to open file '{repart_conf_path}' for writing: {err}") from None

        # Create initcpio hook
        initcpio_hook_path = self.rootfs.mount_point.concatpath("etc", "initcpio", "install", "systemd-repart")
        try:
            with initcpio_hook_path.open(mode="w", encoding="UTF-8") as initcpio_hook_file:
                initcpio_hook_file.write(". /usr/lib/initcpio/install/systemd\n")
                initcpio_hook_file.write("\n")
                initcpio_hook_file.write("function build() {\n")
                initcpio_hook_file.write("\ttest ! -e /etc/repart.d && return 0\n")
                initcpio_hook_file.write("\tmap add_systemd_unit systemd-repart.service\n")
                initcpio_hook_file.write("\tadd_full_dir /etc/repart.d\n")
                initcpio_hook_file.write("}\n")
                initcpio_hook_file.write("\n")
                initcpio_hook_file.write("help() {\n")
                initcpio_hook_file.write('\techo "Installs the systemd-repart service if we have /etc/repart.d."\n')
                initcpio_hook_file.write("}\n")
        except OSError as err:
            raise IliRuntimeError(f"Unable to open file '{initcpio_hook_path}' for writing: {err}") from None

    def setup_target_cloudinit(self) -> None:  # noqa: C901
        """Set up cloud-init."""

        if self.rootfs is None:
            raise IliRuntimeError("rootfs must be set")
        if self.rootfs.mount_point is None:
            raise IliRuntimeError("rootfs mount_point must be set")

        if not self.target_use_cloudinit:
            logging.info("  - Cloud-init is not enabled")
            return

        logging.debug("  - Setting up cloud-init")

        # Read /etc/cloud/cloud.cfg config file
        cloud_cfg_lines = []
        cloud_cfg_path = self.rootfs.mount_point.concatpath("etc", "cloud", "cloud.cfg")
        try:
            with cloud_cfg_path.open(encoding="UTF-8") as cloud_cfg_file:
                cloud_cfg_lines = cloud_cfg_file.readlines()
        except OSError as err:
            raise IliRuntimeError(f"Unable to open file '{cloud_cfg_path}' for reading: {err}") from None
        # Write file out
        cloudinit_disable_root_found = False
        cloudinit_admin_username_found = False
        try:
            with cloud_cfg_path.open(mode="w", encoding="UTF-8") as cloud_cfg_file:
                # Loop with lines from config file
                for line in cloud_cfg_lines:
                    new_line = line
                    # If we need to enable the root account
                    if self.target_cloudinit_enable_root:
                        # Check if this is the configuration line for the disabling of root
                        matches = re.match(r"^(?P<disable_root>\s*disable_root:\s*)true.*", line)
                        if matches:
                            # We got what we were looking for
                            cloudinit_disable_root_found = True
                            # Grab the bit before the value
                            disable_root_str = matches.group("disable_root")
                            # Rebuild the line
                            new_line = f"{disable_root_str} false\n"
                    # If we need to change the admin username
                    if self.target_cloudinit_admin_username:
                        # Check if this is the configuration line for the admin username
                        matches = re.match(r"^(?P<admin_username>\s*name:\s*)osadmin.*", line)
                        if matches:
                            # We got what we were looking for
                            cloudinit_admin_username_found = True
                            # Grab the bit before the value
                            admin_username_str = matches.group("admin_username")
                            # Rebuild the line
                            new_line = f"{admin_username_str} {self.target_cloudinit_admin_username}\n"
                    # Write out line
                    cloud_cfg_file.write(new_line)

        except OSError as err:
            raise IliRuntimeError(f"Unable to open file '{cloud_cfg_path}' for writing: {err}") from None

        # Make sure we found the items we're replacing
        if self.target_cloudinit_enable_root and not cloudinit_disable_root_found:
            logging.debug("++++ cloud.cfg:\n%s", "".join(cloud_cfg_lines))
            raise IliRuntimeError("Unable to find disable_root: to replace")
        if self.target_cloudinit_admin_username and not cloudinit_admin_username_found:
            logging.debug("++++ cloud.cfg:\n%s", "".join(cloud_cfg_lines))
            raise IliRuntimeError("Unable to find name: osadmin to replace")

    def setup_target_sshd_config(self) -> None:
        """Set up sshd_config."""

        if self.rootfs is None:
            raise IliRuntimeError("rootfs must be set")
        if self.rootfs.mount_point is None:
            raise IliRuntimeError("rootfs mount_point must be set")

        # Skip configuration if we have nothing set
        if self.target_sshd_client_alive_interval is None:
            return

        logging.debug("  - Setting up sshd_config defaults")
        sshd_config_override_dir = self.rootfs.mount_point.concatpath("etc", "ssh", "sshd_config.d")
        # If the path doesn't exist create it
        if not sshd_config_override_dir.exists():
            try:
                sshd_config_override_dir.mkdir(mode=0o0755)
            except OSError as err:
                raise IliRuntimeError(
                    f"Failed to create sshd_config override directory '{sshd_config_override_dir}': {err}"
                ) from None

        sshd_config_override_conf_path = sshd_config_override_dir.concatpath("00-defaults-ilitools.conf")
        try:
            with sshd_config_override_conf_path.open(mode="w", encoding="UTF-8") as sshd_config_override_conf_file:
                sshd_config_override_conf_file.write(f"ClientAliveInterval {self.target_sshd_client_alive_interval}\n")
        except OSError as err:
            raise IliRuntimeError(f"Unable to open file '{sshd_config_override_conf_path}' for writing: {err}") from None

    def setup_target_serial_terminal(self) -> None:
        """Set up target serial terminal."""

        if self.rootfs is None:
            raise IliRuntimeError("rootfs must be set")
        if self.rootfs.mount_point is None:
            raise IliRuntimeError("rootfs mount_point must be set")

        if not self.target_serial_terminal:
            logging.info("  - Serial terminal is not enabled")
            return

        logging.debug("  - Setting up serial terminal")

        # Grab the serial port
        serial_port = self.target_serial_terminal.split(",", 1)[0]

        # Create service override to set TERM= if we have it specified
        if self.target_serial_terminal_type:
            getty_serial_override_dir = self.rootfs.mount_point.concatpath(
                "etc", "systemd", "system", f"serial-getty@{serial_port}.service.d"
            )
            # If the path doesn't exist create it, it may of been created due to autologin configuration
            if not getty_serial_override_dir.exists():
                try:
                    getty_serial_override_dir.mkdir(mode=0o0755)
                except OSError as err:
                    raise IliRuntimeError(
                        f"Failed to create getty serial override directory '{getty_serial_override_dir}': {err}"
                    ) from None

            getty_serial_override_conf_path = getty_serial_override_dir.concatpath("term.conf")
            try:
                with getty_serial_override_conf_path.open(mode="w", encoding="UTF-8") as getty_serial_override_conf_file:
                    getty_serial_override_conf_file.write("[Service]\n")
                    getty_serial_override_conf_file.write(f"Environment=TERM={self.target_serial_terminal_type}\n")
            except OSError as err:
                raise IliRuntimeError(f"Unable to open file '{getty_serial_override_conf_path}' for writing: {err}") from None

    def setup_target_initcpio(self) -> None:
        """Set up target initcpio."""

        if self.rootfs is None:
            raise IliRuntimeError("rootfs must be set")
        if self.rootfs.mount_point is None:
            raise IliRuntimeError("rootfs mount_point must be set")

        modules = self.target_mkinitcpio_modules.copy()
        hooks = self.target_mkinitcpio_hooks.copy()

        logging.debug("  - Generating initramfs configuration")

        # Use systemd instead of udev
        systemd_index = hooks.index("udev")
        hooks[systemd_index] = "systemd"
        # Fix systemd-emergency environment
        hooks.insert(systemd_index + 1, "systemd-emergency")

        # Check if we need to add lvm support
        if "lvm" in self.target_mkinitcpio_feature_support:
            # If we do, insert before the filesystems hook
            hooks.insert(hooks.index("block") + 1, "lvm2")

        # Check if we need to add encryption support
        if "encrypt" in self.target_mkinitcpio_feature_support:
            # If we do, insert after the block hook
            hooks.insert(hooks.index("block") + 1, "sd-encrypt")

        # Check if we need to add mdadm support
        if "mdadm" in self.target_mkinitcpio_feature_support:
            # If we do, insert after the block hook
            hooks.insert(hooks.index("block") + 1, "mdadm_udev")

        # Add keyboard drivers before autodetect
        del hooks[hooks.index("keyboard")]
        hooks.insert(hooks.index("autodetect"), "keyboard")

        # Check if we need to remove the autodetection of modules
        if not self.target_mkinitcpio_autodetect:
            del hooks[hooks.index("autodetect")]

        # Add rootfs resizing
        if self.target_autoresize_root:
            # Add systemd-repart before the filesystems hook
            repart_index = hooks.index("filesystems")
            hooks.insert(repart_index, "systemd-repart")

        # Build list of modules
        mkinitcpio_modules_list = " ".join(modules)
        logging.debug("    - MKINITCPIO MODULES: %s", mkinitcpio_modules_list)
        # Build list of hooks
        mkinitcpio_hooks_list = " ".join(hooks)
        logging.debug("    - MKINITCPIO HOOKS: %s", mkinitcpio_hooks_list)

        # Write config out
        mkinitcpio_conf_path = self.rootfs.mount_point.concatpath("etc", "mkinitcpio.conf.d", "00-ilitools.conf")
        try:
            with mkinitcpio_conf_path.open(mode="w", encoding="UTF-8") as fstab_file:
                fstab_file.write(f"MODULES=({mkinitcpio_modules_list})\n")
                fstab_file.write(f"HOOKS=({mkinitcpio_hooks_list})\n")
        except OSError as err:
            raise IliRuntimeError(f"Unable to open file '{mkinitcpio_conf_path}' for writing: {err}") from None

        logging.debug("  - Generating initramfs")

        # Create target initcpio
        self.run_target(["/usr/bin/mkinitcpio", "--preset", self.target_kernel])

    def setup_target_grub(self) -> None:  # noqa: C901,PLR0912,PLR0915
        """Set up grub."""

        if self.rootfs is None:
            raise IliRuntimeError("rootfs must be set")
        if self.rootfs.mount_point is None:
            raise IliRuntimeError("rootfs mount_point must be set")

        logging.debug("  - Generating grub configuration")

        # Create /boot/grub
        grub_dir_path = self.rootfs.mount_point.concatpath("boot", "grub")
        try:
            grub_dir_path.mkdir(mode=0o0755)
        except OSError as err:
            raise IliRuntimeError(f"Failed to create grub directory '{grub_dir_path}': {err}") from None

        # Copy over grub splash image
        grub_splash = self.resolve_data_file("artwork/splash.png")
        rootfs_grub_splash = self.rootfs.mount_point.concatpath("boot", "grub", "splash.png")
        try:
            shutil.copy(grub_splash, rootfs_grub_splash)
        except OSError as err:
            raise IliRuntimeError(f"Failed to copy grub splash image '{grub_splash}' to '{rootfs_grub_splash}': {err}") from None

        # Read /etc/default/grub config file
        default_grub_conf_lines = []
        default_grub_conf_path = self.rootfs.mount_point.concatpath("etc", "default", "grub")
        try:
            with default_grub_conf_path.open(encoding="UTF-8") as default_grub_conf_file:
                default_grub_conf_lines = default_grub_conf_file.readlines()
        except OSError as err:
            raise IliRuntimeError(f"Unable to open file '{default_grub_conf_path}' for reading: {err}") from None
        # Write file out
        grub_preload_modules_changed = False
        try:
            with default_grub_conf_path.open(mode="w", encoding="UTF-8") as default_grub_conf_file:
                # Loop with lines from config file
                for line in default_grub_conf_lines:
                    new_line = line
                    # If we have grub modules to pre-load
                    if self.target_grub_preload_modules:
                        # Check if this is the configuration line for the modules
                        matches = re.match(r'^\s*GRUB_PRELOAD_MODULES="(?P<modules>.*)"', line)
                        if matches:
                            # We found what we're looking for
                            grub_preload_modules_changed = True

                            # If we can split it into a list
                            modules = matches.group("modules").split()

                            # Add grub modules we need
                            modules.extend(self.target_grub_preload_modules)

                            # Rebuild the line
                            new_line = 'GRUB_PRELOAD_MODULES="' + " ".join(modules) + '")\n'

                    # If we have an encrypted disk, enable GRUB's crypto disk support
                    if self.encrypted_hints:
                        # Check if we need to enable encryption support
                        matches = re.match(r"^#?\s*GRUB_ENABLE_CRYPTODISK=", line)
                        if matches:
                            new_line = "GRUB_ENABLE_CRYPTODISK=y"

                    # Write out line
                    default_grub_conf_file.write(new_line)

                # Check if we're using a serial terminal
                if self.encrypted_hints:
                    cmdline_str = " ".join([f"rd.luks.name={x}" for x in self.encrypted_hints])
                    default_grub_conf_file.write("# START - Encrypted disk\n")
                    default_grub_conf_file.write(f'GRUB_CMDLINE_LINUX="$GRUB_CMDLINE_LINUX {cmdline_str}"\n')
                    default_grub_conf_file.write("# END - Encrypted disk\n")

                # Check if we're using a serial terminal
                if self.target_serial_terminal:
                    unit, speed = self.target_serial_terminal.replace("ttyS", "").split(",")
                    default_grub_conf_file.write("# START - Serial Console\n")
                    default_grub_conf_file.write('GRUB_TERMINAL_INPUT="serial"\n')
                    default_grub_conf_file.write('GRUB_TERMINAL_OUTPUT="serial"\n')
                    default_grub_conf_file.write(f'GRUB_SERIAL_COMMAND="serial --unit={unit} --speed={speed}"\n')
                    default_grub_conf_file.write(f'GRUB_CMDLINE_LINUX="$GRUB_CMDLINE_LINUX console=ttyS{unit},{speed}"\n')
                    default_grub_conf_file.write("# END - Serial Console\n")

        except OSError as err:
            raise IliRuntimeError(f"Unable to open file '{default_grub_conf_path}' for writing: {err}") from None

        # Make sure we found the items we're replacing
        if self.target_grub_preload_modules and not grub_preload_modules_changed:
            logging.debug("++++ default/grub:\n%s", "".join(default_grub_conf_lines))
            raise IliRuntimeError("Unable to find GRUB_PRELOAD_MODULES= to replace")

        # Create grub config file
        self.run_target(["/usr/bin/grub-mkconfig", "--output", "/boot/grub/grub.cfg"])

    def setup_target_grub_mbr(self) -> None:
        """Set up grub MBRs."""

        # If we don't have any mbr devices, just return
        if not self.target_mbr_devices:
            logging.info("  - No MBR devices to create")
            return

        # Install grub on each MBR device
        for mbr_device in self.target_mbr_devices:
            logging.debug("    - Setting up grub MBR: %s", mbr_device)
            self.run_target(["/usr/bin/grub-install", "--target", "i386-pc", str(mbr_device)])

    def setup_target_grub_efi(self) -> None:
        """Set up grub EFI."""

        # If we do not have EFI filesystems, just return
        if not self.efifs_list:
            logging.info("  - Grub EFI is not enabled")
            return

        logging.debug("  - Setting up grub EFI")

        # Loop with the EFI filesystems
        for efifs in self.efifs_list:
            logging.debug("  - Setting up grub EFI: %s", efifs.mount_path)

            # Create /efi/EFI
            efi_dir_path = efifs.mount_point.concatpath("EFI")
            try:
                efi_dir_path.mkdir(mode=0o0755)
            except OSError as err:
                raise IliRuntimeError(f"Failed to create EFI directory '{efi_dir_path}': {err}") from None
            # Create /efi/EFI/BOOT
            efi_boot_dir_path = efi_dir_path.concatpath("BOOT")
            try:
                efi_boot_dir_path.mkdir(mode=0o0755)
            except OSError as err:
                raise IliRuntimeError(f"Failed to create EFI/BOOT directory '{efi_boot_dir_path}': {err}") from None

            # Create /efi/grub
            efi_grub_dir_path = efifs.mount_point.concatpath("grub")
            try:
                efi_grub_dir_path.mkdir(mode=0o0755)
            except OSError as err:
                raise IliRuntimeError(f"Failed to create EFI/grub directory '{efi_grub_dir_path}': {err}") from None

            # Write out grub stub config
            umask_orig = os.umask(0o027)
            efi_grub_cfg_path = efifs.mount_point.concatpath("grub", "grub.cfg")
            try:
                with efi_grub_cfg_path.open(mode="w", encoding="UTF-8") as efi_grub_cfg_file:
                    efi_grub_cfg_file.write("insmod part_gpt\n")
                    efi_grub_cfg_file.write("insmod part_msdos\n")
                    # Check if there is any modules we need in the EFI config to find the bootfs
                    for module in self.target_grub_efi_modules:
                        efi_grub_cfg_file.write(f"insmod {module}\n")
                    # Check if we need to mount our /boot first
                    if self.encrypted_boot_uuid:
                        efi_grub_cfg_file.write("insmod cryptodisk\n")
                        efi_grub_cfg_file.write("insmod luks\n")
                        efi_grub_cfg_file.write(f"cryptomount -u {self.encrypted_boot_uuid}\n")
                    # Find bootfs...
                    efi_grub_cfg_file.write(
                        f"search --no-floppy --fs-uuid --set=bootfs --hint='{self.grub_bootfs_hint}' {self.bootfs_uuid}\n"
                    )
                    # Chcek how we're going to load the grub config
                    if self.target_create_bootfs:
                        grub_cfg_path = self.grub_bootfs_prefix.concatpath("grub", "grub.cfg")
                        efi_grub_cfg_file.write(f"configfile ($bootfs){grub_cfg_path}\n")
                    else:
                        grub_cfg_path = self.grub_bootfs_prefix.concatpath("boot", "grub", "grub.cfg")
                        efi_grub_cfg_file.write(f"configfile ($bootfs){grub_cfg_path}\n")
            except OSError as err:
                raise IliRuntimeError(f"Failed to write grub config file '{efi_grub_cfg_path}': {err}") from None
            finally:
                os.umask(umask_orig)

            # Write out grub standalone image
            efi_grub_cfg_path = efifs.mount_path.concatpath("grub", "grub.cfg")
            self.run_target(
                [
                    "/usr/bin/grub-mkstandalone",
                    "--format",
                    "x86_64-efi",
                    "--output",
                    str(efifs.mount_path.concatpath("EFI", "BOOT", "BOOTX64.EFI")),
                    f"boot/grub/grub.cfg={efi_grub_cfg_path}",
                ]
            )

        logging.debug("    - Disabling systemd-boot-update as we're using grub")
        self.run_target(["/usr/bin/systemctl", "disable", "systemd-boot-update"])

    def write_partitions(self, path: IliPath, disk_label: str, partitions: list[str]) -> None:
        """Write out partitions to disk."""

        logging.debug("  - Setting up partitions: %s", path)

        partitions_str = "\n".join(partitions)
        util.run_host(
            ["/usr/bin/sfdisk", "--wipe", "always", "--wipe-partitions", "always", "--label", disk_label, str(path)],
            stdin=f"{partitions_str}\n",
        )

        # Wait for partitions to be written to disk
        self.wait_block_settle()

    @property
    def target_create_efi(self) -> bool:
        """Return if we're using EFI boot on this image."""
        return self._target_create_efi

    @target_create_efi.setter
    def target_create_efi(self, target_create_efi: bool) -> None:
        """Set the state of our EFI usage."""
        self._target_create_efi = target_create_efi

    @property
    def target_create_bootfs(self) -> bool:
        """Return if we're using a boot filesystem on this image."""
        return self._target_create_bootfs

    @target_create_bootfs.setter
    def target_create_bootfs(self, target_create_bootfs: bool) -> None:
        """Set the state of our usage of a boot filesystem."""
        self._target_create_bootfs = target_create_bootfs

    @property
    def target_rootfs_fstype(self) -> str | None:
        """Return the rootfs fstype we're going to use."""
        return self._target_rootfs_fstype

    @target_rootfs_fstype.setter
    def target_rootfs_fstype(self, target_rootfs_fstype: str) -> None:
        """Set the rootfs fstype to use."""
        self._target_rootfs_fstype = target_rootfs_fstype

    @property
    def target_bootfs_fstype(self) -> str | None:
        """Return the boot fstype we're going to use."""
        return self._target_bootfs_fstype

    @target_bootfs_fstype.setter
    def target_bootfs_fstype(self, target_bootfs_fstype: str) -> None:
        """Set the boot fstype to use."""
        self._target_bootfs_fstype = target_bootfs_fstype

    @property
    def target_bootfs_size(self) -> str | None:
        """Return the boot filesystem size."""
        return self._target_bootfs_size

    @target_bootfs_size.setter
    def target_bootfs_size(self, target_bootfs_size: str) -> None:
        """Set the boot filesystem size."""
        self._target_bootfs_size = target_bootfs_size

    @property
    def target_efifs_fstype(self) -> str | None:
        """Return the EFI fstype we're going to use."""
        return self._target_efifs_fstype

    @target_efifs_fstype.setter
    def target_efifs_fstype(self, target_efifs_fstype: str) -> None:
        """Set the EFI fstype to use."""
        self._target_efifs_fstype = target_efifs_fstype

    @property
    def target_efifs_size(self) -> str | None:
        """Return the EFI size."""
        return self._target_efifs_size

    @target_efifs_size.setter
    def target_efifs_size(self, target_efifs_size: str) -> None:
        """Set the EFI image size."""
        self._target_efifs_size = target_efifs_size

    @property
    def target_autoresize_root(self) -> bool:
        """
        Return if the root filesystem should be resizable.

        Defaults to False.
        """
        return self._target_autoresize_root

    @target_autoresize_root.setter
    def target_autoresize_root(self, target_autoresize_root: bool) -> None:
        """Set to enable image automatic resizing."""
        self._target_autoresize_root = target_autoresize_root

    @property
    def target_kernel(self) -> str:
        """Return target_kernel we're going to use."""
        return self._target_kernel

    @target_kernel.setter
    def target_kernel(self, target_kernel: str) -> None:
        """Set the kernel to use."""
        self._target_kernel = target_kernel

    @property
    def target_mbr_devices(self) -> list[IliPath]:
        """Return list of MBR devices."""
        return self._target_mbr_devices

    @property
    def bootfs(self) -> IliFilesystem | None:
        """Return the boot filesystem we're going to use."""
        return self._bootfs

    @bootfs.setter
    def bootfs(self, bootfs: IliFilesystem) -> None:
        """Set the boot filesystem we're going to use."""
        self._bootfs = bootfs

    @property
    def efifs_list(self) -> list[IliFilesystem]:
        """Return the EFI filesystems we're going to use."""
        return self._efifs_list

    @property
    def encrypted_boot_uuid(self) -> str | None:
        """Return the encrypted uuid we're going to use."""
        return self._encrypted_boot_uuid

    @encrypted_boot_uuid.setter
    def encrypted_boot_uuid(self, encrypted_boot_uuid: str) -> None:
        """Set the encrypted uuid to use."""
        self._encrypted_boot_uuid = encrypted_boot_uuid

    @property
    def encrypted_hints(self) -> list[str]:
        """Return the encrypted hint we're going to use."""
        return self._encrypted_hints

    @property
    def bootfs_uuid(self) -> str | None:
        """Return the bootfs uuid we're going to use."""
        return self._bootfs_uuid

    @bootfs_uuid.setter
    def bootfs_uuid(self, bootfs_uuid: str | None) -> None:
        """Set the bootfs uuid to use."""
        self._bootfs_uuid = bootfs_uuid

    @property
    def grub_bootfs_hint(self) -> str | None:
        """Return the boot hint we're going to use."""
        return self._grub_bootfs_hint

    @grub_bootfs_hint.setter
    def grub_bootfs_hint(self, grub_bootfs_hint: str) -> None:
        """Set the boot hint."""
        self._grub_bootfs_hint = grub_bootfs_hint

    @property
    def grub_bootfs_prefix(self) -> IliPath:
        """Return the grub bootfs prefix we're going to use."""
        return self._grub_bootfs_prefix

    @grub_bootfs_prefix.setter
    def grub_bootfs_prefix(self, grub_bootfs_prefix: IliPath) -> None:
        """Set the grub bootfs prefix."""
        self._grub_bootfs_prefix = grub_bootfs_prefix

    @property
    def target_mkinitcpio_feature_support(self) -> list[str]:
        """Return the mkinitcpio features we'll be using."""
        return self._target_mkinitcpio_feature_support

    @property
    def target_mkinitcpio_autodetect(self) -> bool:
        """Return if we're using module auto-detection."""
        return self._target_mkinitcpio_autodetect

    @target_mkinitcpio_autodetect.setter
    def target_mkinitcpio_autodetect(self, mkinitcpio_autodetect: bool) -> None:
        """Set if we're using mkinitcpio autodetection or not."""
        self._target_mkinitcpio_autodetect = mkinitcpio_autodetect

    @property
    def target_mkinitcpio_hooks(self) -> list[str]:
        """Return the list of mkinitcpio hooks we're adding to the initramfs."""
        return self._target_mkinitcpio_hooks

    @property
    def target_mkinitcpio_modules(self) -> list[str]:
        """Return the kernel modules we're adding to the initramfs."""
        return self._target_mkinitcpio_modules

    @property
    def target_grub_preload_modules(self) -> list[str]:
        """Return the grub modules we'll be pre-loading."""
        return self._target_grub_preload_modules

    @property
    def target_grub_efi_modules(self) -> list[str]:
        """Return the grub EFI modules we'll be loading."""
        return self._target_grub_efi_modules

    @property
    def target_serial_terminal(self) -> str | None:
        """Return if we are using a serial terminal."""
        return self._target_serial_terminal

    @target_serial_terminal.setter
    def target_serial_terminal(self, serial_terminal: str) -> None:
        """Set the state of the serial terminal."""
        self._target_serial_terminal = serial_terminal

    @property
    def target_serial_terminal_type(self) -> str | None:
        """Return a specific serial terminal type we're using."""
        return self._target_serial_terminal_type

    @target_serial_terminal_type.setter
    def target_serial_terminal_type(self, serial_terminal_type: str) -> None:
        """Set the serial terminal type we're using."""
        self._target_serial_terminal_type = serial_terminal_type

    @property
    def target_use_cloudinit(self) -> bool:
        """Return if we are using cloud-init."""
        return self._target_use_cloudinit

    @target_use_cloudinit.setter
    def target_use_cloudinit(self, target_use_cloudinit: bool) -> None:
        """Set the state of cloud-init usage."""
        self._target_use_cloudinit = target_use_cloudinit

    @property
    def target_cloudinit_enable_root(self) -> bool:
        """Return if we are enabling the root account or not."""
        return self._target_cloudinit_enable_root

    @target_cloudinit_enable_root.setter
    def target_cloudinit_enable_root(self, cloudinit_enable_root: bool) -> None:
        """Set the state of the root account enabling in cloud-init."""
        self._target_cloudinit_enable_root = cloudinit_enable_root

    @property
    def target_cloudinit_admin_username(self) -> str | None:
        """Return the admin username we're using in cloud-init."""
        return self._target_cloudinit_admin_username

    @target_cloudinit_admin_username.setter
    def target_cloudinit_admin_username(self, cloudinit_admin_username: str) -> None:
        """Set the admin username for cloud-init."""
        self._target_cloudinit_admin_username = cloudinit_admin_username

    @property
    def target_sshd_client_alive_interval(self) -> bool | None:
        """Return the ClientAliveInterval to set in sshd_config."""
        return self._target_sshd_client_alive_interval

    @target_sshd_client_alive_interval.setter
    def target_sshd_client_alive_interval(self, client_alive_interval: bool) -> None:
        """Set the ClientAliveInterval in sshd_config."""
        self._target_sshd_client_alive_interval = client_alive_interval
