#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# Copyright (c) 2019-2024, AllWorldIT
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Generic disk target."""

import logging
from typing import Any

from .. import util
from ..exceptions import IliRuntimeError
from ..filesystem import IliFilesystem, IliPath
from .disk_generic import IliTargetDiskGeneric

__all__ = ["IliTargetDiskImage"]

TARGET_FORMATS: dict[str, dict[str, Any]] = {
    ".raw": {"format": "raw"},
    ".qcow2": {"format": "qcow2"},
    # Azure is says their image is VHD but it is actually VPC
    ".azure.vhd": {"format": "vpc", "options": ["-o", "subformat=fixed,force_size"]},
}


class IliTargetDiskImage(IliTargetDiskGeneric):  # pylint: disable=too-many-instance-attributes,too-many-public-methods
    """Ili disk image target."""

    # Build properties
    _build_dir: IliPath | None
    _build_image: IliPath | None

    # Target disk image properties
    _target_image: IliPath | None
    _target_format: str | None
    _target_format_options: list[str]
    _target_size: str
    _target_disk_label: str

    def target_init(self) -> None:
        """
        Initialize the target.

        This is part of the __init__ process.
        """

        # Hand over to parent
        super().target_init()

        # Build properties
        self._build_dir = None  # Directory we'll be building in
        self._build_image = None  # The .raw image we use for building

        self._target_image = None  # Target of the image file
        self._target_format = None  # Format of the image file
        self._target_format_options = []  # Extra options for the conversion command
        self._target_size = "4G"  # Size of image to create
        self._target_disk_label = "gpt"  # Default to GPT as the disk label

        # Images default to enabling autoresizing of the root partition
        self.target_autoresize_root = True

        # Set our defaults
        self.target_hostname = "localhost.localdomain"
        self.target_root_password = " "  # We allow the root user to login without a password # noqa: S105 # nosec: B105
        self.target_sudo_username = ""
        self.target_mkinitcpio_autodetect = False
        self.target_use_cloudinit = True
        # Setup default networking
        self.target_network_devices = ["e*"]
        self.target_network_config["network_device_e*"] = "DHCP"

    def configure(self) -> None:
        """Set up configuration."""

        # Run parent configuration
        super().configure()

        # Grab target directory for the image
        target_image = self.get_config_from_args_profile({"name": "target_image"})
        if not target_image:
            raise IliRuntimeError("No target image specified, use --target-image or the configuration file")
        self.target_image = IliPath(target_image)

        # Grab the target image size
        target_size = self.get_config_from_args_profile({"name": "image_size"})
        if target_size is not None:
            logging.info(
                "  - Using commandline/profile for value 'image_size': %s [default: %s]",
                target_size,
                self.target_size,
            )
            self.target_size = target_size
        else:
            logging.info("  - Using default value for 'image_size': %s", self.target_size)

        # Grab the target disk label
        disk_label = self.get_config_from_args_profile({"name": "disk_label"})
        if disk_label is not None:
            logging.info(
                "  - Using commandline/profile for value 'disk_label': %s [default: %s]",
                disk_label,
                self.target_disk_label,
            )
            self.target_disk_label = disk_label
        else:
            logging.info("  - Using default value for 'disk_label': %s", self.target_disk_label)

        build_dir = None
        for ext, info in TARGET_FORMATS.items():
            # Check for format match
            # NK: Make mypy and pylance happy by adding self.target_image
            if self.target_image and self.target_image.name.endswith(ext):
                # Set the build directory
                build_dir = self.target_image.with_suffix(".BUILD-TMP")
                self.target_format = info["format"]
                self.target_format_options = info.get("options", [])
        # If we didn't get a build_dir it means we didn't understand the file extension
        if not build_dir:
            raise IliRuntimeError(f"Invalid extension for target file '{self.target_image}'")
        # Make sure build directory does not exist
        if build_dir.exists():
            raise IliRuntimeError(f'Temporary build directory "{build_dir}" already exists')
        # Save build dir
        self.build_dir = build_dir
        # Create raw build image filename
        self.build_image = self.build_dir.concatpath("raw.img")

    def post_configure(self) -> None:
        """Post configure."""

        super().post_configure()

        # If the filesystem is btrfs, add btrfs packages
        if self.target_rootfs_fstype == "btrfs":
            self.add_package("btrfs-progs")
            self.add_package("grub-btrfs")
            self.add_package("inotify-tools")
            # Disable bootfs if we're using btrfs
            logging.info("  - Disabling 'create_bootfs' for 'btrfs' to support consistent system snapshots")
            self.target_create_bootfs = False
            self.grub_bootfs_prefix = IliPath("/@")

    def validate_configuration(self) -> None:
        """Validate configuration."""

        super().validate_configuration()

        # Validate target disk label
        if self.target_disk_label not in ("gpt", "dos"):
            raise IliRuntimeError(f"Target disk label '{self.target_disk_label}' is invalid, can only be 'gpt' or 'dos'")

        # Warn the user that automatic rootfs resizing will probably not work if we have a DOS disk label
        if self.target_autoresize_root and (self.target_disk_label == "dos"):
            raise IliRuntimeError(f"Auto-resizing of rootfs will not work with a '{self.target_disk_label}' disk label")

    def prepare_target(self) -> None:  # noqa: C901,PLR0912,PLR0915 # pylint: disable=too-many-branches,too-many-statements
        """
        Prepare target.

        This must be overridden for each target.
        """

        if self.build_dir is None:
            raise IliRuntimeError("build_dir must be set")
        if self.build_image is None:
            raise IliRuntimeError("build_image must be set")
        if self.target_size is None:
            raise IliRuntimeError("target_size must be set")
        if self.target_disk_label is None:
            raise IliRuntimeError("target_disk_label must be set")
        if self.target_rootfs_fstype is None:
            raise IliRuntimeError("target_rootfs_fstype must be set")

        # Make the build directory and the rootfs directory
        try:
            self.build_dir.mkdir(mode=0o0750)
            self.add_temp_path(self.build_dir)
        except OSError as err:
            raise IliRuntimeError(f"Failed to create build directory '{self.build_dir}': {err}") from None

        rootfs_path = self.build_dir.concatpath("rootfs")
        try:
            rootfs_path.mkdir(mode=0o0755)
            self.add_temp_path(rootfs_path)
        except OSError as err:
            raise IliRuntimeError(f"Failed to create rootfs path '{rootfs_path}': {err}") from None

        logging.debug("  - Build image: %s", self.build_image)

        # Create image
        util.run_host(["/usr/bin/qemu-img", "create", "-f", "raw", str(self.build_image), self.target_size])

        # Setup loop device
        loop_device = self.add_loop_device(self.build_image)
        loop_device_name = loop_device.name

        # Add MBR device
        logging.debug("  - Adding MBR: %s", loop_device)
        self.add_mbr_device(loop_device)

        parts = []
        efi_part_num = 0
        boot_part_num = 0
        # GPT partitioning
        if self.target_disk_label == "gpt":
            # Add DOS boot partition
            parts.append(",1M,21686148-6449-6E6F-744E-656564454649")
        # Add EFI partition
        if self.target_create_efi:
            parts.append(f",{self.target_efifs_size},{self.part_type_efi}")
            efi_part_num = len(parts)
        # The boot partition is the next one in the list
        self.grub_bootfs_hint = f"hd0,{self.target_disk_label}{len(parts) + 1}"
        # Add bootfs
        if self.target_create_bootfs:
            parts.append(f",{self.target_bootfs_size},{self.part_type_bootfs}")
            boot_part_num = len(parts)
        # Add root partition
        parts.append(f",,{self.part_type_rootfs}")
        root_part_num = len(parts)

        # Write out partitions
        logging.debug("  - Writing partition table: %s", loop_device)
        self.write_partitions(loop_device, self.target_disk_label, parts)

        # Add kpartx to the loop device to generate the partition devices
        logging.debug("  - Adding partition block devices to device mapper: %s", loop_device)
        self.add_kpartx_device(loop_device)

        logging.debug("  - Adding rootfs")
        self.rootfs = IliFilesystem(
            source=IliPath("/dev/mapper").concatpath(f"{loop_device_name}p{root_part_num}"),
            fstype=self.target_rootfs_fstype,
            mount_root=rootfs_path,
            mount_path=IliPath("/"),
            options=["-L", "root"],
        )
        if self.target_create_efi:
            if self.target_efifs_fstype is None:
                raise IliRuntimeError("target_efifs_fstype must be set")
            logging.debug("  - Adding EFI")
            self.efifs_list.append(
                IliFilesystem(
                    source=IliPath("/dev/mapper").concatpath(f"{loop_device_name}p{efi_part_num}"),
                    fstype=self.target_efifs_fstype,
                    mount_root=rootfs_path,
                    mount_path=IliPath("/efi"),
                    options=["-n", "EFI", "-F", "32"],
                )
            )
        if self.target_create_bootfs:
            if self.target_bootfs_fstype is None:
                raise IliRuntimeError("target_bootfs_fstype must be set")
            logging.debug("  - Adding bootfs")
            self.bootfs = IliFilesystem(
                source=IliPath("/dev/mapper").concatpath(f"{loop_device_name}p{boot_part_num}"),
                fstype=self.target_bootfs_fstype,
                mount_root=rootfs_path,
                mount_path=IliPath("/boot"),
                options=["-L", "boot"],
            )

    def finalize_target(self) -> None:
        """
        Finalize target.

        Here we create the final image. This is after unmounts have happened.
        """

        if self.target_format is None:
            raise IliRuntimeError("target_format must be set")
        if self.build_image is None:
            raise IliRuntimeError("build_image must be set")
        if self.target_image is None:
            raise IliRuntimeError("target_image must be set")

        logging.debug(
            "  - Converting target image into: %s [%s]",
            self.target_image,
            self.target_format,
        )

        util.run_host(
            [
                "/usr/bin/qemu-img",
                "convert",
                "-f",
                "raw",
                *self.target_format_options,
                "-O",
                self.target_format,
                str(self.build_image),
                str(self.target_image),
            ]
        )

        logging.info("  - Target built: %s", self.target_image)

    @property
    def part_type_efi(self) -> str:
        """Return the EFI partition type depending on the disk label used."""
        if self.target_disk_label == "gpt":
            return "C12A7328-F81F-11D2-BA4B-00A0C93EC93B"
        if self.target_disk_label == "dos":
            return "ef"
        raise IliRuntimeError(f"EFI partition type unknown for disk label '{self.target_disk_label}'")

    @property
    def part_type_bootfs(self) -> str:
        """Return the bootfs partition type depending on the disk label used."""
        if self.target_disk_label == "gpt":
            return "BC13C2FF-59E6-4262-A352-B275FD6F7172"
        if self.target_disk_label == "dos":
            return "ea"
        raise IliRuntimeError(f"bootfs partition type unknown for disk label '{self.target_disk_label}'")

    @property
    def part_type_rootfs(self) -> str:
        """Return the rootfs partition type depending on the disk label used."""
        if self.target_disk_label == "gpt":
            return "4F68BCE3-E8CD-4DB1-96E7-FBCAF984B709"
        if self.target_disk_label == "dos":
            return "83"
        raise IliRuntimeError(f"rootfs partition type unknown for disk label '{self.target_disk_label}'")

    @property
    def build_dir(self) -> IliPath | None:
        """Return the build dir we're going to use."""
        return self._build_dir

    @build_dir.setter
    def build_dir(self, build_dir: IliPath) -> None:
        """Set the build dir to use."""
        self._build_dir = build_dir

    @property
    def build_image(self) -> IliPath | None:
        """Return the build image we're going to use."""
        return self._build_image

    @build_image.setter
    def build_image(self, build_image: IliPath) -> None:
        """Set the build image to use."""
        self._build_image = build_image

    @property
    def target_image(self) -> IliPath | None:
        """Return target image we're going to use."""
        return self._target_image

    @target_image.setter
    def target_image(self, target_image: IliPath) -> None:
        """Set the target image to use."""
        self._target_image = target_image

    @property
    def target_format(self) -> str | None:
        """Return target_format we're going to use."""
        return self._target_format

    @target_format.setter
    def target_format(self, target_format: str) -> None:
        """Set the target_format to use."""
        self._target_format = target_format

    @property
    def target_format_options(self) -> list[str]:
        """Return options to use during the convert process."""
        return self._target_format_options

    @target_format_options.setter
    def target_format_options(self, target_format_options: list[str]) -> None:
        """Set the target format options to use."""
        self._target_format_options = target_format_options

    @property
    def target_size(self) -> str | None:
        """Return the image size we're going to use."""
        return self._target_size

    @target_size.setter
    def target_size(self, target_size: str) -> None:
        """Set the image size to use."""
        self._target_size = target_size

    @property
    def target_disk_label(self) -> str | None:
        """Return the target disk label we're going to use."""
        return self._target_disk_label

    @target_disk_label.setter
    def target_disk_label(self, target_disk_label: str) -> None:
        """Set the target disk label to use."""
        self._target_disk_label = target_disk_label
