#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# Copyright (c) 2019-2024, AllWorldIT
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=too-many-lines

"""Live image target."""

import datetime
import logging
import shutil

from .. import IliBase, util
from ..exceptions import IliRuntimeError
from ..filesystem import IliFilesystem, IliPath

__all__ = ["IliTargetLiveImage"]


class IliTargetLiveImage(IliBase):  # pylint: disable=too-many-instance-attributes,too-many-public-methods
    """
    Ili live image target.

    This class inherits from `IliBase` as almost everything is unique to this target type.
    """

    # Build directory
    _build_dir: IliPath | None

    # Target image filename
    _target_file: IliPath | None

    _target_kernels: list[str]
    _target_autorun: str | None
    _target_ilitools_profile: str | None

    _target_compression_level: int

    _isofs: IliFilesystem | None

    _isofs_rootfs_img: IliPath | None
    # isofs label
    _isofs_img_mdate: str
    _isofs_img_uuid: str
    _isofs_img_volid: str
    _isofs_img_appid: str
    _isofs_img_publisher: str
    _isofs_img_preparer: str

    # mkinitcpio
    _target_mkinitcpio_hooks: list[str]
    _target_mkinitcpio_modules: list[str]

    def target_init(self) -> None:
        """Initialize object. This is called during __init__()."""

        # Hand over to parent
        super().target_init()

        # Setup internals
        self._build_dir = None

        self._target_file = None
        self._target_kernel = ["linux-lts", "linux"]  # Kernels to install
        self._target_autorun = None  # Script to automatically run on boot
        self._target_ilitools_profile = None  # Pre-seeded ilitools profile to use

        self._target_compression_level = 22  # Zstd compression level

        self._isofs = None  # ISO filesystem of the image we're building
        self._isofs_rootfs_img = None  # The filename of the rootfs image on the ISO

        # We need to set the UUID of the image here, so we can pre-load it into the grub.cfg
        mdate = datetime.datetime.now(datetime.UTC)
        self._isofs_img_mdate = mdate.strftime(r"%Y%m%d%H%M%S99")
        self._isofs_img_uuid = mdate.strftime(r"%Y-%m-%d-%H-%M-%S-99")
        self._isofs_img_volid = f"IDMS_{mdate.strftime(r'%Y%m%d')}"
        self._isofs_img_appid = "IDMS Linux Live/Rescue/Install"
        self._isofs_img_publisher = "IDMS Linux <https://www.idms-linux.org>"
        self._isofs_img_preparer = "Prepared by IDMS Linux Installer"

        # mkinitcpio
        self._target_mkinitcpio_hooks = ["base", "udev", "keyboard", "autodetect", "modconf", "block", "filesystems", "fsck"]
        self._target_mkinitcpio_modules = ["i8042", "atkbd", "9p_virtio"]

        # Boot stuff
        self.add_package("edk2-shell")
        self.add_package("memtest86+")
        self.add_package("memtest86+-efi")

        # Add ZSH shell
        self.add_package("zsh")
        self.add_package("grml-zsh-config")

        # We need the mkinitcpio-systemd-emergency package to set the emergency environment password
        self.add_package("mkinitcpio-systemd-emergency")

        # Add packages we need for images
        self.add_package("inetutils")  # for "hostname"
        self.add_package("linux-firmware")  # Firmware for kernel modules
        self.add_package("vim")
        self.add_package("grub")
        # Tools
        self.add_package("mdadm")
        self.add_package("lvm2")
        # Hook into stages
        self.add_stage(
            7200,
            "Setup live media systemd services",
            self.setup_target_systemd_services_live,
        )
        self.add_stage(7400, "Setup initramfs", self.setup_target_initcpio)
        self.add_stage(7500, "Move boot files to ISO", self.setup_target_bootfiles)
        self.add_stage(7800, "Setup Grub config", self.setup_target_grub)
        self.add_stage(7810, "Setup Grub Eltorito", self.setup_target_grub_eltorito)
        self.add_stage(7820, "Setup Grub EFI", self.setup_target_grub_efi)
        self.add_stage(7830, "Setup Grub EFI efi.img", self.setup_target_grub_efi_img)
        # Add our autologin/autorun magic
        self.add_stage(8250, "Setup running of automated scripts", self.setup_target_autorun_script)

        # Setup defaults
        self.target_hostname = "localhost.localdomain"
        # We allow the root user to login without a password
        self.target_root_password = " "  # noqa: S105 # nosec
        self.target_root_autologin = True
        self.target_sudo_username = ""
        self.target_install_microcode = True
        # Setup default networking
        self.target_network_devices = ["e*"]
        self.target_network_config["network_device_e*"] = "DHCP"

    def configure(self) -> None:  # noqa: PLR0912
        """Configure target."""

        # Parent configuration
        super().configure()

        # Grab target filename for the image
        target_file = self.get_config_from_args_profile({"name": "target_file"})
        if not target_file:
            raise IliRuntimeError("No target file specified, use --target-file or the configuration file")
        self.target_file = IliPath(target_file)

        build_dir = self.target_file.with_suffix(".BUILD-TMP")  # type: ignore[none]
        # Make sure build directory does not exist
        if build_dir.exists():
            raise IliRuntimeError(f'Temporary build directory "{build_dir}" already exists')
        # Save build dir
        self.build_dir = build_dir

        # Grab the kernel to use
        target_kernel = self.get_config_from_args_profile({"name": "kernel"})
        if target_kernel is not None:
            logging.info(
                "  - Using commandline/profile for value 'kernel': %s [default: %s]",
                target_kernel,
                self.target_kernels,
            )
            # Check if target_kernel is a list, if it is, use as is, else if it is a string split it on ,
            if isinstance(target_kernel, list):
                self.target_kernels = target_kernel
            elif isinstance(target_kernel, str):
                self.target_kernels = list(set(target_kernel.split(",")))
            else:
                raise IliRuntimeError(f"Invalid kernel '{target_kernel}' specified")
        else:
            logging.info("  - Using default value for 'kernel': %s", ", ".join(self.target_kernels))
        # Add kernel to package list
        for kernel_package in self.target_kernels:
            self.add_package(kernel_package)

        # Grab the autorun to use
        target_autorun = self.get_config_from_args_profile({"name": "autorun"})
        if target_autorun is not None:
            logging.info(
                "  - Using commandline/profile for value 'autorun': %s [default: %s]",
                target_autorun,
                self.target_autorun,
            )
            self.target_autorun = target_autorun
        else:
            logging.info("  - Using default value for 'autorun': %s", self.target_autorun)

        # Grab the ilitools_profile to use
        target_ilitools_profile = self.get_config_from_args_profile({"name": "ilitools_profile"})
        if target_ilitools_profile is not None:
            logging.info(
                "  - Using commandline/profile for value 'ilitools_profile': %s [default: %s]",
                target_ilitools_profile,
                self.target_ilitools_profile,
            )
            self.target_ilitools_profile = target_ilitools_profile
        else:
            logging.info(
                "  - Using default value for 'ilitools_profile': %s",
                self.target_ilitools_profile,
            )

        # Check if we have a compression level set
        target_compression_level = self.get_config_from_args_profile({"name": "compression_level"})
        if target_compression_level is not None:
            logging.info(
                "  - Using commandline/profile for value 'compression_level': %s [default: %s]",
                target_compression_level,
                self.target_compression_level,
            )
            self.target_compression_level = target_compression_level
        else:
            logging.info(
                "  - Using default value for 'compression_level': %s",
                self.target_compression_level,
            )

    def validate_configuration(self) -> None:
        """Validate configuration."""

        super().validate_configuration()

        # Check kernel choice is valid by matching each list item to the valid kernel list
        for kernel in self.target_kernels:
            if kernel not in ("linux", "linux-lts", "linux-hardened"):
                raise IliRuntimeError(f"Invalid kernel '{kernel}' specified")

        # For autorun to work we need
        if self.target_autorun and not self.target_root_autologin:
            raise IliRuntimeError("For autorun to work root autologin must be enabled")

    def prepare_target(self) -> None:
        """
        Prepare target.

        This must be overridden for each target.
        """

        if self.build_dir is None:
            raise IliRuntimeError("build_dir must be set")

        # Make the build directory and the rootfs directory
        try:
            self.build_dir.mkdir(mode=0o0750)
            self.add_temp_path(self.build_dir)
        except OSError as err:
            raise IliRuntimeError(f"Failed to create build directory '{self.build_dir}': {err}") from None

        rootfs_path = self.build_dir.concatpath("rootfs")
        try:
            rootfs_path.mkdir(mode=0o0755)
            self.add_temp_path(rootfs_path)
        except OSError as err:
            raise IliRuntimeError(f"Failed to create rootfs path '{rootfs_path}': {err}") from None

        isofs_path = self.build_dir.concatpath("isofs")
        try:
            isofs_path.mkdir(mode=0o0755)
            self.add_temp_path(isofs_path)
        except OSError as err:
            raise IliRuntimeError(f"Failed to create isofs path '{isofs_path}': {err}") from None
        # Save image isofs path
        self.isofs_path = isofs_path
        # Work out the path of the rootfs image on the ISO
        self.isofs_rootfs_img = self.isofs_path.concatpath("rootfs.sfs")  # type: ignore[none]

        self.rootfs = IliFilesystem(
            source=rootfs_path,
            fstype="bind",
            mount_root=rootfs_path,
            mount_path=IliPath("/"),
        )

    def setup_target_autorun_script(self) -> None:
        """Set running of autorun scripts."""

        if self.rootfs is None:
            raise IliRuntimeError("rootfs must be set")
        if self.rootfs.mount_point is None:
            raise IliRuntimeError("rootfs mount_point must be set")

        # Grab location of autorun script
        autorun_script_src = self.resolve_data_file("scripts/idms-live-autorun")
        autorun_script_dst = self.rootfs.mount_point.concatpath("usr", "local", "bin", "idms-live-autorun")
        try:
            shutil.copy(autorun_script_src, autorun_script_dst)
        except OSError as err:
            raise IliRuntimeError(f"Failed to copy script file '{autorun_script_src}' to '{autorun_script_dst}': {err}") from None
        try:
            autorun_script_dst.chmod(0o0755)  # nosec
        except OSError as err:
            raise IliRuntimeError(f"Failed to chmod script file '{autorun_script_dst}': {err}") from None

    def setup_target_systemd_services_live(self) -> None:
        """Set up target systemd services for live media."""

        if self.rootfs is None:
            raise IliRuntimeError("rootfs must be set")
        if self.rootfs.mount_point is None:
            raise IliRuntimeError("rootfs mount_point must be set")

        # Systemd directories
        rootfs_systemd_system_path = self.rootfs.mount_point.concatpath("usr", "local", "lib", "systemd", "system")

        # Create tmpfs pacman gnupg mount point
        systemd_pacman_gnupg_mount = self.resolve_data_file("systemd/system/etc-pacman.d-gnupg.mount")
        systemd_pacman_gnupg_mount_dst = rootfs_systemd_system_path.concatpath("etc-pacman.d-gnupg.mount")
        shutil.copy(systemd_pacman_gnupg_mount, systemd_pacman_gnupg_mount_dst)
        # Activate mount
        self.add_service("etc-pacman.d-gnupg.mount")

        # Mount OEM volume
        systemd_oem_mount_src = self.resolve_data_file("systemd/system/OEM.mount")
        systemd_oem_mount_dst = rootfs_systemd_system_path.concatpath("OEM.mount")
        shutil.copy(systemd_oem_mount_src, systemd_oem_mount_dst)

        # Mount host volume
        systemd_host_mount_src = self.resolve_data_file("systemd/system/host.mount")
        systemd_host_mount_dst = rootfs_systemd_system_path.concatpath("host.mount")
        shutil.copy(systemd_host_mount_src, systemd_host_mount_dst)

        # Setup the autorun script using the below
        systemd_autorun_sshd_config_service_src = self.resolve_data_file("systemd/system/autorun-sshd-config.service")
        systemd_autorun_sshd_config_service_dst = rootfs_systemd_system_path.concatpath("autorun-sshd-config.service")
        shutil.copy(systemd_autorun_sshd_config_service_src, systemd_autorun_sshd_config_service_dst)

        # Setup the autorun service
        systemd_autorun_service_src = self.resolve_data_file("systemd/system/autorun.service")
        systemd_autorun_service_dst = rootfs_systemd_system_path.concatpath("autorun.service")
        shutil.copy(systemd_autorun_service_src, systemd_autorun_service_dst)
        # Activate service
        self.add_service("autorun")

        # Create /etc/systemd/system/getty@.service.d/noclear.conf to override tty1 being cleared on boot
        systemd_getty_noclear_path = rootfs_systemd_system_path.concatpath("getty@.service.d", "noclear.conf")
        try:
            systemd_getty_noclear_path.parent.mkdir(mode=0o0755, exist_ok=True)
        except OSError as err:
            raise IliRuntimeError(
                f"Failed to create getty@.service.d directory '{systemd_getty_noclear_path.parent}': {err}"
            ) from None
        try:
            with systemd_getty_noclear_path.open(mode="w", encoding="UTF-8") as noclear_file:
                noclear_file.write("[Service]\n")
                noclear_file.write("TTYVTDisallocate=no\n")
        except OSError as err:
            raise IliRuntimeError(f"Unable to open file '{systemd_getty_noclear_path}' for writing: {err}") from None

    def setup_target_initcpio(self) -> None:  # pylint: disable=too-many-locals,too-many-statements
        """Set up target initcpio."""

        if self.rootfs is None:
            raise IliRuntimeError("rootfs must be set")
        if self.rootfs.mount_point is None:
            raise IliRuntimeError("rootfs mount_point must be set")

        modules = self.target_mkinitcpio_modules.copy()
        hooks = self.target_mkinitcpio_hooks.copy()

        logging.debug("  - Generating initramfs configuration")

        # Use systemd instead of udev
        systemd_index = hooks.index("udev")
        hooks[systemd_index] = "systemd"
        # Fix systemd-emergency environment
        hooks.insert(systemd_index + 1, "systemd-emergency")

        # Add keyboard driver loading after autodetect
        keyboard_index = hooks.index("keyboard")
        del hooks[keyboard_index]
        keyboard_index = hooks.index("autodetect") + 1
        hooks.insert(keyboard_index, "keyboard")

        # Remove autodetect
        autodetect_index = hooks.index("autodetect")
        del hooks[autodetect_index]

        # Remove fsck
        fsck_index = hooks.index("fsck")
        del hooks[fsck_index]

        # Add live hook
        live_index = hooks.index("block")
        hooks.insert(live_index + 1, "sd-live")

        mkinitcpio_modules_list = " ".join(modules)
        logging.debug("    - MKINITCPIO MODULES: %s", mkinitcpio_modules_list)

        # Build list of modules
        mkinitcpio_modules_list = " ".join(modules)
        logging.debug("    - MKINITCPIO MODULES: %s", mkinitcpio_modules_list)
        # Build list of hooks
        mkinitcpio_hooks_list = " ".join(hooks)
        logging.debug("    - MKINITCPIO HOOKS: %s", mkinitcpio_hooks_list)

        # Write config out
        mkinitcpio_conf_path = self.rootfs.mount_point.concatpath("etc", "mkinitcpio.conf.d", "00-ilitools.conf")
        try:
            with mkinitcpio_conf_path.open(mode="w", encoding="UTF-8") as fstab_file:
                fstab_file.write(f"MODULES=({mkinitcpio_modules_list})\n")
                fstab_file.write(f"HOOKS=({mkinitcpio_hooks_list})\n")
        except OSError as err:
            raise IliRuntimeError(f"Unable to open file '{mkinitcpio_conf_path}' for writing: {err}") from None

        rootfs_initcpio_path = self.rootfs.mount_point.concatpath("etc", "initcpio")

        # Create hooks dir
        rootfs_initcpio_hooks_path = rootfs_initcpio_path.concatpath("hooks")
        try:
            rootfs_initcpio_hooks_path.mkdir(mode=0o0755, parents=True, exist_ok=True)
        except OSError as err:
            raise IliRuntimeError(f"Failed to create initcpio hooks directory '{rootfs_initcpio_hooks_path}': {err}") from None

        # Create install dir
        rootfs_initcpio_install_path = rootfs_initcpio_path.concatpath("install")
        try:
            rootfs_initcpio_install_path.mkdir(mode=0o0755, parents=True, exist_ok=True)
        except OSError as err:
            raise IliRuntimeError(f"Failed to create initcpio install directory '{rootfs_initcpio_install_path}': {err}") from None

        # Copy in sd-live.install to rootfs /etc/mkinicpio/install/sd-live
        mkinitcpio_install_sdlive_path = self.resolve_data_file("mkinitcpio/sd-live.install")
        shutil.copy(mkinitcpio_install_sdlive_path, rootfs_initcpio_install_path.concatpath("sd-live"))

        for kernel in self.target_kernels:
            logging.debug("  - Generating initramfs for kernel: %s", kernel)

            # Create target initcpio
            self.run_target(
                ["/usr/bin/mkinitcpio", "--preset", kernel],
                env={"LIVE_UUID": self.isofs_img_uuid},
            )

    def setup_target_bootfiles(self) -> None:  # noqa: C901,PLR0912,PLR0915 # pylint: disable=too-many-locals,too-many-statements
        """Move boot files to ISO."""

        if self.isofs_path is None:
            raise IliRuntimeError("isofs_path must be set")
        if self.rootfs is None:
            raise IliRuntimeError("rootfs must be set")
        if self.rootfs.mount_point is None:
            raise IliRuntimeError("rootfs mount_point must be set")

        # Create boot dir
        isofs_boot_path = self.isofs_path.concatpath("boot")
        try:
            isofs_boot_path.mkdir(mode=0o0755, parents=True, exist_ok=True)
        except OSError as err:
            raise IliRuntimeError(f"Failed to create boot directory '{isofs_boot_path}': {err}") from None

        # Loop with kernels and copy the kernel and initramfs's
        for kernel in self.target_kernels:
            # Copy the kernel to the isofs
            logging.debug("  - Boot files: kernel (%s)", kernel)
            rootfs_kernel = self.rootfs.mount_point.concatpath("boot", f"vmlinuz-{kernel}")
            isofs_kernel = isofs_boot_path.concatpath(f"vmlinuz-{kernel}")
            try:
                shutil.copy(rootfs_kernel, isofs_kernel)
            except OSError as err:
                raise IliRuntimeError(f"Failed to copy kernel '{rootfs_kernel}' to '{isofs_kernel}': {err}") from None

            # Copy the initramfs to the isofs
            logging.debug("  - Boot files: initramfs (%s)", kernel)
            rootfs_initramfs = self.rootfs.mount_point.concatpath("boot", f"initramfs-{kernel}.img")
            isofs_initramfs = isofs_boot_path.concatpath(f"initramfs-{kernel}.img")
            try:
                shutil.copy(rootfs_initramfs, isofs_initramfs)
            except OSError as err:
                raise IliRuntimeError(f"Failed to copy initramfs '{rootfs_initramfs}' to '{isofs_initramfs}': {err}") from None

            # Remove fallback image, the only diff normally between the two is the fallback has all the modules loaded
            # In this case we include all the modules in the normal image anyway.
            rootfs_initramfs_fallback = self.rootfs.mount_point.concatpath("boot", f"initramfs-{kernel}-fallback.img")
            try:
                rootfs_initramfs_fallback.unlink()
            except OSError as err:
                raise IliRuntimeError(f"Failed to remove initramfs fallback '{rootfs_initramfs_fallback}': {err}") from None

        # Short circuit if we did not install microcode
        if self.target_install_microcode:
            logging.debug("  - Boot files: microcode")
            # Copy AMD microcode
            rootfs_microcode_amd = self.rootfs.mount_point.concatpath("boot", "amd-ucode.img")
            isofs_microcode_amd = isofs_boot_path.concatpath("amd-ucode.img")
            try:
                shutil.copy(rootfs_microcode_amd, isofs_microcode_amd)
            except OSError as err:
                raise IliRuntimeError(
                    f"Failed to copy AMD microcode '{rootfs_microcode_amd}' to '{isofs_microcode_amd}': {err}"
                ) from None
            # And license file
            rootfs_microcode_amd_license = self.rootfs.mount_point.concatpath(
                "usr", "share", "licenses", "amd-ucode", "LICENSE.amd-ucode"
            )
            isofs_microcode_amd_license = isofs_boot_path.concatpath("amd-ucode.img.LICENSE")
            try:
                shutil.copy(rootfs_microcode_amd_license, isofs_microcode_amd_license)
            except OSError as err:
                raise IliRuntimeError(
                    f"Failed to copy AMD microcode LICENSE '{rootfs_microcode_amd_license}' "
                    f"to '{isofs_microcode_amd_license}': {err}"
                ) from None

            # Copy Intel microcode
            rootfs_microcode_intel = self.rootfs.mount_point.concatpath("boot", "intel-ucode.img")
            isofs_microcode_intel = isofs_boot_path.concatpath("intel-ucode.img")
            try:
                shutil.copy(rootfs_microcode_intel, isofs_microcode_intel)
            except OSError as err:
                raise IliRuntimeError(
                    f"Failed to copy Intel microcode '{rootfs_microcode_intel}' to '{isofs_microcode_intel}': {err}"
                ) from None
            # And license file
            rootfs_microcode_intel_license = self.rootfs.mount_point.concatpath(
                "usr", "share", "licenses", "intel-ucode", "LICENSE"
            )
            isofs_microcode_intel_license = isofs_boot_path.concatpath("intel-ucode.img.LICENSE")
            try:
                shutil.copy(rootfs_microcode_intel_license, isofs_microcode_intel_license)
            except OSError as err:
                raise IliRuntimeError(
                    f"Failed to copy Intel microcode LICENSE '{rootfs_microcode_intel_license}' "
                    f"to '{isofs_microcode_intel_license}': {err}"
                ) from None

    def setup_target_grub(self) -> None:  # noqa: C901,PLR0912,PLR0915
        """Set up Grub configuration."""

        if self.isofs_path is None:
            raise IliRuntimeError("isofs_path must be set")
        if self.rootfs is None:
            raise IliRuntimeError("rootfs must be set")
        if self.rootfs.mount_point is None:
            raise IliRuntimeError("rootfs mount_point must be set")

        # Create boot/Grub dir
        isofs_grub_path = self.isofs_path.concatpath("boot", "grub")
        try:
            isofs_grub_path.mkdir(mode=0o0755, parents=True, exist_ok=True)
        except OSError as err:
            raise IliRuntimeError(f"Failed to create boot/Grub directory '{isofs_grub_path}': {err}") from None

        # Copy over Grub splash image
        grub_splash = self.resolve_data_file("artwork/splash.png")
        isofs_grub_splash = isofs_grub_path.concatpath("splash.png")
        try:
            shutil.copy(grub_splash, isofs_grub_splash)
        except OSError as err:
            raise IliRuntimeError(f"Failed to copy Grub splash image '{grub_splash}' to '{isofs_grub_splash}': {err}") from None

        # If we're installing microcode, add the initrd images
        initrd_extra = ""
        if self.target_install_microcode:
            initrd_extra = "/boot/intel-ucode.img /boot/amd-ucode.img "

        # Extra kernel command line options
        extra_cmdline = ""
        if self.target_autorun:
            extra_cmdline += f" autorun={self.target_autorun}"
        if self.target_ilitools_profile:
            extra_cmdline += f" ilitools.profile={self.target_ilitools_profile}"

        # Get kernel versions
        logging.debug("  - Getting kernel versions")
        kernel_versions: dict[str, str] = {}
        for kernel in self.target_kernels:
            kernel_version_output = self.run_target(["/usr/bin/pacman", "-Q", kernel])
            # Check we got one line back
            if len(kernel_version_output) != 1:
                raise IliRuntimeError(f"Failed to get kernel version for '{kernel}'")
            # Check kernel package version is valid
            kernel_package_version = kernel_version_output[0].split(" ")
            if len(kernel_package_version) != 2:  # noqa: PLR2004
                raise IliRuntimeError(f"Failed to get kernel package version for '{kernel}'")
            # Split off kernel version
            kernel_versions[kernel] = kernel_package_version[1]

        # Create Grub configuration
        isofs_grub_cfg = isofs_grub_path.concatpath("grub.cfg")
        with isofs_grub_cfg.open(mode="w", encoding="UTF-8") as grub_cfg_file:
            grub_cfg_file.write("# Early load\n")
            grub_cfg_file.write("insmod echo\n")
            grub_cfg_file.write("echo\n")
            grub_cfg_file.write('echo "IDMS Linux initializing GRUB..."\n')
            grub_cfg_file.write("insmod sleep\n")
            grub_cfg_file.write("insmod test\n")
            grub_cfg_file.write("insmod serial\n")
            grub_cfg_file.write("insmod part_msdos\n")
            grub_cfg_file.write("insmod part_gpt\n")
            grub_cfg_file.write("insmod ext2\n")
            grub_cfg_file.write("insmod fat\n")
            grub_cfg_file.write("insmod search\n")
            grub_cfg_file.write("insmod search_label\n")
            grub_cfg_file.write("insmod configfile\n")
            grub_cfg_file.write("insmod linux\n")
            grub_cfg_file.write("echo\n")
            grub_cfg_file.write("\n")
            grub_cfg_file.write('echo "GRUB: Looking for OEM volume..."\n')
            grub_cfg_file.write("search --no-floppy --set oem --label OEM\n")
            grub_cfg_file.write("echo\n")
            grub_cfg_file.write("sleep 2\n")
            grub_cfg_file.write("\n")

            grub_cfg_file.write("# Setup video\n")
            grub_cfg_file.write('echo "GRUB: Loading video..."\n')
            grub_cfg_file.write("insmod all_video\n")
            grub_cfg_file.write("insmod gfxterm\n")
            grub_cfg_file.write("insmod font\n")
            grub_cfg_file.write("if loadfont unicode; then\n")
            grub_cfg_file.write('\techo "GRUB: Loading GFX terminal..."\n')
            grub_cfg_file.write("\tinsmod gfxterm_background\n")
            grub_cfg_file.write("\tinsmod png\n")
            grub_cfg_file.write("\tset gfxmode=640x480\n")
            grub_cfg_file.write("\t# Load background and change to graphical terminal\n")
            grub_cfg_file.write("\tterminal_input console\n")
            grub_cfg_file.write("\tterminal_output gfxterm\n")
            grub_cfg_file.write("\tbackground_image /boot/grub/splash.png\n")
            grub_cfg_file.write("fi\n")
            grub_cfg_file.write("\n")

            grub_cfg_file.write("# Setup serial\n")
            grub_cfg_file.write('echo "GRUB: Loading serial console support..."\n')
            grub_cfg_file.write("if serial --unit=0 --speed=115200; then\n")
            grub_cfg_file.write('\techo "GRUB: Setting up serial..."\n')
            grub_cfg_file.write("\tterminal_input --append serial\n")
            grub_cfg_file.write("\tterminal_output --append serial\n")
            grub_cfg_file.write("fi\n")
            grub_cfg_file.write("\n")

            # If we're doing autologin, we need to set a timeout
            if self.target_root_autologin:
                grub_cfg_file.write("\n")
                grub_cfg_file.write("# Autologin is enabled so timeout is set")
                grub_cfg_file.write("set timeout_style=countdown\n")
                grub_cfg_file.write("set timeout=3\n")

            # Search for the OEM partition, load additional configuration if found
            grub_cfg_file.write('if [ -n "$oem" -a -f "($oem)/grub.cfg" ]; then\n')
            grub_cfg_file.write('\techo "GRUB: Loading OEM configuration [$oem]..."\n')
            grub_cfg_file.write("\tsleep 2\n")
            grub_cfg_file.write('\tsource "($oem)/grub.cfg"\n')
            grub_cfg_file.write("else\n")
            grub_cfg_file.write('\techo "GRUB: No OEM configuration"\n')
            grub_cfg_file.write("fi\n")
            grub_cfg_file.write("echo\n")

            grub_cfg_file.write("sleep 2\n")
            grub_cfg_file.write("\n")

            grub_cfg_file.write("# Menu entries\n")

            # Work out main kernel
            main_kernel = self.target_kernels[0]
            extra_kernels = self.target_kernels[1:]

            # Main live entry
            grub_cfg_file.write(f'menuentry "Boot IDMS Linux live [{kernel_versions[main_kernel]}]" {{\n')
            grub_cfg_file.write("\tset gfxpayload=keep\n")
            grub_cfg_file.write("\techo\n")
            grub_cfg_file.write('\techo "IDMS Linux: Loading kernel..."\n')
            grub_cfg_file.write(f"\tlinux /boot/vmlinuz-{main_kernel}{extra_cmdline}\n")
            grub_cfg_file.write('\techo "IDMS Linux: Loading initrd images..."\n')
            grub_cfg_file.write(f"\tinitrd {initrd_extra}/boot/initramfs-{main_kernel}.img\n")
            grub_cfg_file.write('\techo "IDMS Linux: Booting..."\n')
            grub_cfg_file.write("\tsleep 1\n")
            grub_cfg_file.write("}\n")
            grub_cfg_file.write("\n")
            # Serial entry
            grub_cfg_file.write(f'menuentry "Boot IDMS Linux live (serial) [{kernel_versions[main_kernel]}]" {{\n')
            grub_cfg_file.write("\tset gfxpayload=text\n")
            grub_cfg_file.write("\techo\n")
            grub_cfg_file.write('\techo "IDMS Linux: Loading kernel..."\n')
            grub_cfg_file.write(f"\tlinux /boot/vmlinuz-{main_kernel} console=tty0 console=ttyS0,115200{extra_cmdline}\n")
            grub_cfg_file.write('\techo "IDMS Linux: Loading initrd images..."\n')
            grub_cfg_file.write(f"\tinitrd {initrd_extra}/boot/initramfs-{main_kernel}.img\n")
            grub_cfg_file.write('\techo "IDMS Linux: Booting..."\n')
            grub_cfg_file.write("\tsleep 1\n")
            grub_cfg_file.write("}\n")
            grub_cfg_file.write("\n")
            grub_cfg_file.write("\n")

            # Autoprovision menu
            grub_cfg_file.write('submenu ">> Autoprovision" {\n')
            grub_cfg_file.write("\n")
            # Non-serial entry
            grub_cfg_file.write(f'\tmenuentry "Autoprovision IDMS Linux - 169.254.1.1 [{kernel_versions[main_kernel]}]" {{\n')
            grub_cfg_file.write("\t\tset gfxpayload=keep\n")
            grub_cfg_file.write("\t\techo\n")
            grub_cfg_file.write('\t\techo "IDMS Linux: Loading kernel..."\n')
            grub_cfg_file.write(f"\t\tlinux /boot/vmlinuz-{main_kernel} autorun=http://169.254.1.1/autoprovision\n")
            grub_cfg_file.write('\t\techo "IDMS Linux: Loading initrd images..."\n')
            grub_cfg_file.write(f"\t\tinitrd {initrd_extra}/boot/initramfs-{main_kernel}.img\n")
            grub_cfg_file.write('\techo "IDMS Linux: Booting..."\n')
            grub_cfg_file.write("\tsleep 1\n")
            grub_cfg_file.write("\t}\n")
            grub_cfg_file.write("\n")
            # Serial entry
            grub_cfg_file.write(
                f'\tmenuentry "Autoprovision IDMS Linux (serial) - 169.254.1.1 [{kernel_versions[main_kernel]}]" {{\n'
            )
            grub_cfg_file.write("\t\tset gfxpayload=text\n")
            grub_cfg_file.write("\t\techo\n")
            grub_cfg_file.write('\t\techo "IDMS Linux: Loading kernel..."\n')
            grub_cfg_file.write(
                f"\t\tlinux /boot/vmlinuz-{main_kernel} console=tty0 console=ttyS0,115200 "
                "autorun=http://169.254.1.1/autoprovision\n"
            )
            grub_cfg_file.write('\t\techo "IDMS Linux: Loading initrd images..."\n')
            grub_cfg_file.write(f"\t\tinitrd {initrd_extra}/boot/initramfs-{main_kernel}.img\n")
            grub_cfg_file.write('\techo "IDMS Linux: Booting..."\n')
            grub_cfg_file.write("\tsleep 1\n")
            grub_cfg_file.write("\t}\n")
            grub_cfg_file.write("\n")
            grub_cfg_file.write("}\n")
            grub_cfg_file.write("\n")
            grub_cfg_file.write("\n")

            # Tools menu
            grub_cfg_file.write('submenu ">> Tools" {\n')
            grub_cfg_file.write("\n")
            # Memtest86+ BIOS
            grub_cfg_file.write('\tmenuentry "Boot Memtest86+ - BIOS" {\n')
            grub_cfg_file.write("\t\tset gfxpayload=keep\n")
            grub_cfg_file.write("\t\techo\n")
            grub_cfg_file.write('\t\techo "IDMS Linux: Loading memtest86+ (BIOS)..."\n')
            grub_cfg_file.write("\t\tlinux /boot/memtest86+/memtest.bin\n")
            grub_cfg_file.write('\techo "IDMS Linux: Starting memtest86+..."\n')
            grub_cfg_file.write("\tsleep 1\n")
            grub_cfg_file.write("\t}\n")
            grub_cfg_file.write("\n")
            # Memtest86+ BIOS
            grub_cfg_file.write('\tmenuentry "Boot Memtest86+ - BIOS (serial)" {\n')
            grub_cfg_file.write("\t\tset gfxpayload=keep\n")
            grub_cfg_file.write("\t\techo\n")
            grub_cfg_file.write('\t\techo "IDMS Linux: Loading memtest86+ (BIOS)..."\n')
            grub_cfg_file.write("\t\tlinux /boot/memtest86+/memtest.bin console=ttyS0,115200n8\n")
            grub_cfg_file.write('\techo "IDMS Linux: Starting memtest86+..."\n')
            grub_cfg_file.write("\tsleep 1\n")
            grub_cfg_file.write("\t}\n")
            grub_cfg_file.write("\n")

            # START - EFI stuff
            grub_cfg_file.write('\tif [ "${grub_platform}" == "efi" ]; then\n')

            # Memtest86+ EFI
            grub_cfg_file.write('\t\tmenuentry "Boot Memtest86+ - EFI" {\n')
            grub_cfg_file.write("\t\t\tinsmod chain\n")
            grub_cfg_file.write("\t\t\tset gfxpayload=keep\n")
            grub_cfg_file.write("\t\t\techo\n")
            grub_cfg_file.write('\t\t\techo "IDMS Linux: Loading memtest86+ (EFI)..."\n')
            grub_cfg_file.write("\t\t\tchainloader /boot/memtest86+/memtest.efi\n")
            grub_cfg_file.write('\techo "IDMS Linux: Starting memtest86+..."\n')
            grub_cfg_file.write("\tsleep 1\n")
            grub_cfg_file.write("\t\t}\n")
            grub_cfg_file.write("\n")
            # EFI shell
            grub_cfg_file.write('\t\tmenuentry "UEFI Shell" {\n')
            grub_cfg_file.write("\t\t\tinsmod chain\n")
            grub_cfg_file.write("\t\t\tset gfxpayload=text\n")
            grub_cfg_file.write("\t\t\techo\n")
            grub_cfg_file.write('\t\t\techo "IDMS Linux: Loading EFI shell..."\n')
            grub_cfg_file.write("\t\t\tchainloader /SHELLX64.EFI\n")
            grub_cfg_file.write('\techo "IDMS Linux: Starting EFI shell..."\n')
            grub_cfg_file.write("\tsleep 1\n")
            grub_cfg_file.write("\t\t}\n")
            grub_cfg_file.write("\n")

            # END - EFI stuff
            grub_cfg_file.write("\tfi\n")
            grub_cfg_file.write("\n")
            grub_cfg_file.write("}\n")
            grub_cfg_file.write("\n")
            grub_cfg_file.write("\n")

            # START - Extra kernels

            # Check if we have extra kernels
            if extra_kernels:
                # Alternative kernels menu
                grub_cfg_file.write('submenu ">> Additional Kernels" {\n')
                grub_cfg_file.write("\n")
                # Loop with extra kernels to add
                for kernel in extra_kernels:
                    grub_cfg_file.write(f'\tsubmenu ">> Kernel: {kernel}" {{\n')
                    # Main live entry
                    grub_cfg_file.write(f'\t\tmenuentry "Boot IDMS Linux live [{kernel_versions[kernel]}]" {{\n')
                    grub_cfg_file.write("\t\t\tset gfxpayload=keep\n")
                    grub_cfg_file.write("\t\t\techo\n")
                    grub_cfg_file.write(f'\t\t\techo "IDMS Linux: Loading kernel [{kernel}]..."\n')
                    grub_cfg_file.write(f"\t\t\tlinux /boot/vmlinuz-{kernel}{extra_cmdline}\n")
                    grub_cfg_file.write(f'\t\t\techo "IDMS Linux: Loading initrd images [{kernel}]..."\n')
                    grub_cfg_file.write(f"\t\t\tinitrd {initrd_extra}/boot/initramfs-{kernel}.img\n")
                    grub_cfg_file.write(f'\t\t\techo "IDMS Linux: Booting [{kernel}]..."\n')
                    grub_cfg_file.write("\t\t\tsleep 1\n")
                    grub_cfg_file.write("\t\t}\n")
                    grub_cfg_file.write("\n")
                    # Serial entry
                    grub_cfg_file.write(f'\t\tmenuentry "Boot IDMS Linux live (serial) [{kernel_versions[kernel]}]" {{\n')
                    grub_cfg_file.write("\t\t\tset gfxpayload=text\n")
                    grub_cfg_file.write("\t\t\techo\n")
                    grub_cfg_file.write(f'\t\t\techo "IDMS Linux: Loading kernel [{kernel}]..."\n')
                    grub_cfg_file.write(f"\t\t\tlinux /boot/vmlinuz-{kernel} console=tty0 console=ttyS0,115200{extra_cmdline}\n")
                    grub_cfg_file.write(f'\t\t\techo "IDMS Linux: Loading initrd images [{kernel}]..."\n')
                    grub_cfg_file.write(f"\t\t\tinitrd {initrd_extra}/boot/initramfs-{kernel}.img\n")
                    grub_cfg_file.write(f'\t\t\techo "IDMS Linux: Booting [{kernel}]..."\n')
                    grub_cfg_file.write("\t\t\tsleep 1\n")
                    grub_cfg_file.write("\t\t}\n")
                    grub_cfg_file.write("\n")
                    # Autoprovision menu
                    grub_cfg_file.write('\t\tsubmenu ">> Autoprovision" {\n')
                    grub_cfg_file.write("\n")
                    # Non-serial entry
                    grub_cfg_file.write(
                        f'\t\t\tmenuentry "Autoprovision IDMS Linux - 169.254.1.1 [{kernel_versions[kernel]}]" {{\n'
                    )
                    grub_cfg_file.write("\t\t\t\tset gfxpayload=keep\n")
                    grub_cfg_file.write("\t\t\t\techo\n")
                    grub_cfg_file.write(f'\t\t\t\techo "IDMS Linux: Loading kernel [{kernel}]..."\n')
                    grub_cfg_file.write(f"\t\t\t\tlinux /boot/vmlinuz-{kernel} autorun=http://169.254.1.1/autoprovision\n")
                    grub_cfg_file.write(f'\t\t\t\techo "IDMS Linux: Loading initrd images [{kernel}]..."\n')
                    grub_cfg_file.write(f"\t\t\t\tinitrd {initrd_extra}/boot/initramfs-{kernel}.img\n")
                    grub_cfg_file.write(f'\t\t\techo "IDMS Linux: Booting [{kernel}]..."\n')
                    grub_cfg_file.write("\t\t\tsleep 1\n")
                    grub_cfg_file.write("\t\t\t}\n")
                    grub_cfg_file.write("\n")
                    # Serial entry
                    grub_cfg_file.write(
                        f'\t\t\tmenuentry "Autoprovision IDMS Linux (serial) - 169.254.1.1 [{kernel_versions[kernel]}]" ' "{\n"
                    )
                    grub_cfg_file.write("\t\t\t\tset gfxpayload=text\n")
                    grub_cfg_file.write("\t\t\t\techo\n")
                    grub_cfg_file.write(f'\t\t\t\techo "IDMS Linux: Loading kernel [{kernel}]..."\n')
                    grub_cfg_file.write(
                        f"\t\t\t\tlinux /boot/vmlinuz-{kernel} console=tty0 console=ttyS0,115200 "
                        "autorun=http://169.254.1.1/autoprovision\n"
                    )
                    grub_cfg_file.write(f'\t\t\t\techo "IDMS Linux: Loading initrd images [{kernel}]..."\n')
                    grub_cfg_file.write(f"\t\t\t\tinitrd {initrd_extra}/boot/initramfs-{kernel}.img\n")
                    grub_cfg_file.write(f'\t\t\techo "IDMS Linux: Booting [{kernel}]..."\n')
                    grub_cfg_file.write("\t\t\tsleep 1\n")
                    grub_cfg_file.write("\t\t\t}\n")
                    grub_cfg_file.write("\n")
                    grub_cfg_file.write("\t\t}\n")
                    grub_cfg_file.write("\n")
                    grub_cfg_file.write("\t}\n")
                    grub_cfg_file.write("\n")
                    grub_cfg_file.write("\n")
                grub_cfg_file.write("}\n")
                grub_cfg_file.write("\n")
                grub_cfg_file.write("\n")

            # END - Extra kernels

            # Shutdown entry
            grub_cfg_file.write('menuentry "Shutdown" {\n')
            grub_cfg_file.write("\tinsmod halt\n")
            grub_cfg_file.write("\techo\n")
            grub_cfg_file.write('\techo "IDMS Linux: Shutting down system..."\n')
            grub_cfg_file.write("\tsleep 1\n")
            grub_cfg_file.write("\thalt\n")
            grub_cfg_file.write("}\n")
            grub_cfg_file.write("\n")

            # Reboot entry
            grub_cfg_file.write('menuentry "Reboot" {\n')
            grub_cfg_file.write("\tinsmod reboot\n")
            grub_cfg_file.write("\techo\n")
            grub_cfg_file.write('\techo "IDMS Linux: Rebooting system..."\n')
            grub_cfg_file.write("\tsleep 1\n")
            grub_cfg_file.write("\treboot\n")
            grub_cfg_file.write("}\n")
            grub_cfg_file.write("\n")

        # Create Grub fonts dir
        isofs_grub_fonts_path = isofs_grub_path.concatpath("fonts")
        try:
            isofs_grub_fonts_path.mkdir(mode=0o0755, parents=True, exist_ok=True)
        except OSError as err:
            raise IliRuntimeError(f"Failed to create boot/Grub directory '{isofs_grub_fonts_path}': {err}") from None

        # Copy over Grub font
        rootfs_grub_font = self.rootfs.mount_point.concatpath("usr", "share", "grub", "unicode.pf2")
        isofs_grub_font = isofs_grub_fonts_path.concatpath("unicode.pf2")
        try:
            shutil.copy(rootfs_grub_font, isofs_grub_font)
        except OSError as err:
            raise IliRuntimeError(f"Failed to copy Grub font '{rootfs_grub_font}' to '{isofs_grub_font}': {err}") from None

        # Copy EFI shell over
        rootfs_efi_img_shellx64 = self.rootfs.mount_point.concatpath("usr", "share", "edk2-shell", "x64", "Shell_Full.efi")
        isofs_efi_img_shellx64 = self.isofs_path.concatpath("SHELLX64.EFI")
        try:
            shutil.copy(rootfs_efi_img_shellx64, isofs_efi_img_shellx64)
        except OSError as err:
            raise IliRuntimeError(
                f"Failed to copy EFI shell '{rootfs_efi_img_shellx64}' to '{isofs_efi_img_shellx64}': {err}"
            ) from None

        # Create memtest86+ dir
        isofs_memtest86_path = self.isofs_path.concatpath("boot", "memtest86+")
        try:
            isofs_memtest86_path.mkdir(mode=0o0755, parents=True, exist_ok=True)
        except OSError as err:
            raise IliRuntimeError(f"Failed to create memtest86+ directory '{isofs_memtest86_path}': {err}") from None
        # Copy memtest86+ bin over
        rootfs_memtest86 = self.rootfs.mount_point.concatpath("boot", "memtest86+", "memtest.bin")
        isofs_memtest86 = isofs_memtest86_path.concatpath("memtest.bin")
        try:
            shutil.copy(rootfs_memtest86, isofs_memtest86)
        except OSError as err:
            raise IliRuntimeError(f"Failed to copy memtest86+ '{rootfs_memtest86}' to '{isofs_memtest86}': {err}") from None
        # Copy memtest86+ EFI over
        rootfs_memtest86_efi = self.rootfs.mount_point.concatpath("boot", "memtest86+", "memtest.efi")
        isofs_memtest86_efi = isofs_memtest86_path.concatpath("memtest.efi")
        try:
            shutil.copy(rootfs_memtest86_efi, isofs_memtest86_efi)
        except OSError as err:
            raise IliRuntimeError(f"Failed to copy memtest86+ '{rootfs_memtest86_efi}' to '{isofs_memtest86_efi}': {err}") from None

    def setup_target_grub_eltorito(self) -> None:
        """Set up Grub Eltorito image."""

        if self.isofs_path is None:
            raise IliRuntimeError("isofs_path must be set")
        if self.rootfs is None:
            raise IliRuntimeError("rootfs must be set")
        if self.rootfs.mount_point is None:
            raise IliRuntimeError("rootfs mount_point must be set")

        # Create Grub dir
        isofs_grub_i386_path = self.isofs_path.concatpath("boot", "grub", "i386-pc")
        try:
            isofs_grub_i386_path.mkdir(mode=0o0755, parents=True, exist_ok=True)
        except OSError as err:
            raise IliRuntimeError(f"Failed to create Grub Eltorito directory '{isofs_grub_i386_path}': {err}") from None

        # Create eltorito image
        # We build this within the rootfs, then move it to the isofs
        self.run_target(
            [
                "/usr/bin/grub-mkimage",
                "--prefix=/boot/grub",
                "--format=i386-pc-eltorito",
                "--output=/eltorito.img",
                "biosdisk",
                "iso9660",
            ]
        )

        # Move image to the ISO
        rootfs_eltorito = self.rootfs.mount_point.concatpath("eltorito.img")
        isofs_eltorito = isofs_grub_i386_path.concatpath("eltorito.img")
        try:
            shutil.copy(rootfs_eltorito, isofs_eltorito)
        except OSError as err:
            raise IliRuntimeError(f"Failed to copy Grub image '{rootfs_eltorito}' to '{isofs_eltorito}': {err}") from None
        try:
            rootfs_eltorito.unlink()
        except OSError as err:
            raise IliRuntimeError(f"Failed to remove temporary Grub image '{rootfs_eltorito}': {err}") from None

        # Copy Grub modules we need
        grub_mods_path = self.rootfs.mount_point.concatpath("usr", "lib", "grub", "i386-pc")
        for grub_filename in grub_mods_path.glob("*.mod"):
            shutil.copy(grub_filename, isofs_grub_i386_path)

    def setup_target_grub_efi(self) -> None:
        """Set up Grub EFI image."""

        if self.isofs_path is None:
            raise IliRuntimeError("isofs_path must be set")
        if self.rootfs is None:
            raise IliRuntimeError("rootfs must be set")
        if self.rootfs.mount_point is None:
            raise IliRuntimeError("rootfs mount_point must be set")

        # Create Grub dir
        isofs_grub_x64_path = self.isofs_path.concatpath("boot", "grub", "x86_64-efi")
        try:
            isofs_grub_x64_path.mkdir(mode=0o0755, parents=True, exist_ok=True)
        except OSError as err:
            raise IliRuntimeError(f"Failed to create Grub Eltorito directory '{isofs_grub_x64_path}': {err}") from None

        # Create EFI dir
        isofs_efiboot_path = self.isofs_path.concatpath("EFI", "BOOT")
        try:
            isofs_efiboot_path.mkdir(mode=0o0755, parents=True, exist_ok=True)
        except OSError as err:
            raise IliRuntimeError(f"Failed to create EFI/BOOT directory '{isofs_efiboot_path}': {err}") from None

        # Create EFI image
        # We build this within the rootfs, then move it to the isofs
        self.run_target(
            [
                "/usr/bin/grub-mkimage",
                "--prefix=/boot/grub",
                "--format=x86_64-efi",
                "--sbat=/usr/share/grub/sbat.csv",
                "--disable-shim-lock",
                "--output=/BOOTX64.EFI",
                "iso9660",
            ]
        )

        # Move image to the ISO
        rootfs_bootx64 = self.rootfs.mount_point.concatpath("BOOTX64.EFI")
        isofs_bootx64 = isofs_efiboot_path.concatpath("BOOTX64.EFI")
        try:
            shutil.copy(rootfs_bootx64, isofs_efiboot_path)
        except OSError as err:
            raise IliRuntimeError(f"Failed to move Grub image '{rootfs_bootx64}' to '{isofs_bootx64}': {err}") from None
        try:
            rootfs_bootx64.unlink()
        except OSError as err:
            raise IliRuntimeError(f"Failed to remove temporary Grub image '{rootfs_bootx64}': {err}") from None

        # Copy Grub modules we need
        grub_mods_path = self.rootfs.mount_point.concatpath("usr", "lib", "grub", "x86_64-efi")
        for grub_filename in grub_mods_path.glob("*.mod"):
            shutil.copy(grub_filename, isofs_grub_x64_path)

    def setup_target_grub_efi_img(self) -> None:
        """Set up Grub efi.img image."""

        if self.build_dir is None:
            raise IliRuntimeError("build_dir must be set")
        if self.rootfs is None:
            raise IliRuntimeError("rootfs must be set")
        if self.rootfs.mount_point is None:
            raise IliRuntimeError("rootfs mount_point must be set")

        # Create Grub EFI mount point
        efifs_path = self.build_dir.concatpath("efifs")
        try:
            efifs_path.mkdir(mode=0o0755, parents=True, exist_ok=True)
        except OSError as err:
            raise IliRuntimeError(f"Failed to create Grub EFI mount point directory '{efifs_path}': {err}") from None

        # Create image file of 64M, which is the minimum we can use
        efi_img = self.build_dir.concatpath("efi.img")
        util.run_host(["/usr/bin/truncate", "--size", "64M", str(efi_img)])

        # Setup loop device
        loop_device = self.add_loop_device(efi_img)

        # Setup filesystem
        imgfs = IliFilesystem(
            source=loop_device,
            fstype="vfat",
            mount_root=efifs_path,
            mount_path=IliPath("/"),
            options=["-n", "IDMSISO_EFI", "-F", "32"],
        )
        # Make filesystem
        imgfs.mkfs()
        # Mount filesystem
        self.mount(imgfs)

        # Create EFI image
        self.run_target(
            [
                "/usr/bin/grub-mkimage",
                "--prefix=/boot/grub",
                "--format=x86_64-efi",
                "--sbat=/usr/share/grub/sbat.csv",
                "--disable-shim-lock",
                "--output=/BOOTX64.EFI",
                "iso9660",
            ]
        )

        # Create EFI boot directory
        efifs_efiboot_path = efifs_path.concatpath("EFI", "BOOT")
        try:
            efifs_efiboot_path.mkdir(mode=0o0755, parents=True, exist_ok=True)
        except OSError as err:
            raise IliRuntimeError(f"Failed to create Grub EFI efi.img directory '{efifs_efiboot_path}': {err}") from None

        # Copy Grub EFI over
        rootfs_efi_img_bootx64 = self.rootfs.mount_point.concatpath("BOOTX64.EFI")
        efifs_efi_img_bootx64 = efifs_efiboot_path.concatpath("BOOTX64.EFI")
        try:
            shutil.copy(rootfs_efi_img_bootx64, efifs_efi_img_bootx64)
        except OSError as err:
            raise IliRuntimeError(
                f"Failed to copy Grub EFI '{rootfs_efi_img_bootx64}' to '{efifs_efi_img_bootx64}': {err}"
            ) from None
        try:
            rootfs_efi_img_bootx64.unlink()
        except OSError as err:
            raise IliRuntimeError(f"Failed to copy Grub EFI '{rootfs_efi_img_bootx64}': {err}") from None

        # Copy EFI shell over
        rootfs_efi_img_shellx64 = self.rootfs.mount_point.concatpath("usr", "share", "edk2-shell", "x64", "Shell_Full.efi")
        efifs_efi_img_shellx64 = efifs_efiboot_path.concatpath("SHELLX64.EFI")
        try:
            shutil.copy(rootfs_efi_img_shellx64, efifs_efi_img_shellx64)
        except OSError as err:
            raise IliRuntimeError(
                f"Failed to copy EFI shell '{rootfs_efi_img_shellx64}' to '{efifs_efi_img_shellx64}': {err}"
            ) from None

        # Unmount and remove loop device
        self.unmount(imgfs)
        self.remove_loop_device(loop_device)

    def finalize_target(self) -> None:
        """
        Finalize target.

        Here we create the final tar archive. This is after
        """

        if self.build_dir is None:
            raise IliRuntimeError("build_dir must be set")
        if self.rootfs is None:
            raise IliRuntimeError("rootfs must be set")
        if self.rootfs.mount_point is None:
            raise IliRuntimeError("rootfs mount_point must be set")
        if self.isofs_rootfs_img is None:
            raise IliRuntimeError("isofs_rootfs_img must be set")
        if self.target_file is None:
            raise IliRuntimeError("target_file must be set")
        if self.isofs_path is None:
            raise IliRuntimeError("isofs_path must be set")

        logging.debug("  - Creating squashfs rootfs")

        util.run_host(
            [
                "/usr/bin/nice",
                "--adjustment=19",
                "/usr/bin/mksquashfs",
                str(self.rootfs.mount_point),
                str(self.isofs_rootfs_img),
                "-noappend",
                "-comp",
                "zstd",
                "-Xcompression-level",
                f"{self.target_compression_level}",
            ]
        )

        logging.debug("  - Creating final target '%s'", self.target_file)

        xorriso_args = []

        # Add Xorriso args
        xorriso_args.extend(["/usr/bin/xorriso", "-as", "mkisofs"])
        xorriso_args.extend(["-iso-level", "3"])
        xorriso_args.append("-full-iso9660-filenames")
        xorriso_args.extend(["-joliet", "-joliet-long"])
        xorriso_args.append("-rational-rock")
        xorriso_args.extend(["-volid", self.isofs_img_volid])
        xorriso_args.extend(["-appid", self.isofs_img_appid])
        xorriso_args.extend(["-publisher", self.isofs_img_publisher])
        xorriso_args.extend(["-preparer", self.isofs_img_preparer])
        xorriso_args.append(f"--modification-date={self.isofs_img_mdate}")

        # BIOS MBR
        grub_mbr_path = self.rootfs.mount_point.concatpath("usr", "lib", "grub", "i386-pc", "boot_hybrid.img")
        xorriso_args.append("--grub2-boot-info")
        xorriso_args.extend(["--grub2-mbr", str(grub_mbr_path)])
        # - When GPT is used, create an additional partition in the MBR (besides 0xEE) for sectors 0 to 1 (MBR
        #   bootstrap code area) and mark it as bootable
        xorriso_args.append("--mbr-force-bootable")
        # - Move partition away from start of ISO otherwise the GPT will not be valid and ISO 9660 partition will not be
        #   mountable
        xorriso_args.extend(["-partition_offset", "16"])

        # BIOS Eltorito
        # - El Torito boot image for x86 BIOS
        grub_eltorito_path = IliPath("/").concatpath("boot", "grub", "i386-pc", "eltorito.img")
        xorriso_args.extend(["-eltorito-boot", str(grub_eltorito_path)])
        # - Required to boot
        xorriso_args.append("-no-emul-boot")
        xorriso_args.extend(["-boot-load-size", "4"])
        xorriso_args.append("-boot-info-table")

        # EFI ESP
        # - Attach efiboot.img as a second partition and set its partition type to "EFI system partition"
        efi_path = self.build_dir.concatpath("efi.img")
        xorriso_args.extend(["-append_partition", "2", "C12A7328-F81F-11D2-BA4B-00A0C93EC93B", str(efi_path)])
        xorriso_args.append("-isohybrid-gpt-basdat")

        # EFI Eltorito
        # - Start a new El Torito boot entry for UEFI
        xorriso_args.append("-eltorito-alt-boot")
        # - Set the second partition as the El Torito UEFI boot image
        xorriso_args.extend(["-e", "--interval:appended_partition_2:all::"])
        # - Boot image is not emulating floppy or hard disk; required for all known boot loaders
        xorriso_args.append("-no-emul-boot")
        xorriso_args.append("-isohybrid-gpt-basdat")

        # Xorriso args
        xorriso_args.extend(["-output", str(self.target_file)])
        xorriso_args.append(str(self.isofs_path))

        # Build ISO
        util.run_host(xorriso_args)

        logging.info("  - Target built: %s", self.target_file)

    @property
    def build_dir(self) -> IliPath | None:
        """Return the build dir we're going to use."""
        return self._build_dir

    @build_dir.setter
    def build_dir(self, build_dir: IliPath) -> None:
        """Set the build dir to use."""
        self._build_dir = build_dir

    @property
    def target_file(self) -> IliPath | None:
        """Return target file we're going to use."""
        return self._target_file

    @target_file.setter
    def target_file(self, target_file: IliPath) -> None:
        """Set the target file to use."""
        self._target_file = target_file

    @property
    def target_kernels(self) -> list[str]:
        """Return kernels we're going to use."""
        return self._target_kernel

    @target_kernels.setter
    def target_kernels(self, target_kernel: list[str]) -> None:
        """Set the kernels to use."""
        self._target_kernel = target_kernel

    @property
    def target_autorun(self) -> str | None:
        """Return the autorun script we're going to use."""
        return self._target_autorun

    @target_autorun.setter
    def target_autorun(self, target_autorun: str) -> None:
        """Set the autorun script to use."""
        self._target_autorun = target_autorun

    @property
    def target_ilitools_profile(self) -> str | None:
        """Return the ilitools_profile script we're going to use."""
        return self._target_ilitools_profile

    @target_ilitools_profile.setter
    def target_ilitools_profile(self, target_ilitools_profile: str) -> None:
        """Set the ilitools_profile script to use."""
        self._target_ilitools_profile = target_ilitools_profile

    @property
    def target_compression_level(self) -> int:
        """Return the compression level we're going to use."""
        return self._target_compression_level

    @target_compression_level.setter
    def target_compression_level(self, target_compression_level: int) -> None:
        """Set the compression level to use."""
        self._target_compression_level = target_compression_level

    @property
    def isofs_path(self) -> IliPath | None:
        """Return the image isofs dir we're going to use."""
        return self._isofs_path

    @isofs_path.setter
    def isofs_path(self, isofs_path: IliPath) -> None:
        """Set the image isofs dir to use."""
        self._isofs_path = isofs_path

    @property
    def isofs_rootfs_img(self) -> IliPath | None:
        """Return the isofs image path we're going to use."""
        return self._isofs_rootfs_img

    @isofs_rootfs_img.setter
    def isofs_rootfs_img(self, isofs_rootfs_img: IliPath) -> None:
        """Set the isofs image path to use."""
        self._isofs_rootfs_img = isofs_rootfs_img

    @property
    def isofs_img_mdate(self) -> str:
        """Return the isofs modification date."""
        return self._isofs_img_mdate

    @property
    def isofs_img_uuid(self) -> str:
        """Return the isofs UUID."""
        return self._isofs_img_uuid

    @property
    def isofs_img_volid(self) -> str:
        """Return the isofs volume ID."""
        return self._isofs_img_volid

    @property
    def isofs_img_appid(self) -> str:
        """Return the isofs application ID."""
        return self._isofs_img_appid

    @property
    def isofs_img_publisher(self) -> str:
        """Return the isofs publisher."""
        return self._isofs_img_publisher

    @property
    def isofs_img_preparer(self) -> str:
        """Return the isofs preparer."""
        return self._isofs_img_preparer

    @property
    def target_mkinitcpio_hooks(self) -> list[str]:
        """Return the list of mkinitcpio hooks we're adding to the initramfs."""
        return self._target_mkinitcpio_hooks

    @property
    def target_mkinitcpio_modules(self) -> list[str]:
        """Return the kernel modules we're adding to the initramfs."""
        return self._target_mkinitcpio_modules
