#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# Copyright (c) 2019-2024, AllWorldIT
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Entry point into the IDMS Linux installer VM functionality from the commandline."""

import argparse
import logging
import sys
from typing import Any

from ..commandline import setup_logging
from ..exceptions import IliRuntimeError, IliUsageError
from ..filesystem import IliPath
from ..version import __version__
from .vm import IliVM

__all__ = ["run_ilivm", "run_ilivm_entrypoint"]


def get_argparser(prog: str | None) -> argparse.ArgumentParser:
    """Return an instance of argparse.ArgumentParser."""

    print(f"IDMS Linux Installer VM v{__version__} - Copyright © 2019-2024, AllWorldIT.\n", file=sys.stderr)  # noqa: T201

    argparser = argparse.ArgumentParser(prog=prog, add_help=False)

    optional_group = argparser.add_argument_group("Optional arguments")
    optional_group.add_argument("-h", "--help", action="help", help="Show this help message and exit")
    optional_group.add_argument(
        "-w", "--work-dir", nargs=1, default=None, help="Working directory, by default TMPDIR or /tmp is used"
    )
    optional_group.add_argument(
        "--wait-poweroff",
        nargs="?",
        default=None,
        type=int,
        help="Do not stop the VM after running the scripts and wait X seconds for poweroff, by default the VM is terminated. "
        "0 = no limit.",
    )
    optional_group.add_argument("-q", "--quiet", action="store_true", help="Be more quiet")

    return argparser


def run_ilivm(argv: list[str], prog: str = "ilivm") -> None:  # noqa: C901,PLR0912,PLR0915
    """Entry point for execution from the commandline."""

    our_dir = IliPath(__file__).parent

    # Grab our arg parser
    argparser = get_argparser(prog)

    argparser.add_argument("--live-image", nargs=1, help="Path to the live image to use")
    argparser.add_argument("--listen-port", nargs=1, type=int, help="Port to listen on for SSH")
    argparser.add_argument(
        "--copy-to-vm", action="append", help="File/paths to copy from the HOST to the VM in the format of SOURCE:DEST"
    )
    argparser.add_argument(
        "--copy-from-vm", action="append", help="File/paths to copy from the VM to the HOST in the format of SOURCE:DEST"
    )

    argparser.add_argument(
        "--add-disk",
        action="append",
        help="Add a disk to the VM, either a path or an integer size in GB. "
        "Also supports adding ',model=MODEL' and ',serial=SERIAL' to the end of the disk string",
    )

    argparser.add_argument(
        "--kernel",
        nargs=1,
        choices=["linux", "linux-lts"],
        default=["linux-lts"],
        help="Kernel to use, 'linux' or 'linux-lts'. Default is 'linux'.",
    )
    argparser.add_argument(
        "--extra-kernel-args",
        nargs=1,
        help="Extra kernel arguments to pass to the kernel",
    )

    argparser.add_argument("--run-script", action="append", help="Scripts to run")

    # If no arguments were specified, display help
    if len(sys.argv) == 1:
        raise IliUsageError(argparser.format_help())

    # Parse args
    args = argparser.parse_args(argv)
    if args.quiet:
        setup_logging()
    else:
        setup_logging(logging.DEBUG)

    # Work out arguments to pass to the VM class
    vm_kwargs: dict[str, Any] = {}
    # Live image to boot off
    if args.live_image:
        vm_kwargs["live_image"] = IliPath(args.live_image[0]).expanduser()
    # Base working directory
    if args.work_dir:
        vm_kwargs["work_dir"] = IliPath(args.work_dir[0]).expanduser()
    # Listen port for SSH
    if args.listen_port:
        vm_kwargs["listen_port"] = args.listen_port[0]

    # Create VM object
    vm = IliVM(**vm_kwargs)

    # Work out files to copy to the VM
    copy_to_vm: list[tuple[IliPath, IliPath]] = []
    if args.copy_to_vm:
        # Loop and add each pair
        for copy_file in args.copy_to_vm:
            source, dest = copy_file.split(":")
            if not dest:
                raise ValueError(f"Invalid format for --add-to-vm (SOURCE:DEST): {copy_file}")
            source_full = IliPath(source).expanduser()
            if not source_full.exists():
                raise IliRuntimeError(f"Cannot find path to copy to VM: {source_full}")
            copy_to_vm.append((IliPath(source).expanduser(), IliPath(dest)))
    # Work out files to copy from the VM
    copy_fom_vm: list[tuple[IliPath, IliPath]] = []
    if args.copy_from_vm:
        # Loop and add each pair
        for copy_file in args.copy_from_vm:
            source, dest = copy_file.split(":")
            if not dest:
                raise ValueError(f"Invalid format for --get-from-vm (SOURCE:DEST): {copy_file}")
            copy_fom_vm.append((IliPath(source), IliPath(dest).expanduser()))

    # Load scripts to run
    scripts: list[IliPath] = []
    for script in args.run_script:
        script_path = IliPath(script).expanduser()
        if not script_path.exists():
            raise FileNotFoundError(f"Script not found: {script_path}")
        scripts.append(script_path)

    # Prepare VM
    vm.add_oem_file(
        our_dir.concatpath("data", "oem", "grub.cfg"),
        replaces={"@KERNEL@": args.kernel[0], "@EXTRA_KERNEL_ARGS@": args.extra_kernel_args[0] if args.extra_kernel_args else ""},
    )
    vm.add_oem_file(our_dir.concatpath("data", "oem", "autorun"))

    # Add disks
    if args.add_disk:
        # Loop with disk options, split off optional serial and model and add to VM
        for disk in args.add_disk:
            # Split disk components
            disk_components = disk.split(",")
            # Start with nothing
            disk_path = None
            disk_size = None
            disk_product = None
            disk_model = None
            disk_serial = None
            # Work out if its a size specification
            if disk_components[0].isdigit():
                # Size in GB
                disk_size = int(disk_components[0])
            else:
                disk_path = IliPath(disk_components[0]).expanduser()
            # If not a size specification, its a path specification
            for component in disk_components[1:]:
                if component.startswith("product="):
                    disk_product = component.split("=", 1)[1]
                elif component.startswith("model="):
                    disk_model = component.split("=", 1)[1]
                elif component.startswith("serial="):
                    disk_serial = component.split("=", 1)[1]
                else:
                    raise IliRuntimeError(f"Invalid disk option: {component}")
            # Add disk to VM
            vm.add_disk(path=disk_path, size=disk_size, product=disk_product, model=disk_model, serial=disk_serial)

    # Prepare VM
    vm.prepare()

    # Start VM
    vm.start()

    # Try copy files we need to the VM before running scripts
    for source, dest in copy_to_vm:
        try:
            vm.copy_to_vm(source, dest)
        except IliRuntimeError as exc:
            logging.error("[HOST]   - Failed to copy file '%s' to VM", source)  # noqa: TRY400
            logging.error("[HOST]     - %s", exc)  # noqa: TRY400
            if exc.data:
                logging.error("[HOST]     - Error data: %s", exc.data)  # noqa: TRY400
            break

    # Deferred exception to raise later
    exception: Exception | None = None

    # Run scripts
    for script in scripts:
        try:
            vm.run_script(script)
        except IliRuntimeError as exc:
            logging.error("[HOST]   - Failed to run script '%s'", script)  # noqa: TRY400
            logging.error("[HOST]     - %s", exc)  # noqa: TRY400
            if exc.data:
                logging.error("[HOST]     - Error data: %s", exc.data)  # noqa: TRY400
            exception = exc
            break

    if exception:
        logging.error("[HOST] One or more scripts failed to run")
    else:
        logging.info("[HOST] All scripts ran successfully")
        # Check if we need to copy files from VM
        if copy_fom_vm:
            logging.info("[HOST] Copying files from VM")
        for source, dest in copy_fom_vm:
            try:
                vm.copy_from_vm(source, dest)
            except IliRuntimeError as exc:
                logging.error("[HOST]   - Failed to copy file '%s' from VM", source)  # noqa: TRY400
                logging.error("[HOST]     - %s", exc)  # noqa: TRY400
                if exc.data:
                    logging.error("[HOST]     - Error data: %s", exc.data)  # noqa: TRY400
                exception = exc
                break

    # Stop VM
    if args.wait_poweroff is None:
        vm.stop()
    else:
        poweroff_timeout = None if args.wait_poweroff == 0 else args.wait_poweroff
        vm.poweroff()
        vm.wait_for_vm_poweroff(poweroff_timeout)

    # Raise exception if it was deferred
    if exception:
        raise exception


def run_ilivm_entrypoint(prog: str = "ilivm") -> None:
    """Entry point for execution from the commandline."""
    try:
        run_ilivm(sys.argv[1:], prog)
    except IliUsageError as exc:
        print(exc.data, file=sys.stderr)  # noqa: T201
    # Catch installer exception
    except IliRuntimeError as exc:
        logging.error("!!!!! ERROR !!!!!")  # noqa: TRY400
        logging.error("Exception encountered: %s", exc)  # noqa: TRY400
        if exc.data:
            logging.error("  - Exception data: %s", exc.data)  # noqa: TRY400
        logging.error("!!!!! ERROR !!!!!")  # noqa: TRY400
        sys.exit(1)
