#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# Copyright (c) 2019-2024, AllWorldIT
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Virtual machine library."""

import json
import logging
import os
import random
import re
import shlex
import shutil
import socket
import string
import subprocess
import tempfile
import threading
import time
from enum import Enum

from ..exceptions import IliRuntimeError
from ..filesystem import IliPath
from ..util import run_host

__all__ = ["IliVM", "IliVMDisk", "IliVMOEMFile", "IliVMState"]


def sanitize_serial_output(input_str: str) -> str:
    """Sanitize serial output."""

    # Step 1: Remove ANSI color codes (escape sequences)
    ansi_escape_re = re.compile(r"\x1B\[[0-?]*[ -/]*[@-~]")
    sanitized_str = ansi_escape_re.sub("", input_str)

    # Step 2: Remove control characters
    control_char_re = re.compile(r"[\x00-\x1F\x7F]")
    sanitized_str = control_char_re.sub("", sanitized_str)

    # Step 3: Replace encoding issues with a placeholder (optional)
    sanitized_str = sanitized_str.encode("utf-8", errors="replace").decode("utf-8")

    # Step 4: Trim leading and trailing whitespace
    return sanitized_str.strip()


class IliVMState(Enum):
    """Virtual machine state enum."""

    STOPPED = 0
    RUNNING = 1
    BOOTING = 2


class IliVMOEMFile:
    """Class representing a file to copy to the OEM media."""

    _path: IliPath
    _replaces: dict[str, str] | None

    def __init__(self, path: IliPath, replaces: dict[str, str] | None = None) -> None:
        """Initialize the OEM file."""
        self._path = path
        self._replaces = replaces

    def copy_to(self, dst: IliPath) -> None:
        """Copy the file to the destination."""

        # If we have replacements, read the file and replace the contents
        if self._replaces:
            with self._path.open(encoding="UTF-8") as file:
                content = file.read()
                for key, value in self._replaces.items():
                    content = content.replace(key, value)
            with dst.open(mode="w", encoding="UTF-8") as file:
                file.write(content)
            return
        # Otherwise, just copy the file
        shutil.copy(self._path, dst)

    @property
    def path(self) -> IliPath:
        """Get the OEM file path."""
        return self._path


class IliVMDisk:
    """Virtual machine disk class."""

    _size: int | None  # Size in GB
    _path: IliPath | None
    _product: str | None
    _model: str | None
    _serial: str | None

    def __init__(
        self,
        size: int | None = None,
        path: IliPath | None = None,
        product: str | None = None,
        model: str | None = None,
        serial: str | None = None,
    ) -> None:
        """Initialize the virtual machine disk."""
        self._size = size
        self._path = path
        self._product = product
        self._model = model
        self._serial = serial

    @property
    def size(self) -> int | None:
        """Get the virtual machine disk size."""
        return self._size

    @property
    def path(self) -> IliPath | None:
        """Get the virtual machine disk path."""
        return self._path

    @path.setter
    def path(self, path: IliPath) -> None:
        """Set the virtual machine disk path."""
        self._path = path

    @property
    def product(self) -> str | None:
        """Get the virtual machine disk product."""
        return self._product

    @property
    def model(self) -> str | None:
        """Get the virtual machine disk model."""
        return self._model

    @property
    def serial(self) -> str | None:
        """Get the virtual machine disk serial."""
        return self._serial


class IliVM:
    """Virtual machine class."""

    _temp_paths: list[IliPath]

    # Live image to boot
    _live_image: IliPath | None
    # SSH port
    _ssh_port: int

    # Temporary working dir
    _workdir_base: IliPath | None
    _workdir: IliPath | None
    # Hostfs directory off the working dir
    _hostfs: IliPath | None

    # OEM ISO to override boot
    _oem_iso: IliPath | None

    # OEM files to copy into the OEM ISO
    _oem_files: list[IliVMOEMFile]

    # VM disks
    _disks: list[IliVMDisk]

    # QEMU process
    _qemu: subprocess.Popen | None
    _qmp_socket: IliPath | None
    _state: IliVMState
    _thread_qemu_watcher: threading.Thread | None
    _ssh_privkey: IliPath | None
    _ssh_knownhosts: IliPath | None

    def __init__(self, live_image: IliPath | None = None, work_dir: IliPath | None = None, listen_port: int = 2022) -> None:
        """Initialize the virtual machine."""
        self._temp_paths = []

        self._live_image = live_image
        self._ssh_port = listen_port

        self._workdir_base = work_dir
        self._workdir = None
        self._hostfs = None

        self._oem_iso = None

        self._oem_files = []

        self._disks = []

        self._qemu = None
        self._qmp_socket = None
        self._state = IliVMState.STOPPED
        self._thread_qemu_watcher = None
        self._ssh_privkey = None
        self._ssh_knownhosts = None

    def __del__(self) -> None:
        """Delete the virtual machine."""
        self.cleanup()

    def cleanup(self) -> None:
        """Clean up the virtual machine."""
        # Stop the virtual machine if its running
        if self._qemu:
            logging.info("[HOST] Terminating virtual machine")
            self.stop()
        # Cleanup temp files
        logging.info("[HOST] Deleting temporary directories")
        while self.temp_paths:
            temp_path = self.temp_paths.pop()
            logging.debug("  - Cleaning up temp paths - removing directory: %s", temp_path)
            shutil.rmtree(temp_path)

    def add_oem_file(self, src: IliPath, replaces: dict[str, str] | None = None) -> None:
        """Add a file to the OEM volume."""
        if not src.exists():
            raise IliRuntimeError(f"Source file '{src}' does not exist")
        self._oem_files.append(IliVMOEMFile(src, replaces))

    def add_disk(
        self,
        size: int | None = None,
        path: IliPath | None = None,
        product: str | None = None,
        model: str | None = None,
        serial: str | None = None,
    ) -> None:
        """
        Add a disk to the virtual machine.

        If the `size` is specified the disk will be created if it does not exist.
        """
        self._disks.append(IliVMDisk(size=size, path=path, product=product, model=model, serial=serial))

    def prepare(self) -> None:  # noqa: C901,PLR0912,PLR0915
        """Prepare the virtual machine."""
        # Create a temporary directory for us to work in
        work_basedir = self._workdir_base or os.environ.get("TMPDIR", "/tmp")  # noqa: S108
        try:
            self.workdir = IliPath(tempfile.mkdtemp(prefix="ilitools-vm-", dir=work_basedir))
            self.temp_paths.append(IliPath(self.workdir))
        except OSError as e:
            raise IliRuntimeError(f"Failed to create temporary directory: {e}") from None
        if not self.workdir.exists():
            self.workdir.mkdir(parents=True)
        if not self.workdir.is_dir():
            raise IliRuntimeError(f"Specified working dir is not a directory: {self.workdir}")

        # Create VM disks
        if self.disks:
            logging.info("[HOST] Adding VM disks...")
            # Setup disk directory
            disk_dir = self.workdir.concatpath("disks")
            logging.debug("[HOST] DISKDIR: %s", disk_dir)
            try:
                disk_dir.mkdir()
            except OSError as e:
                raise IliRuntimeError(f"Failed to create disk directory: {e}") from None
            # Create disks
            disknum = 0
            for disk in self.disks:
                disknum = disknum + 1
                # Check for an specified path
                if disk.path:
                    logging.debug("[HOST]   - Disk path specified: %s", disk.path)
                    if disk.path.exists():
                        # Check if we have a size specified, which is invalid
                        if disk.size:
                            raise IliRuntimeError("Disk size specified with existing disk path")
                        # Use the disk we were given
                        logging.debug("[HOST]   - Using existing disk: %s", disk.path)

                        continue
                    # Check if we have a size specified to create the disk, if not throw an error
                    if not disk.size:
                        raise IliRuntimeError("Disk path specified but no size specified to create disk")
                    disk_path = disk.path
                else:
                    # If we don't have a disk path, assign one
                    disk_path = disk_dir.concatpath(f"disk{disknum}.qcow2")

                # Check again if the disk does not exist, if it doesn't create it
                logging.debug("[HOST]   - Creating disk: %s [%sGiB]", disk_path, disk.size)
                qemu_img_command = ["/usr/bin/qemu-img", "create", "-f", "qcow2", str(disk_path), f"{disk.size}G"]
                run_host(qemu_img_command, log_prefix="[HOST]")
                # Save disk path
                disk.path = disk_path

        # Create dir for SSH key
        sshkey_dir = self.workdir.concatpath("ssh")
        self._ssh_privkey = sshkey_dir.concatpath("id_ed25519")
        self._ssh_knownhosts = sshkey_dir.concatpath("known_hosts")
        sshkey_pub = self._ssh_privkey.with_suffix(".pub")
        logging.debug("[HOST] SSHKEYDIR: %s", sshkey_dir)
        try:
            sshkey_dir.mkdir()
        except OSError as e:
            raise IliRuntimeError(f"Failed to create SSH key directory: {e}") from None
        # Generate key
        logging.info("[HOST] Generating VM SSH key...")
        ssh_keygen_command = ["/usr/bin/ssh-keygen", "-N", "", "-t", "ed25519", "-f", str(self._ssh_privkey)]
        run_host(ssh_keygen_command, log_prefix="[HOST]")

        # Create ISO for the OEM volume
        oem_dir = self.workdir.concatpath("oem")
        self.oem_iso = self.workdir.concatpath("oem.iso")
        logging.debug("[HOST] OEMDIR: %s", oem_dir)
        try:
            oem_dir.mkdir()
        except OSError as e:
            raise IliRuntimeError(f"Failed to create OEM directory: {e}") from None

        # Copy in SSH key
        # NK: if add_oem_file gets destination name capability, we can probably replace this bit
        logging.info("[HOST] Copying SSH 'authorized_keys' to OEM...")
        try:
            shutil.copy(sshkey_pub, oem_dir.concatpath("authorized_keys"))
        except OSError as e:
            raise IliRuntimeError(f"Failed to copy SSH key to OEM directory: {e}") from None

        # Copy the rest of the oem files
        logging.info("[HOST] Copying OEM files")
        for oemfile in self._oem_files:
            logging.debug("  - Copying file: %s", str(oemfile))
            try:
                oemfile.copy_to(oem_dir.concatpath(oemfile.path.name))
            except OSError as e:
                raise IliRuntimeError(f"Failed to copy OEM file '{oemfile}' to OEM directory: {e}") from None

        # Create the ISO file
        xorissofs_command = [
            "xorrisofs",
            # "-iso-level", "3",
            # "-full-iso9660-filenames",
            "-joliet",
            # "-joliet-long",
            # "-rational-rock",
            "-volid",
            "OEM",
            "-uid",
            "0",
            "-gid",
            "0",
            "-output",
            str(self.oem_iso),
            str(oem_dir),
        ]
        logging.info("[HOST] Building OEM volume...")
        run_host(xorissofs_command, log_prefix="[HOST]")

        # Set up hostfs
        hostfs_dir = self.workdir.concatpath("hostfs")
        logging.debug("[HOST] HOSTFS: %s", hostfs_dir)
        try:
            hostfs_dir.mkdir()
        except OSError as e:
            raise IliRuntimeError(f"Failed to create hostfs directory: {e}") from None
        self._hostfs = hostfs_dir

    def start(self) -> None:  # noqa: PLR0915
        """Start the virtual machine."""

        # Work out qmp socket to use
        self._qmp_socket = self.workdir.concatpath("qmp.sock")

        qemu_command = ["/usr/bin/qemu-system-x86_64"]
        # Disable default config
        qemu_command.extend(["-no-user-config", "-nodefaults"])
        # Set q35 machine type
        qemu_command.extend(["-machine", "q35,accel=kvm"])
        # Set memory
        qemu_command.extend(["-m", "4096"])
        # Use UEFI bios (for multiple CDROM support)
        qemu_command.extend(["-bios", "/usr/share/ovmf/x64/OVMF.4m.fd"])

        # Add network device
        qemu_command.extend(["-netdev", f"type=user,id=net0,hostfwd=tcp::{self.ssh_port}-:22"])
        qemu_command.extend(["-device", "virtio-net-pci,netdev=net0"])

        # Add host filesystem
        qemu_command.extend(
            [
                "-fsdev",
                f"local,id=fsdev0,path={self.hostfs},security_model=mapped,multidevs=remap",
            ]
        )
        qemu_command.extend(["-device", "virtio-9p-pci,fsdev=fsdev0,mount_tag=host"])

        # Disable graphics
        qemu_command.append("-nographic")

        # Add CDROM's
        cdnum = 1
        # Add live image
        if self.live_image:
            qemu_command.extend(["-device", f"ide-cd,bus=ide.{cdnum - 1},drive=cdrom{cdnum}.format,id=cdrom{cdnum},bootindex=0"])
            qemu_command.extend(["-blockdev", f"driver=file,filename={self.live_image},node-name=cdrom{cdnum}.storage"])
            qemu_command.extend(
                [
                    "-blockdev",
                    f"node-name=cdrom{cdnum}.format,driver=raw,file=cdrom{cdnum}.storage,read-only=on",
                ]
            )
            cdnum = cdnum + 1
        # Add OEM disk (always)
        qemu_command.extend(["-device", f"ide-cd,bus=ide.{cdnum - 1},drive=cdrom{cdnum}.format,id=cdrom{cdnum}"])
        qemu_command.extend(["-blockdev", f"driver=file,filename={self.oem_iso},node-name=cdrom{cdnum}.storage"])
        qemu_command.extend(
            [
                "-blockdev",
                f"node-name=cdrom{cdnum}.format,driver=raw,file=cdrom{cdnum}.storage,read-only=on",
            ]
        )

        # Work out disks to add to qemu command
        if self.disks:
            # Add virtio disk controller
            qemu_command.extend(["-device", "virtio-scsi-pci,id=scsi0"])
            disknum = 1
            for disk in self.disks:
                disk_device_args: list[str] = []
                if disk.product:
                    disk_device_args.append(f"product={disk.product}")
                if disk.model:
                    disk_device_args.append(f"model={disk.model}")
                if disk.serial:
                    disk_device_args.append(f"serial={disk.serial}")
                # Combine the disk device args
                disk_device_args_str = ""
                if disk_device_args:
                    disk_device_args_str = "," + (",".join(disk_device_args))

                qemu_command.extend(
                    [
                        "-device",
                        f"scsi-hd,bus=scsi0.0,drive=disk{disknum}.format,device_id=disk{disknum}.device,id=disk{disknum},bootindex={disknum}{disk_device_args_str}",
                    ]
                )
                qemu_command.extend(
                    [
                        "-blockdev",
                        f"driver=file,filename={disk.path},node-name=disk{disknum}.storage,discard=unmap",
                    ]
                )
                qemu_command.extend(
                    [
                        "-blockdev",
                        f"node-name=disk{disknum}.format,driver=qcow2,file=disk{disknum}.storage,read-only=false,discard=unmap,file=disk{disknum}.storage",
                    ]
                )
                disknum = disknum + 1

        # Add serial port
        qemu_command.extend(["-device", "isa-serial,chardev=serial0"])
        qemu_command.extend(["-chardev", "stdio,id=serial0"])
        # Add QMP monitor API unix socket
        qemu_command.extend(["-qmp", f"unix:{self._qmp_socket},server,nowait"])

        logging.info("[HOST] Starting virtual machine...")
        logging.debug("[HOST]   - Running: %s", qemu_command)
        try:
            self._qemu = subprocess.Popen(  # noqa: S603
                qemu_command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True
            )
        except (OSError, ValueError) as e:
            raise IliRuntimeError(f"Failed to start virtual machine: {e}") from None

        # Start a separate thread to capture output
        self._state = IliVMState.BOOTING
        self._thread_qemu_watcher = threading.Thread(target=self._qemu_watcher)
        self._thread_qemu_watcher.start()

        self._wait_for_vm()
        self._wait_for_vm_ssh()

    def run_script(self, script: IliPath) -> None:
        """Run a script on the virtual machine."""
        logging.info("[HOST] Running script in VM: %s", script)

        # Create random script directory name with 8 random alnum characters
        randsuffix = "".join(random.choices(string.ascii_letters + string.digits, k=8))  # noqa: S311
        local_script_dirname = f"scripts.{randsuffix}"
        local_script_dir = self.workdir.concatpath(local_script_dirname)

        remote_script_dir = IliPath(f"/root/.ilitools-{local_script_dirname}")
        remote_script_path = remote_script_dir.concatpath(script.name)

        # Create scripts directory
        try:
            local_script_dir.mkdir()
        except OSError as e:
            raise IliRuntimeError(f"Failed to create local script directory '{local_script_dir}': {e}") from None

        # Copy script to our scripts directory
        local_script_path = local_script_dir.concatpath(script.name)
        try:
            shutil.copy(script, local_script_path)
        except OSError as e:
            raise IliRuntimeError(f"Failed to copy script to scripts directory: {e}") from None
        # Make script executable
        local_script_path.chmod(0o755)

        # Copy script to remote host
        logging.debug("[HOST]   - Copying script '%s' to VM '%s'", local_script_path.name, remote_script_path)
        self.copy_to_vm(local_script_path, remote_script_path, args=["--mkpath"])

        # Remove local script directory for this script
        try:
            shutil.rmtree(local_script_dir)
        except OSError as e:
            raise IliRuntimeError(f"Failed to remove local script directory '{local_script_dir}': {e}") from None

        # Execute script via SSH on remote host
        self.run_command([str(remote_script_path)])

    def run_command(self, args: list[str]) -> None:
        """Run a command on the virtual machine."""
        # Build full commandline
        ssh_args = self._ssh_args_vm()
        ssh_args.append("root@localhost")
        ssh_args.append("--")
        ssh_args.extend(args)
        # Execute on remote host
        run_host(ssh_args, log_prefix="[HOST]")

    def poweroff(self) -> None:
        """Poweroff the virtual machine."""
        if not self._qemu:
            return
        logging.info("[HOST] Powering off VM via ACPI")
        # Connect to unix socket to send qmp_command
        with socket.socket(socket.AF_UNIX, socket.SOCK_STREAM) as sock:
            try:
                sock.connect(str(self._qmp_socket))
                # Receive greeting
                response = sock.recv(1024)
                logging.debug("[HOST]   - QMP Greeting: %s", response)
                # Send qmp_capabilities
                qmp = {"execute": "qmp_capabilities", "arguments": {"enable": ["oob"]}}
                sock.sendall(json.dumps(qmp).encode() + b"\n")
                # Receive response
                response = sock.recv(1024)
                logging.debug("[HOST]   - QMP Capabilities: %s", response)
                # Send poweroff
                qmp = {"execute": "system_powerdown"}
                sock.sendall(json.dumps(qmp).encode() + b"\n")
                # Get response
                response = sock.recv(1024)
                logging.debug("[HOST]   - QMP Poweroff: %s", response)

            except OSError as e:
                raise IliRuntimeError(f"Failed to send poweroff command to VM: {e}") from None

    def stop(self) -> None:
        """Terminate the QEMU process and wait for thread to finish."""
        if not self._qemu:
            return
        logging.info("[HOST] Stopping VM")
        # Terminate process and wait for it to finish
        logging.debug("[HOST]   - Terminating VM")
        self._qemu.terminate()
        logging.debug("[HOST]   - Waiting for VM to stop")
        self._qemu.wait()
        # Join the watcher and wait for it to exit
        if self._thread_qemu_watcher:
            self._thread_qemu_watcher.join()
        # Clear current process and watcher thread
        self._qemu = None
        self._qmp_socket = None
        self._thread_qemu_watcher = None

    def copy_to_vm(self, src: IliPath, dst: IliPath, args: list[str] | None = None) -> None:
        """Copy a file or path to the virtual machine."""

        logging.debug("[HOST] Copying to VM '%s' => '%s'", src, dst)
        # Copy file to the VM
        rsync_command = [
            "/usr/bin/rsync",
            "--rsh",
            self._rsh_command(),
            "--recursive",
            "--links",
            "--sparse",
        ]
        # Check if we have extra args
        if args:
            rsync_command.extend(args)
        # Add source and dest
        rsync_command.append(str(src))
        rsync_command.append(f"root@localhost:{dst}")
        # Run rsync
        run_host(rsync_command, log_prefix="[HOST]")

    def copy_from_vm(self, src: IliPath, dst: IliPath, args: list[str] | None = None) -> None:
        """Copy a file or path from the virtual machine."""

        logging.debug("[HOST] Copying from VM '%s' => '%s'", src, dst)
        # Copy file to the VM
        rsync_command = [
            "/usr/bin/rsync",
            "--rsh",
            self._rsh_command(),
            "--archive",
            "--sparse",
        ]
        # Check if we have extra args
        if args:
            rsync_command.extend(args)
        # Add source and dest
        rsync_command.append(f"root@localhost:{src}")
        rsync_command.append(str(dst))
        # Run rsync
        run_host(rsync_command, log_prefix="[HOST]")

    def wait_for_vm_poweroff(self, timeout: int | None = None) -> None:
        """Wait for VM to poweroff for a specific timeout, or 0 for indefinitely."""
        if self._qemu is None:
            raise IliRuntimeError("wait_for_vm_poweroff can only be called on a started VM")
        logging.debug("[HOST] Waiting for VM to poweroff normally")
        # Wait for VM to poweroff by waiting for the subprocess to exit for the timeout
        try:
            self._qemu.wait(timeout=timeout)
            logging.info("[HOST] VM has powered off normally")
            self.stop()
        except subprocess.TimeoutExpired:
            logging.warning("[HOST] Timeout waiting for VM to poweroff, terminating")
            self.stop()
            raise IliRuntimeError("Timeout waiting for VM to poweroff") from None

    def _wait_for_vm(self) -> None:
        logging.info("[HOST] Waiting for VM to boot")
        # Wait for VM to finish booting
        timeout = 60
        while timeout:
            logging.debug("[HOST]   - Waiting for VM to boot: state=%s, remaining=%ss", self.state, timeout)
            if self.state == IliVMState.RUNNING:
                logging.info("[HOST] VM is now running")
                break
            timeout = timeout - 1
            time.sleep(1)
        if not timeout:
            logging.warning("[HOST] Timeout waiting for VM to boot, terminating")
            self.stop()
            raise IliRuntimeError("Timeout waiting for VM to boot")

    def _wait_for_vm_ssh(self) -> None:
        logging.info("[HOST] Waiting for connection to VM")
        # Wait until we get a connection to the port
        timeout = 30
        while timeout:
            logging.debug("[HOST]   - Waiting for connection to VM: remaining=%ss", timeout)
            # Try connect over TCP/IP with a 1s timeout
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
                sock.settimeout(1)
                try:
                    sock.connect(("localhost", self.ssh_port))
                    # Try read raw SSH banner
                    rb = sock.recv(1024)
                    if rb:
                        logging.info("[HOST] VM connection is UP")
                        break
                except (ConnectionRefusedError, ConnectionResetError):
                    pass
                except InterruptedError:
                    raise IliRuntimeError("Interrupted while checking OpenSSH connectivity") from None
            timeout = timeout - 1
            time.sleep(1)
        if not timeout:
            logging.warning("[HOST] Timeout waiting for OpenSSH to start, terminating")
            self.stop()
            raise IliRuntimeError("Timeout waiting for OpenSSH to start")

    def _qemu_watcher(self) -> None:
        """Watch qemu output."""
        # Make sure we have a qemu handle
        if self._qemu is None or self._qemu.stdout is None:
            return
        # Loop with output
        while True:
            # Grab line
            line = self._qemu.stdout.readline()
            if not line:
                break
            # Sanitize line
            line_clean = sanitize_serial_output(line)
            if not line_clean:
                continue
            # Check for boot completion
            if self.state == IliVMState.BOOTING and re.match(r"VM READY", line_clean):
                self._state = IliVMState.RUNNING
            # Log line
            logging.debug("[ VM ]   <<<< %s", line_clean)
        # Close handle once we're done
        self._qemu.stdout.close()

    def _ssh_args_vm(self, port_arg: str = "-p") -> list[str]:
        """Get SSH arguments."""
        return [
            "/usr/bin/ssh",
            "-o",
            "StrictHostKeyChecking=no",
            "-o",
            "PasswordAuthentication=no",
            "-o",
            f"UserKnownHostsFile={self._ssh_knownhosts}",
            "-i",
            str(self._ssh_privkey),
            port_arg,
            str(self.ssh_port),
        ]

    def _rsh_command(self) -> str:
        """Get RSH command."""
        return " ".join([shlex.quote(x) for x in self._ssh_args_vm()])

    @property
    def temp_paths(self) -> list[IliPath]:
        """Get the virtual machine temporary directories created."""
        return self._temp_paths

    @property
    def live_image(self) -> IliPath | None:
        """Get the virtual machine live image."""
        return self._live_image

    @live_image.setter
    def live_image(self, live_image: IliPath) -> None:
        """Set the virtual machine live image."""
        self._live_image = live_image

    @property
    def ssh_port(self) -> int:
        """Get the virtual machine SSH listen port."""
        return self._ssh_port

    @property
    def workdir(self) -> IliPath:
        """Get the virtual machine work directory."""
        if self._workdir is None:
            raise IliRuntimeError("Virtual machine work used when not prepare()'d")
        return self._workdir

    @workdir.setter
    def workdir(self, workdir: IliPath) -> None:
        """Set the virtual machine work directory."""
        self._workdir = workdir

    @property
    def hostfs(self) -> IliPath:
        """Get the virtual machine hostfs directory."""
        if self._hostfs is None:
            raise IliRuntimeError("Virtual machine hostfs used when not prepare()'d")
        return self._hostfs

    @property
    def oem_iso(self) -> IliPath:
        """Get the virtual machine OEM ISO."""
        if self._oem_iso is None:
            raise IliRuntimeError("Virtual machine OEM ISO used when not prepare()'d")
        return self._oem_iso

    @oem_iso.setter
    def oem_iso(self, oem_iso: IliPath) -> None:
        """Set the virtual machine OEM ISO."""
        self._oem_iso = oem_iso

    @property
    def disks(self) -> list[IliVMDisk]:
        """Get the virtual machine disks."""
        return self._disks

    @property
    def state(self) -> IliVMState:
        """Get the virtual machine state."""
        return self._state
